@@ -5,6 +5,7 @@
 
 #include <boost/algorithm/string/join.hpp>
 #include <fmt/format.h>
+#include <fmt/ranges.h>
 #include <seastar/net/api.hh>
 #include <seastar/net/inet_address.hh>
 #include <seastar/core/future-util.hh>
@@ -13,6 +14,7 @@
 #include <seastar/core/thread.hh>
 #include <seastar/util/std-compat.hh>
 
+#include "common/options.h"
 #include "common/version.h"
 #include "messages/MCommand.h"
 #include "messages/MCommandReply.h"
@@ -20,6 +22,12 @@
 #include "crimson/net/Socket.h"
 
 using namespace crimson::common;
+using namespace std::literals;
+using ceph::common::cmdmap_from_json;
+using ceph::common::cmd_getval;
+using ceph::common::bad_cmd_get;
+using ceph::common::validate_cmd;
+using ceph::common::dump_cmd_and_help_to_json;
 
 namespace {
 seastar::logger& logger()
@@ -28,6 +36,11 @@ seastar::logger& logger()
 }
 }  // namespace
 
+using std::string;
+using std::string_view;
+using std::stringstream;
+using std::unique_ptr;
+
 namespace crimson::admin {
 
 tell_result_t::tell_result_t(int ret, std::string&& err)
@@ -43,25 +56,14 @@ tell_result_t::tell_result_t(std::unique_ptr<Formatter> formatter)
   formatter->flush(out);
 }
 
-seastar::future<>
-AdminSocket::register_command(std::unique_ptr<AdminSocketHook>&& hook)
+void AdminSocket::register_command(std::unique_ptr<AdminSocketHook>&& hook)
 {
-  return seastar::with_lock(servers_tbl_rwlock,
-			    [this, hook = std::move(hook)]() mutable {
-    auto prefix = hook->prefix;
-    auto [it, added] = hooks.emplace(prefix, std::move(hook));
-    //  was this server tag already registered?
-    assert(added);
-    if (added) {
-      logger().info("register_command(): {})", it->first);
-    }
-    return seastar::now();
-  });
+  auto prefix = hook->prefix;
+  auto [it, added] = hooks.emplace(prefix, std::move(hook));
+  assert(added);
+  logger().info("register_command(): {})", it->first);
 }
 
-/*
- * Note: parse_cmd() is executed with servers_tbl_rwlock held as shared
- */
 auto AdminSocket::parse_cmd(const std::vector<std::string>& cmd)
   -> std::variant<parsed_command_t, tell_result_t>
 {
@@ -92,6 +94,10 @@ auto AdminSocket::parse_cmd(const std::vector<std::string>& cmd)
   try {
     cmd_getval(cmdmap, "format", format);
     cmd_getval(cmdmap, "prefix", prefix);
+    // tolerate old-style pg <pgid> command <args> style formatting
+    if (prefix == "pg") {
+      cmd_getval(cmdmap, "cmd", prefix);
+    }
   } catch (const bad_cmd_get& e) {
     logger().error("{}: invalid syntax: {}", __func__, cmd);
     out.append(string{e.what()});
@@ -118,7 +124,8 @@ seastar::future<> AdminSocket::finalize_response(
   uint32_t response_length = htonl(outbuf_cont.length());
   logger().info("asok response length: {}", outbuf_cont.length());
 
-  return out.write((char*)&response_length, sizeof(uint32_t))
+  return out.write(reinterpret_cast<char*>(&response_length),
+                   sizeof(response_length))
     .then([&out, outbuf_cont] { return out.write(outbuf_cont.c_str()); });
 }
 
@@ -129,17 +136,17 @@ seastar::future<> AdminSocket::handle_command(crimson::net::ConnectionRef conn,
   return execute_command(m->cmd, std::move(m->get_data())).then(
     [conn, tid=m->get_tid()](auto result) {
     auto [ret, err, out] = std::move(result);
-    auto reply = make_message<MCommandReply>(ret, err);
+    auto reply = crimson::make_message<MCommandReply>(ret, err);
     reply->set_tid(tid);
     reply->set_data(out);
-    return conn->send(reply);
+    return conn->send(std::move(reply));
   });
 }
 
 seastar::future<> AdminSocket::execute_line(std::string cmdline,
                                             seastar::output_stream<char>& out)
 {
-  return execute_command({cmdline}, {}).then([&out, this](auto result) {
+  return execute_command({std::move(cmdline)}, {}).then([&out, this](auto result) {
      auto [ret, stderr, stdout] = std::move(result);
      if (ret < 0) {
        stdout.append(fmt::format("ERROR: {}\n", cpp_strerror(ret)));
@@ -153,26 +160,23 @@ auto AdminSocket::execute_command(const std::vector<std::string>& cmd,
 				  ceph::bufferlist&& buf)
   -> seastar::future<tell_result_t>
 {
-  return seastar::with_shared(servers_tbl_rwlock,
-			      [cmd, buf=std::move(buf), this]() mutable {
-    auto maybe_parsed = parse_cmd(cmd);
-    if (auto parsed = std::get_if<parsed_command_t>(&maybe_parsed); parsed) {
-      stringstream os;
-      string desc{parsed->hook.desc};
-      if (!validate_cmd(nullptr, desc, parsed->params, os)) {
-	logger().error("AdminSocket::execute_command: "
-		       "failed to validate '{}': {}", cmd, os.str());
-	ceph::bufferlist out;
-	out.append(os);
-	return seastar::make_ready_future<tell_result_t>(
-          tell_result_t{-EINVAL, "invalid command json", std::move(out)});
-      }
-      return parsed->hook.call(parsed->params, parsed->format, std::move(buf));
-    } else {
-      auto& result = std::get<tell_result_t>(maybe_parsed);
-      return seastar::make_ready_future<tell_result_t>(std::move(result));
+  auto maybe_parsed = parse_cmd(cmd);
+  if (auto* parsed = std::get_if<parsed_command_t>(&maybe_parsed); parsed) {
+    stringstream os;
+    string desc{parsed->hook.desc};
+    if (!validate_cmd(desc, parsed->params, os)) {
+      logger().error("AdminSocket::execute_command: "
+                     "failed to validate '{}': {}", cmd, os.str());
+      ceph::bufferlist out;
+      out.append(os);
+      return seastar::make_ready_future<tell_result_t>(
+        tell_result_t{-EINVAL, "invalid command json", std::move(out)});
     }
-  });
+    return parsed->hook.call(parsed->params, parsed->format, std::move(buf));
+  } else {
+    auto& result = std::get<tell_result_t>(maybe_parsed);
+    return seastar::make_ready_future<tell_result_t>(std::move(result));
+  }
 }
 
 // an input_stream consumer that reads buffer into a std::string up to the first
@@ -232,35 +236,42 @@ seastar::future<> AdminSocket::start(const std::string& path)
   try {
     server_sock = seastar::engine().listen(sock_path);
   } catch (const std::system_error& e) {
+    if (e.code() == std::errc::address_in_use) {
+      logger().debug("{}: Admin Socket socket path={} already exists, retrying",
+                     __func__, path);
+      return seastar::remove_file(path).then([this, path] {
+        server_sock.reset();
+        return start(path);
+      });
+    }
     logger().error("{}: unable to listen({}): {}", __func__, path, e.what());
     server_sock.reset();
     return seastar::make_ready_future<>();
   }
   // listen in background
-  task = seastar::do_until(
-    [this] { return stop_gate.is_closed(); },
-    [this] {
-      return seastar::with_gate(stop_gate, [this] {
-        assert(!connected_sock.has_value());
-        return server_sock->accept().then([this](seastar::accept_result acc) {
-          connected_sock = std::move(acc.connection);
-          return seastar::do_with(connected_sock->input(),
-                                  connected_sock->output(),
-            [this](auto& input, auto& output) mutable {
-            return handle_client(input, output);
-          }).finally([this] {
-            assert(connected_sock.has_value());
-            connected_sock.reset();
-          });
-        }).handle_exception([this](auto ep) {
-          if (!stop_gate.is_closed()) {
-            logger().error("AdminSocket: terminated: {}", ep);
-          }
+  task = seastar::keep_doing([this] {
+    return seastar::try_with_gate(stop_gate, [this] {
+      assert(!connected_sock.has_value());
+      return server_sock->accept().then([this](seastar::accept_result acc) {
+        connected_sock = std::move(acc.connection);
+        return seastar::do_with(connected_sock->input(),
+                                connected_sock->output(),
+          [this](auto& input, auto& output) mutable {
+          return handle_client(input, output);
+        }).finally([this] {
+          assert(connected_sock.has_value());
+          connected_sock.reset();
         });
+      }).handle_exception([this](auto ep) {
+        if (!stop_gate.is_closed()) {
+          logger().error("AdminSocket: terminated: {}", ep);
+        }
       });
-    }).finally([path] {
-      return seastar::remove_file(path);
     });
+  }).handle_exception_type([](const seastar::gate_closed_exception&) {
+  }).finally([path] {
+    return seastar::remove_file(path);
+  });
   return seastar::make_ready_future<>();
 }
 
@@ -340,18 +351,16 @@ class HelpHook final : public AdminSocketHook {
 				      std::string_view format,
 				      ceph::bufferlist&&) const final
   {
-    return seastar::with_shared(m_as.servers_tbl_rwlock,
-				[format, this] {
-      unique_ptr<Formatter> f{Formatter::create(format, "json-pretty", "json-pretty")};
-      f->open_object_section("help");
-      for (const auto& [prefix, hook] : m_as) {
-        if (!hook->help.empty()) {
-          f->dump_string(prefix.data(), hook->help);
-	}
+    unique_ptr<Formatter> f{Formatter::create(format,
+					      "json-pretty", "json-pretty")};
+    f->open_object_section("help");
+    for (const auto& [prefix, hook] : m_as) {
+      if (!hook->help.empty()) {
+        f->dump_string(prefix.data(), hook->help);
       }
-      f->close_section();
-      return seastar::make_ready_future<tell_result_t>(std::move(f));
-    });
+    }
+    f->close_section();
+    return seastar::make_ready_future<tell_result_t>(std::move(f));
   }
 };
 
@@ -369,20 +378,18 @@ class GetdescsHook final : public AdminSocketHook {
 				      std::string_view format,
 				      ceph::bufferlist&&) const final
   {
-    return seastar::with_shared(m_as.servers_tbl_rwlock, [format, this] {
-      unique_ptr<Formatter> f{Formatter::create(format, "json-pretty", "json-pretty")};
-      int cmdnum = 0;
-      f->open_object_section("command_descriptions");
-      for (const auto& [prefix, hook] : m_as) {
-	auto secname = fmt::format("cmd {:>03}", cmdnum);
-        auto cmd = fmt::format("{} {}", hook->prefix, hook->desc);
-        dump_cmd_and_help_to_json(f.get(), CEPH_FEATURES_ALL, secname,
-                                  cmd, std::string{hook->help});
-        cmdnum++;
-      }
-      f->close_section();
-      return seastar::make_ready_future<tell_result_t>(std::move(f));
-    });
+    unique_ptr<Formatter> f{Formatter::create(format, "json-pretty", "json-pretty")};
+    int cmdnum = 0;
+    f->open_object_section("command_descriptions");
+    for (const auto& [prefix, hook] : m_as) {
+      auto secname = fmt::format("cmd {:>03}", cmdnum);
+      auto cmd = fmt::format("{} {}", hook->prefix, hook->desc);
+      dump_cmd_and_help_to_json(f.get(), CEPH_FEATURES_ALL, secname,
+				cmd, std::string{hook->help});
+      cmdnum++;
+    }
+    f->close_section();
+    return seastar::make_ready_future<tell_result_t>(std::move(f));
   }
 };
 
@@ -501,21 +508,43 @@ public:
   }
 };
 
+/**
+ * listing the configuration values
+ */
+class ConfigHelpHook : public AdminSocketHook {
+public:
+  ConfigHelpHook() :
+    AdminSocketHook{"config help",
+                    "",
+                    "get config setting schema and descriptions"}
+  {}
+  seastar::future<tell_result_t> call(const cmdmap_t&,
+                                      std::string_view format,
+                                      ceph::bufferlist&& input) const final
+  {
+    unique_ptr<Formatter> f{Formatter::create(format, "json-pretty", "json-pretty")};
+    // Output all
+    f->open_array_section("options");
+    for (const auto &option : ceph_options) {
+      f->dump_object("option", option);
+    }
+    f->close_section();
+    return seastar::make_ready_future<tell_result_t>(std::move(f));
+  }
+};
+
 /// the hooks that are served directly by the admin_socket server
-seastar::future<> AdminSocket::register_admin_commands()
+void AdminSocket::register_admin_commands()
 {
-  return seastar::when_all_succeed(
-    register_command(std::make_unique<VersionHook>()),
-    register_command(std::make_unique<GitVersionHook>()),
-    register_command(std::make_unique<HelpHook>(*this)),
-    register_command(std::make_unique<GetdescsHook>(*this)),
-    register_command(std::make_unique<ConfigGetHook>()),
-    register_command(std::make_unique<ConfigSetHook>()),
-    register_command(std::make_unique<ConfigShowHook>()),
-    register_command(std::make_unique<InjectArgsHook>())
-  ).then_unpack([] {
-    return seastar::now();
-  });
+  register_command(std::make_unique<VersionHook>());
+  register_command(std::make_unique<GitVersionHook>());
+  register_command(std::make_unique<HelpHook>(*this));
+  register_command(std::make_unique<GetdescsHook>(*this));
+  register_command(std::make_unique<ConfigGetHook>());
+  register_command(std::make_unique<ConfigSetHook>());
+  register_command(std::make_unique<ConfigShowHook>());
+  register_command(std::make_unique<ConfigHelpHook>());
+  register_command(std::make_unique<InjectArgsHook>());
 }
 
 }  // namespace crimson::admin
