@@ -17,10 +17,11 @@
 
 namespace {
   seastar::logger& logger() {
-    return crimson::get_logger(ceph_subsys_filestore);
+    return crimson::get_logger(ceph_subsys_cyanstore);
   }
 }
 
+using std::string;
 using crimson::common::local_conf;
 
 namespace crimson::os {
@@ -33,57 +34,51 @@ CyanStore::CyanStore(const std::string& path)
 
 CyanStore::~CyanStore() = default;
 
-seastar::future<> CyanStore::mount()
-{
-  ceph::bufferlist bl;
-  std::string fn = path + "/collections";
-  std::string err;
-  if (int r = bl.read_file(fn.c_str(), &err); r < 0) {
-    throw std::runtime_error("read_file");
-  }
-
-  std::set<coll_t> collections;
-  auto p = bl.cbegin();
-  ceph::decode(collections, p);
-
-  for (auto& coll : collections) {
-    std::string fn = fmt::format("{}/{}", path, coll);
-    ceph::bufferlist cbl;
-    if (int r = cbl.read_file(fn.c_str(), &err); r < 0) {
-      throw std::runtime_error("read_file");
+template <const char* MsgV>
+struct singleton_ec : std::error_code {
+  singleton_ec()
+    : error_code(42, this_error_category{}) {
+  };
+private:
+  struct this_error_category : std::error_category {
+    const char* name() const noexcept final {
+      // XXX: we could concatenate with MsgV at compile-time but the burden
+      // isn't worth the benefit.
+      return "singleton_ec";
     }
-    boost::intrusive_ptr<Collection> c{new Collection{coll}};
-    auto p = cbl.cbegin();
-    c->decode(p);
-    coll_map[coll] = c;
-    used_bytes += c->used_bytes();
-  }
-  return seastar::now();
-}
+    std::string message([[maybe_unused]] const int ev) const final {
+      assert(ev == 42);
+      return MsgV;
+    }
+  };
+};
 
-seastar::future<> CyanStore::umount()
+seastar::future<store_statfs_t> CyanStore::stat() const
 {
-  return seastar::do_with(std::set<coll_t>{}, [this](auto& collections) {
-    return seastar::do_for_each(coll_map, [&collections, this](auto& coll) {
-      auto& [col, ch] = coll;
-      collections.insert(col);
-      ceph::bufferlist bl;
-      ceph_assert(ch);
-      ch->encode(bl);
-      std::string fn = fmt::format("{}/{}", path, col);
-      return crimson::write_file(std::move(bl), fn);
-    }).then([&collections, this] {
-      ceph::bufferlist bl;
-      ceph::encode(collections, bl);
-      std::string fn = fmt::format("{}/collections", path);
-      return crimson::write_file(std::move(bl), fn);
-    });
+  ceph_assert(seastar::this_shard_id() == primary_core);
+  logger().debug("{}", __func__);
+  return shard_stores.map_reduce0(
+    [](const CyanStore::Shard &local_store) {
+      return local_store.get_used_bytes();
+    },
+    (uint64_t)0,
+    std::plus<uint64_t>()
+  ).then([](uint64_t used_bytes) {
+    store_statfs_t st;
+    st.total = crimson::common::local_conf().get_val<Option::size_t>("memstore_device_bytes");
+    st.available = st.total - used_bytes;
+    return seastar::make_ready_future<store_statfs_t>(std::move(st));
   });
 }
 
-seastar::future<> CyanStore::mkfs(uuid_d new_osd_fsid)
+
+CyanStore::mkfs_ertr::future<> CyanStore::mkfs(uuid_d new_osd_fsid)
 {
-  return read_meta("fsid").then([=](auto&& ret) {
+  ceph_assert(seastar::this_shard_id() == primary_core);
+  static const char read_meta_errmsg[]{"read_meta"};
+  static const char parse_fsid_errmsg[]{"failed to parse fsid"};
+  static const char match_ofsid_errmsg[]{"unmatched osd_fsid"};
+  return read_meta("fsid").then([=, this](auto&& ret) -> mkfs_ertr::future<> {
     auto& [r, fsid_str] = ret;
     if (r == -ENOENT) {
       if (new_osd_fsid.is_zero()) {
@@ -93,43 +88,115 @@ seastar::future<> CyanStore::mkfs(uuid_d new_osd_fsid)
       }
       return write_meta("fsid", fmt::format("{}", osd_fsid));
     } else if (r < 0) {
-      throw std::runtime_error("read_meta");
+      return crimson::stateful_ec{ singleton_ec<read_meta_errmsg>() };
     } else {
-      logger().info("{} already has fsid {}", __func__, fsid_str);
+      logger().info("mkfs already has fsid {}", fsid_str);
       if (!osd_fsid.parse(fsid_str.c_str())) {
-        throw std::runtime_error("failed to parse fsid");
+        return crimson::stateful_ec{ singleton_ec<parse_fsid_errmsg>() };
       } else if (osd_fsid != new_osd_fsid) {
         logger().error("on-disk fsid {} != provided {}", osd_fsid, new_osd_fsid);
-        throw std::runtime_error("unmatched osd_fsid");
+        return crimson::stateful_ec{ singleton_ec<match_ofsid_errmsg>() };
       } else {
-	return seastar::now();
+        return mkfs_ertr::now();
       }
     }
-  }).then([this]{
-    std::string fn = path + "/collections";
-    ceph::bufferlist bl;
-    std::set<coll_t> collections;
-    ceph::encode(collections, bl);
-    return crimson::write_file(std::move(bl), fn);
-  }).then([this] {
+  }).safe_then([this]{
     return write_meta("type", "memstore");
+  }).safe_then([this] {
+    return shard_stores.invoke_on_all(
+      [](auto &local_store) {
+      return local_store.mkfs();
+    });
   });
 }
 
-seastar::future<store_statfs_t> CyanStore::stat() const
+seastar::future<> CyanStore::Shard::mkfs()
 {
-  logger().debug("{}", __func__);
-  store_statfs_t st;
-  st.total = crimson::common::local_conf().get_val<Option::size_t>("memstore_device_bytes");
-  st.available = st.total - used_bytes;
-  return seastar::make_ready_future<store_statfs_t>(std::move(st));
+  std::string fn =
+    path + "/collections" + std::to_string(seastar::this_shard_id());
+  ceph::bufferlist bl;
+  std::set<coll_t> collections;
+  ceph::encode(collections, bl);
+  return crimson::write_file(std::move(bl), fn);
+}
+
+using coll_core_t = FuturizedStore::coll_core_t;
+seastar::future<std::vector<coll_core_t>>
+CyanStore::list_collections()
+{
+  ceph_assert(seastar::this_shard_id() == primary_core);
+  return seastar::do_with(std::vector<coll_core_t>{}, [this](auto &collections) {
+    return shard_stores.map([](auto &local_store) {
+      return local_store.list_collections();
+    }).then([&collections](std::vector<std::vector<coll_core_t>> results) {
+      for (auto& colls : results) {
+        collections.insert(collections.end(), colls.begin(), colls.end());
+      }
+      return seastar::make_ready_future<std::vector<coll_core_t>>(
+        std::move(collections));
+    });
+  });
+}
+
+CyanStore::mount_ertr::future<> CyanStore::Shard::mount()
+{
+  static const char read_file_errmsg[]{"read_file"};
+  ceph::bufferlist bl;
+  std::string fn =
+    path + "/collections" + std::to_string(seastar::this_shard_id());
+  std::string err;
+  if (int r = bl.read_file(fn.c_str(), &err); r < 0) {
+    return crimson::stateful_ec{ singleton_ec<read_file_errmsg>() };
+  }
+
+  std::set<coll_t> collections;
+  auto p = bl.cbegin();
+  ceph::decode(collections, p);
+
+  for (auto& coll : collections) {
+    std::string fn = fmt::format("{}/{}{}", path, coll,
+      std::to_string(seastar::this_shard_id()));
+    ceph::bufferlist cbl;
+    if (int r = cbl.read_file(fn.c_str(), &err); r < 0) {
+      return crimson::stateful_ec{ singleton_ec<read_file_errmsg>() };
+    }
+    boost::intrusive_ptr<Collection> c{new Collection{coll}};
+    auto p = cbl.cbegin();
+    c->decode(p);
+    coll_map[coll] = c;
+    used_bytes += c->used_bytes();
+  }
+  return mount_ertr::now();
+}
+
+seastar::future<> CyanStore::Shard::umount()
+{
+  return seastar::do_with(std::set<coll_t>{}, [this](auto& collections) {
+    return seastar::do_for_each(coll_map, [&collections, this](auto& coll) {
+      auto& [col, ch] = coll;
+      collections.insert(col);
+      ceph::bufferlist bl;
+      ceph_assert(ch);
+      ch->encode(bl);
+      std::string fn = fmt::format("{}/{}{}", path, col,
+        std::to_string(seastar::this_shard_id()));
+      return crimson::write_file(std::move(bl), fn);
+    }).then([&collections, this] {
+      ceph::bufferlist bl;
+      ceph::encode(collections, bl);
+      std::string fn = fmt::format("{}/collections{}",
+        path, std::to_string(seastar::this_shard_id()));
+      return crimson::write_file(std::move(bl), fn);
+    });
+  });
 }
 
 seastar::future<std::tuple<std::vector<ghobject_t>, ghobject_t>>
-CyanStore::list_objects(CollectionRef ch,
-                        const ghobject_t& start,
-                        const ghobject_t& end,
-                        uint64_t limit) const
+CyanStore::Shard::list_objects(
+  CollectionRef ch,
+  const ghobject_t& start,
+  const ghobject_t& end,
+  uint64_t limit) const
 {
   auto c = static_cast<Collection*>(ch.get());
   logger().debug("{} {} {} {} {}",
@@ -151,28 +218,32 @@ CyanStore::list_objects(CollectionRef ch,
     std::make_tuple(std::move(objects), next));
 }
 
-seastar::future<CollectionRef> CyanStore::create_new_collection(const coll_t& cid)
+seastar::future<CollectionRef>
+CyanStore::Shard::create_new_collection(const coll_t& cid)
 {
   auto c = new Collection{cid};
   new_coll_map[cid] = c;
   return seastar::make_ready_future<CollectionRef>(c);
 }
 
-seastar::future<CollectionRef> CyanStore::open_collection(const coll_t& cid)
+seastar::future<CollectionRef>
+CyanStore::Shard::open_collection(const coll_t& cid)
 {
   return seastar::make_ready_future<CollectionRef>(_get_collection(cid));
 }
 
-seastar::future<std::vector<coll_t>> CyanStore::list_collections()
+seastar::future<std::vector<coll_core_t>>
+CyanStore::Shard::list_collections()
 {
-  std::vector<coll_t> collections;
+  std::vector<coll_core_t> collections;
   for (auto& coll : coll_map) {
-    collections.push_back(coll.first);
+    collections.push_back(std::make_pair(coll.first, seastar::this_shard_id()));
   }
-  return seastar::make_ready_future<std::vector<coll_t>>(std::move(collections));
+  return seastar::make_ready_future<std::vector<coll_core_t>>(std::move(collections));
 }
 
-CyanStore::read_errorator::future<ceph::bufferlist> CyanStore::read(
+CyanStore::Shard::read_errorator::future<ceph::bufferlist>
+CyanStore::Shard::read(
   CollectionRef ch,
   const ghobject_t& oid,
   uint64_t offset,
@@ -199,7 +270,8 @@ CyanStore::read_errorator::future<ceph::bufferlist> CyanStore::read(
   return read_errorator::make_ready_future<ceph::bufferlist>(o->read(offset, l));
 }
 
-CyanStore::read_errorator::future<ceph::bufferlist> CyanStore::readv(
+CyanStore::Shard::read_errorator::future<ceph::bufferlist>
+CyanStore::Shard::readv(
   CollectionRef ch,
   const ghobject_t& oid,
   interval_set<uint64_t>& m,
@@ -219,8 +291,8 @@ CyanStore::read_errorator::future<ceph::bufferlist> CyanStore::readv(
   });
 }
 
-
-CyanStore::get_attr_errorator::future<ceph::bufferptr> CyanStore::get_attr(
+CyanStore::Shard::get_attr_errorator::future<ceph::bufferlist>
+CyanStore::Shard::get_attr(
   CollectionRef ch,
   const ghobject_t& oid,
   std::string_view name) const
@@ -233,13 +305,14 @@ CyanStore::get_attr_errorator::future<ceph::bufferptr> CyanStore::get_attr(
     return crimson::ct_error::enoent::make();
   }
   if (auto found = o->xattr.find(name); found != o->xattr.end()) {
-    return get_attr_errorator::make_ready_future<ceph::bufferptr>(found->second);
+    return get_attr_errorator::make_ready_future<ceph::bufferlist>(found->second);
   } else {
     return crimson::ct_error::enodata::make();
   }
 }
 
-CyanStore::get_attrs_ertr::future<CyanStore::attrs_t> CyanStore::get_attrs(
+CyanStore::Shard::get_attrs_ertr::future<CyanStore::Shard::attrs_t>
+CyanStore::Shard::get_attrs(
   CollectionRef ch,
   const ghobject_t& oid)
 {
@@ -253,9 +326,10 @@ CyanStore::get_attrs_ertr::future<CyanStore::attrs_t> CyanStore::get_attrs(
   return get_attrs_ertr::make_ready_future<attrs_t>(o->xattr);
 }
 
-auto CyanStore::omap_get_values(CollectionRef ch,
-				const ghobject_t& oid,
-				const omap_keys_t& keys)
+auto CyanStore::Shard::omap_get_values(
+  CollectionRef ch,
+  const ghobject_t& oid,
+  const omap_keys_t& keys)
   -> read_errorator::future<omap_values_t>
 {
   auto c = static_cast<Collection*>(ch.get());
@@ -273,11 +347,11 @@ auto CyanStore::omap_get_values(CollectionRef ch,
   return seastar::make_ready_future<omap_values_t>(std::move(values));
 }
 
-auto
-CyanStore::omap_get_values(CollectionRef ch,
-			   const ghobject_t &oid,
-			   const std::optional<string> &start)
-  -> read_errorator::future<std::tuple<bool, omap_values_t>>
+auto CyanStore::Shard::omap_get_values(
+  CollectionRef ch,
+  const ghobject_t &oid,
+  const std::optional<string> &start)
+  -> CyanStore::Shard::read_errorator::future<std::tuple<bool, omap_values_t>>
 {
   auto c = static_cast<Collection*>(ch.get());
   logger().debug("{} {} {}", __func__, c->get_cid(), oid);
@@ -287,7 +361,7 @@ CyanStore::omap_get_values(CollectionRef ch,
   }
   omap_values_t values;
   for (auto i = start ? o->omap.upper_bound(*start) : o->omap.begin();
-       values.size() < MAX_KEYS_PER_OMAP_GET_CALL && i != o->omap.end();
+       i != o->omap.end();
        ++i) {
     values.insert(*i);
   }
@@ -295,10 +369,10 @@ CyanStore::omap_get_values(CollectionRef ch,
     std::make_tuple(true, std::move(values)));
 }
 
-auto
-CyanStore::omap_get_header(CollectionRef ch,
-			   const ghobject_t& oid)
-  -> read_errorator::future<ceph::bufferlist>
+auto CyanStore::Shard::omap_get_header(
+  CollectionRef ch,
+  const ghobject_t& oid)
+  -> CyanStore::Shard::get_attr_errorator::future<ceph::bufferlist>
 {
   auto c = static_cast<Collection*>(ch.get());
   auto o = c->get_object(oid);
@@ -306,12 +380,13 @@ CyanStore::omap_get_header(CollectionRef ch,
     return crimson::ct_error::enoent::make();
   }
 
-  return read_errorator::make_ready_future<ceph::bufferlist>(
+  return get_attr_errorator::make_ready_future<ceph::bufferlist>(
     o->omap_header);
 }
 
-seastar::future<> CyanStore::do_transaction(CollectionRef ch,
-                                            ceph::os::Transaction&& t)
+seastar::future<> CyanStore::Shard::do_transaction_no_callbacks(
+  CollectionRef ch,
+  ceph::os::Transaction&& t)
 {
   using ceph::os::Transaction;
   int r = 0;
@@ -333,6 +408,7 @@ seastar::future<> CyanStore::do_transaction(CollectionRef ch,
       }
       break;
       case Transaction::OP_TOUCH:
+      case Transaction::OP_CREATE:
       {
         coll_t cid = i.get_cid(op->cid);
         ghobject_t oid = i.get_oid(op->oid);
@@ -368,6 +444,14 @@ seastar::future<> CyanStore::do_transaction(CollectionRef ch,
         r = _truncate(cid, oid, off);
       }
       break;
+      case Transaction::OP_CLONE:
+      {
+        coll_t cid = i.get_cid(op->cid);
+        ghobject_t oid = i.get_oid(op->oid);
+        ghobject_t noid = i.get_oid(op->dest_oid);
+        r = _clone(cid, oid, noid);
+      }
+      break;
       case Transaction::OP_SETATTR:
       {
         coll_t cid = i.get_cid(op->cid);
@@ -375,9 +459,18 @@ seastar::future<> CyanStore::do_transaction(CollectionRef ch,
         std::string name = i.decode_string();
         ceph::bufferlist bl;
         i.decode_bl(bl);
-        std::map<std::string, bufferptr> to_set;
-        to_set[name] = bufferptr(bl.c_str(), bl.length());
-        r = _setattrs(cid, oid, to_set);
+        std::map<std::string, bufferlist> to_set;
+        to_set.emplace(name, std::move(bl));
+        r = _setattrs(cid, oid, std::move(to_set));
+      }
+      break;
+      case Transaction::OP_SETATTRS:
+      {
+        coll_t cid = i.get_cid(op->cid);
+        ghobject_t oid = i.get_oid(op->oid);
+        std::map<std::string, bufferlist> aset;
+        i.decode_attrset(aset);
+        r = _setattrs(cid, oid, std::move(aset));
       }
       break;
       case Transaction::OP_RMATTR:
@@ -388,12 +481,24 @@ seastar::future<> CyanStore::do_transaction(CollectionRef ch,
         r = _rm_attr(cid, oid, name);	
       }
       break;
+      case Transaction::OP_RMATTRS:
+      {
+        coll_t cid = i.get_cid(op->cid);
+        ghobject_t oid = i.get_oid(op->oid);
+        r = _rm_attrs(cid, oid);
+      }
+      break;
       case Transaction::OP_MKCOLL:
       {
         coll_t cid = i.get_cid(op->cid);
         r = _create_collection(cid, op->split_bits);
       }
       break;
+      case Transaction::OP_SETALLOCHINT:
+      {
+        r = 0;
+      }
+      break;
       case Transaction::OP_OMAP_CLEAR:
       {
         coll_t cid = i.get_cid(op->cid);
@@ -468,18 +573,10 @@ seastar::future<> CyanStore::do_transaction(CollectionRef ch,
     logger().error("{}", str.str());
     ceph_assert(r == 0);
   }
-  for (auto i : {
-      t.get_on_applied(),
-      t.get_on_commit(),
-      t.get_on_applied_sync()}) {
-    if (i) {
-      i->complete(0);
-    }
-  }
   return seastar::now();
 }
 
-int CyanStore::_remove(const coll_t& cid, const ghobject_t& oid)
+int CyanStore::Shard::_remove(const coll_t& cid, const ghobject_t& oid)
 {
   logger().debug("{} cid={} oid={}",
                 __func__, cid, oid);
@@ -496,7 +593,7 @@ int CyanStore::_remove(const coll_t& cid, const ghobject_t& oid)
   return 0;
 }
 
-int CyanStore::_touch(const coll_t& cid, const ghobject_t& oid)
+int CyanStore::Shard::_touch(const coll_t& cid, const ghobject_t& oid)
 {
   logger().debug("{} cid={} oid={}",
                 __func__, cid, oid);
@@ -508,9 +605,13 @@ int CyanStore::_touch(const coll_t& cid, const ghobject_t& oid)
   return 0;
 }
 
-int CyanStore::_write(const coll_t& cid, const ghobject_t& oid,
-                      uint64_t offset, size_t len, const ceph::bufferlist& bl,
-                      uint32_t fadvise_flags)
+int CyanStore::Shard::_write(
+  const coll_t& cid,
+  const ghobject_t& oid,
+  uint64_t offset,
+  size_t len,
+  const ceph::bufferlist& bl,
+  uint32_t fadvise_flags)
 {
   logger().debug("{} {} {} {} ~ {}",
                 __func__, cid, oid, offset, len);
@@ -530,8 +631,11 @@ int CyanStore::_write(const coll_t& cid, const ghobject_t& oid,
   return 0;
 }
 
-int CyanStore::_zero(const coll_t& cid, const ghobject_t& oid,
-                     uint64_t offset, size_t len)
+int CyanStore::Shard::_zero(
+  const coll_t& cid,
+  const ghobject_t& oid,
+  uint64_t offset,
+  size_t len)
 {
   logger().debug("{} {} {} {} ~ {}",
                 __func__, cid, oid, offset, len);
@@ -541,7 +645,7 @@ int CyanStore::_zero(const coll_t& cid, const ghobject_t& oid,
   return _write(cid, oid, offset, len, bl, 0);
 }
 
-int CyanStore::_omap_clear(
+int CyanStore::Shard::_omap_clear(
   const coll_t& cid,
   const ghobject_t& oid)
 {
@@ -560,7 +664,7 @@ int CyanStore::_omap_clear(
   return 0;
 }
 
-int CyanStore::_omap_set_values(
+int CyanStore::Shard::_omap_set_values(
   const coll_t& cid,
   const ghobject_t& oid,
   std::map<std::string, ceph::bufferlist> &&aset)
@@ -574,13 +678,13 @@ int CyanStore::_omap_set_values(
     return -ENOENT;
 
   ObjectRef o = c->get_or_create_object(oid);
-  for (auto &&i: aset) {
-    o->omap.insert(std::move(i));
+  for (auto&& [key, val]: aset) {
+    o->omap.insert_or_assign(std::move(key), std::move(val));
   }
   return 0;
 }
 
-int CyanStore::_omap_set_header(
+int CyanStore::Shard::_omap_set_header(
   const coll_t& cid,
   const ghobject_t& oid,
   const ceph::bufferlist &header)
@@ -598,7 +702,7 @@ int CyanStore::_omap_set_header(
   return 0;
 }
 
-int CyanStore::_omap_rmkeys(
+int CyanStore::Shard::_omap_rmkeys(
   const coll_t& cid,
   const ghobject_t& oid,
   const omap_keys_t& aset)
@@ -618,7 +722,7 @@ int CyanStore::_omap_rmkeys(
   return 0;
 }
 
-int CyanStore::_omap_rmkeyrange(
+int CyanStore::Shard::_omap_rmkeyrange(
   const coll_t& cid,
   const ghobject_t& oid,
   const std::string &first,
@@ -634,12 +738,15 @@ int CyanStore::_omap_rmkeyrange(
 
   ObjectRef o = c->get_or_create_object(oid);
   for (auto i = o->omap.lower_bound(first);
-       i != o->omap.end() && i->first <= last;
+       i != o->omap.end() && i->first < last;
        o->omap.erase(i++));
   return 0;
 }
 
-int CyanStore::_truncate(const coll_t& cid, const ghobject_t& oid, uint64_t size)
+int CyanStore::Shard::_truncate(
+  const coll_t& cid,
+  const ghobject_t& oid,
+  uint64_t size)
 {
   logger().debug("{} cid={} oid={} size={}",
                 __func__, cid, oid, size);
@@ -658,8 +765,36 @@ int CyanStore::_truncate(const coll_t& cid, const ghobject_t& oid, uint64_t size
   return r;
 }
 
-int CyanStore::_setattrs(const coll_t& cid, const ghobject_t& oid,
-                         std::map<std::string,bufferptr>& aset)
+int CyanStore::Shard::_clone(
+  const coll_t& cid,
+  const ghobject_t& oid,
+  const ghobject_t& noid)
+{
+  logger().debug("{} cid={} oid={} noid={}",
+                __func__, cid, oid, noid);
+  auto c = _get_collection(cid);
+  if (!c)
+    return -ENOENT;
+
+  ObjectRef oo = c->get_object(oid);
+  if (!oo)
+    return -ENOENT;
+  if (local_conf()->memstore_debug_omit_block_device_write)
+    return 0;
+  ObjectRef no = c->get_or_create_object(noid);
+  used_bytes += ((ssize_t)oo->get_size() - (ssize_t)no->get_size());
+  no->clone(oo.get(), 0, oo->get_size(), 0);
+
+  no->omap_header = oo->omap_header;
+  no->omap = oo->omap;
+  no->xattr = oo->xattr;
+  return 0;
+}
+
+int CyanStore::Shard::_setattrs(
+  const coll_t& cid,
+  const ghobject_t& oid,
+  std::map<std::string,bufferlist>&& aset)
 {
   logger().debug("{} cid={} oid={}",
                 __func__, cid, oid);
@@ -670,14 +805,16 @@ int CyanStore::_setattrs(const coll_t& cid, const ghobject_t& oid,
   ObjectRef o = c->get_object(oid);
   if (!o)
     return -ENOENT;
-  for (std::map<std::string, bufferptr>::const_iterator p = aset.begin();
-       p != aset.end(); ++p)
-    o->xattr[p->first] = p->second;
+  for (auto&& [key, val]: aset) {
+    o->xattr.insert_or_assign(std::move(key), std::move(val));
+  }
   return 0;
 }
 
-int CyanStore::_rm_attr(const coll_t& cid, const ghobject_t& oid,
-			std::string_view name)
+int CyanStore::Shard::_rm_attr(
+  const coll_t& cid,
+  const ghobject_t& oid,
+  std::string_view name)
 {
   logger().debug("{} cid={} oid={} name={}", __func__, cid, oid, name);
   auto c = _get_collection(cid);
@@ -696,7 +833,24 @@ int CyanStore::_rm_attr(const coll_t& cid, const ghobject_t& oid,
   return 0;
 }
 
-int CyanStore::_create_collection(const coll_t& cid, int bits)
+int CyanStore::Shard::_rm_attrs(
+  const coll_t& cid,
+  const ghobject_t& oid)
+{
+  logger().debug("{} cid={} oid={}", __func__, cid, oid);
+  auto c = _get_collection(cid);
+  if (!c) {
+    return -ENOENT;
+  }
+  ObjectRef o = c->get_object(oid);
+  if (!o) {
+    return -ENOENT;
+  }
+  o->xattr.clear();
+  return 0;
+}
+
+int CyanStore::Shard::_create_collection(const coll_t& cid, int bits)
 {
   auto result = coll_map.try_emplace(cid);
   if (!result.second)
@@ -709,7 +863,8 @@ int CyanStore::_create_collection(const coll_t& cid, int bits)
   return 0;
 }
 
-boost::intrusive_ptr<Collection> CyanStore::_get_collection(const coll_t& cid)
+boost::intrusive_ptr<Collection>
+CyanStore::Shard::_get_collection(const coll_t& cid)
 {
   auto cp = coll_map.find(cid);
   if (cp == coll_map.end())
@@ -717,9 +872,11 @@ boost::intrusive_ptr<Collection> CyanStore::_get_collection(const coll_t& cid)
   return cp->second;
 }
 
-seastar::future<> CyanStore::write_meta(const std::string& key,
-					const std::string& value)
+seastar::future<> CyanStore::write_meta(
+  const std::string& key,
+  const std::string& value)
 {
+  ceph_assert(seastar::this_shard_id() == primary_core);
   std::string v = value;
   v += "\n";
   if (int r = safe_write_file(path.c_str(), key.c_str(),
@@ -733,6 +890,7 @@ seastar::future<> CyanStore::write_meta(const std::string& key,
 seastar::future<std::tuple<int, std::string>>
 CyanStore::read_meta(const std::string& key)
 {
+  ceph_assert(seastar::this_shard_id() == primary_core);
   std::string fsid(4096, '\0');
   int r = safe_read_file(path.c_str(), key.c_str(), fsid.data(), fsid.size());
   if (r > 0) {
@@ -749,34 +907,22 @@ CyanStore::read_meta(const std::string& key)
 
 uuid_d CyanStore::get_fsid() const
 {
+  ceph_assert(seastar::this_shard_id() == primary_core);
   return osd_fsid;
 }
 
-unsigned CyanStore::get_max_attr_name_length() const
+unsigned CyanStore::Shard::get_max_attr_name_length() const
 {
   // arbitrary limitation exactly like in the case of MemStore.
   return 256;
 }
 
-seastar::future<FuturizedStore::OmapIteratorRef> CyanStore::get_omap_iterator(
-    CollectionRef ch,
-    const ghobject_t& oid)
-{
-  auto c = static_cast<Collection*>(ch.get());
-  auto o = c->get_object(oid);
-  if (!o) {
-    throw std::runtime_error(fmt::format("object does not exist: {}", oid));
-  }
-  return seastar::make_ready_future<FuturizedStore::OmapIteratorRef>(
-	    new CyanStore::CyanOmapIterator(o));
-}
-
-seastar::future<std::map<uint64_t, uint64_t>>
-CyanStore::fiemap(
-    CollectionRef ch,
-    const ghobject_t& oid,
-    uint64_t off,
-    uint64_t len)
+CyanStore::Shard::read_errorator::future<std::map<uint64_t, uint64_t>>
+CyanStore::Shard::fiemap(
+  CollectionRef ch,
+  const ghobject_t& oid,
+  uint64_t off,
+  uint64_t len)
 {
   auto c = static_cast<Collection*>(ch.get());
 
@@ -789,7 +935,7 @@ CyanStore::fiemap(
 }
 
 seastar::future<struct stat>
-CyanStore::stat(
+CyanStore::Shard::stat(
   CollectionRef ch,
   const ghobject_t& oid)
 {
@@ -803,33 +949,4 @@ CyanStore::stat(
   return seastar::make_ready_future<struct stat>(std::move(st));
 }
 
-seastar::future<> CyanStore::CyanOmapIterator::seek_to_first()
-{
-  iter = obj->omap.begin();
-  return seastar::make_ready_future<>();
-}
-
-seastar::future<> CyanStore::CyanOmapIterator::upper_bound(const std::string& after)
-{
-  iter = obj->omap.upper_bound(after);
-  return seastar::make_ready_future<>();
-}
-
-seastar::future<> CyanStore::CyanOmapIterator::lower_bound(const std::string &to)
-{
-  iter = obj->omap.lower_bound(to);
-  return seastar::make_ready_future<>();
-}
-
-bool CyanStore::CyanOmapIterator::valid() const
-{
-  return iter != obj->omap.end();
-}
-
-seastar::future<> CyanStore::CyanOmapIterator::next()
-{
-  ++iter;
-  return seastar::make_ready_future<>();
-}
-
 }
