@@ -35,6 +35,7 @@
 #include "common/dout.h"
 #include "common/errno.h"
 #include "common/version.h"
+#include "common/win32/wstring.h"
 
 #include "global/global_init.h"
 
@@ -93,7 +94,7 @@ static NTSTATUS do_open_file(
   if (fd < 0) {
     dout(2) << __func__ << " " << path
             << ": ceph_open failed. Error: " << fd << dendl;
-    return cephfs_errno_to_ntsatus(fd);
+    return cephfs_errno_to_ntstatus_map(fd);
   }
 
   fdc->fd = fd;
@@ -111,11 +112,11 @@ static NTSTATUS WinCephCreateDirectory(
     return 0;
   }
 
-  int ret = ceph_mkdir(cmount, path.c_str(), 0755);
+  int ret = ceph_mkdir(cmount, path.c_str(), g_cfg->dir_mode);
   if (ret < 0) {
     dout(2) << __func__ << " " << path
             << ": ceph_mkdir failed. Error: " << ret << dendl;
-    return cephfs_errno_to_ntsatus(ret);
+    return cephfs_errno_to_ntstatus_map(ret);
   }
   return 0;
 }
@@ -166,13 +167,14 @@ static NTSTATUS WinCephCreateFile(
         return STATUS_OBJECT_NAME_COLLISION;
       case TRUNCATE_EXISTING:
         // open O_TRUNC & return 0
-        return do_open_file(path, O_CREAT | O_TRUNC | O_RDWR, 0755, fdc);
+        return do_open_file(path, O_CREAT | O_TRUNC | O_RDWR,
+                            g_cfg->file_mode, fdc);
       case OPEN_ALWAYS:
         // open & return STATUS_OBJECT_NAME_COLLISION
         if (!WRITE_ACCESS_REQUESTED(AccessMode))
           fdc->read_only = 1;
         if ((st = do_open_file(path, fdc->read_only ? O_RDONLY : O_RDWR,
-                               0755, fdc)))
+                               g_cfg->file_mode, fdc)))
           return st;
         return STATUS_OBJECT_NAME_COLLISION;
       case OPEN_EXISTING:
@@ -180,12 +182,13 @@ static NTSTATUS WinCephCreateFile(
         if (!WRITE_ACCESS_REQUESTED(AccessMode))
           fdc->read_only = 1;
         if ((st = do_open_file(path, fdc->read_only ? O_RDONLY : O_RDWR,
-                               0755, fdc)))
+                               g_cfg->file_mode, fdc)))
           return st;
         return 0;
       case CREATE_ALWAYS:
         // open O_TRUNC & return STATUS_OBJECT_NAME_COLLISION
-        if ((st = do_open_file(path, O_CREAT | O_TRUNC | O_RDWR, 0755, fdc)))
+        if ((st = do_open_file(path, O_CREAT | O_TRUNC | O_RDWR,
+                               g_cfg->file_mode, fdc)))
           return st;
         return STATUS_OBJECT_NAME_COLLISION;
       }
@@ -204,7 +207,7 @@ static NTSTATUS WinCephCreateFile(
         return 0;
       case OPEN_ALWAYS:
       case OPEN_EXISTING:
-        return do_open_file(path, O_RDONLY, 0755, fdc);
+        return do_open_file(path, O_RDONLY, g_cfg->file_mode, fdc);
       case CREATE_ALWAYS:
         return STATUS_OBJECT_NAME_COLLISION;
       }
@@ -220,18 +223,21 @@ static NTSTATUS WinCephCreateFile(
       if ((st = WinCephCreateDirectory(FileName, DokanFileInfo)))
         return st;
       // Dokan expects a file handle even when creating new directories.
-      return do_open_file(path, O_RDONLY, 0755, fdc);
+      return do_open_file(path, O_RDONLY, g_cfg->file_mode, fdc);
     }
     dout(20) << __func__ << " " << path << ". New file." << dendl;
     switch (CreationDisposition) {
       case CREATE_NEW:
         // create & return 0
-        return do_open_file(path, O_CREAT | O_RDWR | O_EXCL, 0755, fdc);
+        return do_open_file(path, O_CREAT | O_RDWR | O_EXCL,
+                            g_cfg->file_mode, fdc);
       case CREATE_ALWAYS:
         // create & return 0
-        return do_open_file(path, O_CREAT | O_TRUNC | O_RDWR, 0755, fdc);
+        return do_open_file(path, O_CREAT | O_TRUNC | O_RDWR,
+                            g_cfg->file_mode, fdc);
       case OPEN_ALWAYS:
-        return do_open_file(path, O_CREAT | O_RDWR, 0755, fdc);
+        return do_open_file(path, O_CREAT | O_RDWR,
+                            g_cfg->file_mode, fdc);
       case OPEN_EXISTING:
       case TRUNCATE_EXISTING:
         dout(2) << __func__ << " " << path << ": Not found." << dendl;
@@ -335,7 +341,7 @@ static NTSTATUS WinCephReadFile(
     if (fd_new < 0) {
       dout(2) << __func__ << " " << path
               << ": ceph_open failed. Error: " << fd_new << dendl;
-      return cephfs_errno_to_ntsatus(fd_new);
+      return cephfs_errno_to_ntstatus_map(fd_new);
     }
 
     int ret = ceph_read(cmount, fd_new, (char*) Buffer, BufferLength, Offset);
@@ -345,7 +351,7 @@ static NTSTATUS WinCephReadFile(
               << ". Offset: " << Offset
               << "Buffer length: " << BufferLength << dendl;
       ceph_close(cmount, fd_new);
-      return cephfs_errno_to_ntsatus(ret);
+      return cephfs_errno_to_ntstatus_map(ret);
     }
     *ReadLength = ret;
     ceph_close(cmount, fd_new);
@@ -357,7 +363,7 @@ static NTSTATUS WinCephReadFile(
               << ": ceph_read failed. Error: " << ret
               << ". Offset: " << Offset
               << "Buffer length: " << BufferLength << dendl;
-      return cephfs_errno_to_ntsatus(ret);
+      return cephfs_errno_to_ntstatus_map(ret);
     }
     *ReadLength = ret;
     return 0;
@@ -386,7 +392,7 @@ static NTSTATUS WinCephWriteFile(
       if (ret) {
         dout(2) << __func__ << " " << path
                 << ": ceph_statx failed. Error: " << ret << dendl;
-        return cephfs_errno_to_ntsatus(ret);
+        return cephfs_errno_to_ntstatus_map(ret);
       }
 
       Offset = stbuf.stx_size;
@@ -422,7 +428,7 @@ static NTSTATUS WinCephWriteFile(
     if (fd_new < 0) {
       dout(2) << __func__ << " " << path
               << ": ceph_open failed. Error: " << fd_new << dendl;
-      return cephfs_errno_to_ntsatus(fd_new);
+      return cephfs_errno_to_ntstatus_map(fd_new);
     }
 
     int ret = ceph_write(cmount, fd_new, (char*) Buffer,
@@ -433,7 +439,7 @@ static NTSTATUS WinCephWriteFile(
               << ". Offset: " << Offset
               << "Buffer length: " << NumberOfBytesToWrite << dendl;
       ceph_close(cmount, fd_new);
-      return cephfs_errno_to_ntsatus(ret);
+      return cephfs_errno_to_ntstatus_map(ret);
     }
     *NumberOfBytesWritten = ret;
     ceph_close(cmount, fd_new);
@@ -446,7 +452,7 @@ static NTSTATUS WinCephWriteFile(
               << ": ceph_write failed. Error: " << ret
               << ". Offset: " << Offset
               << "Buffer length: " << NumberOfBytesToWrite << dendl;
-      return cephfs_errno_to_ntsatus(ret);
+      return cephfs_errno_to_ntstatus_map(ret);
     }
     *NumberOfBytesWritten = ret;
     return 0;
@@ -467,7 +473,7 @@ static NTSTATUS WinCephFlushFileBuffers(
   if (ret) {
     dout(2) << __func__ << " " << get_path(FileName)
             << ": ceph_sync failed. Error: " << ret << dendl;
-    return cephfs_errno_to_ntsatus(ret);
+    return cephfs_errno_to_ntstatus_map(ret);
   }
   return 0;
 }
@@ -490,14 +496,14 @@ static NTSTATUS WinCephGetFileInformation(
     if (ret) {
       dout(2) << __func__ << " " << path
               << ": ceph_statx failed. Error: " << ret << dendl;
-      return cephfs_errno_to_ntsatus(ret);
+      return cephfs_errno_to_ntstatus_map(ret);
     }
   } else {
     int ret = ceph_fstatx(cmount, fdc->fd, &stbuf, requested_attrs, 0);
     if (ret) {
       dout(2) << __func__ << " " << path
               << ": ceph_fstatx failed. Error: " << ret << dendl;
-      return cephfs_errno_to_ntsatus(ret);
+      return cephfs_errno_to_ntstatus_map(ret);
     }
   }
 
@@ -534,7 +540,7 @@ static NTSTATUS WinCephFindFiles(
   if (ret != 0) {
     dout(2) << __func__ << " " << path
             << ": ceph_mkdir failed. Error: " << ret << dendl;
-    return cephfs_errno_to_ntsatus(ret);
+    return cephfs_errno_to_ntstatus_map(ret);
   }
 
   WIN32_FIND_DATAW findData;
@@ -554,7 +560,7 @@ static NTSTATUS WinCephFindFiles(
     if (ret < 0) {
       dout(2) << __func__ << " " << path
               << ": ceph_readdirplus_r failed. Error: " << ret << dendl;
-      return cephfs_errno_to_ntsatus(ret);
+      return cephfs_errno_to_ntstatus_map(ret);
     }
 
     to_wstring(result.d_name).copy(findData.cFileName, MAX_PATH);
@@ -617,7 +623,7 @@ static NTSTATUS WinCephDeleteDirectory(
   if (ret != 0) {
     dout(2) << __func__ << " " << path
             << ": ceph_opendir failed. Error: " << ret << dendl;
-    return cephfs_errno_to_ntsatus(ret);
+    return cephfs_errno_to_ntstatus_map(ret);
   }
 
   WIN32_FIND_DATAW findData;
@@ -654,7 +660,7 @@ static NTSTATUS WinCephMoveFile(
             << ": ceph_rename failed. Error: " << ret << dendl;
   }
 
-  return cephfs_errno_to_ntsatus(ret);
+  return cephfs_errno_to_ntstatus_map(ret);
 }
 
 static NTSTATUS WinCephSetEndOfFile(
@@ -673,7 +679,7 @@ static NTSTATUS WinCephSetEndOfFile(
     dout(2) << __func__ << " " << get_path(FileName)
             << ": ceph_ftruncate failed. Error: " << ret
             << " Offset: " << ByteOffset << dendl;
-    return cephfs_errno_to_ntsatus(ret);
+    return cephfs_errno_to_ntstatus_map(ret);
   }
 
   return 0;
@@ -696,7 +702,7 @@ static NTSTATUS WinCephSetAllocationSize(
   if (ret) {
     dout(2) << __func__ << " " << get_path(FileName)
             << ": ceph_fstatx failed. Error: " << ret << dendl;
-    return cephfs_errno_to_ntsatus(ret);
+    return cephfs_errno_to_ntstatus_map(ret);
   }
 
   if ((unsigned long long) AllocSize < stbuf.stx_size) {
@@ -704,7 +710,7 @@ static NTSTATUS WinCephSetAllocationSize(
     if (ret) {
       dout(2) << __func__ << " " << get_path(FileName)
               << ": ceph_ftruncate failed. Error: " << ret << dendl;
-      return cephfs_errno_to_ntsatus(ret);
+      return cephfs_errno_to_ntstatus_map(ret);
     }
     return 0;
   }
@@ -756,7 +762,7 @@ static NTSTATUS WinCephSetFileTime(
   if (ret) {
     dout(2) << __func__ << " " << path
             << ": ceph_setattrx failed. Error: " << ret << dendl;
-    return cephfs_errno_to_ntsatus(ret);
+    return cephfs_errno_to_ntstatus_map(ret);
   }
   return 0;
 }
@@ -789,7 +795,8 @@ static NTSTATUS WinCephGetVolumeInformation(
   g_cfg->win_vol_name.copy(VolumeNameBuffer, VolumeNameSize);
   *VolumeSerialNumber = g_cfg->win_vol_serial;
 
-  *MaximumComponentLength = 256;
+  *MaximumComponentLength = g_cfg->max_path_len;
+
   *FileSystemFlags = FILE_CASE_SENSITIVE_SEARCH |
             FILE_CASE_PRESERVED_NAMES |
             FILE_SUPPORTS_REMOTE_STORAGE |
@@ -810,7 +817,7 @@ static NTSTATUS WinCephGetDiskFreeSpace(
   int ret = ceph_statfs(cmount, "/", &vfsbuf);
   if (ret) {
     derr << "ceph_statfs failed. Error: " << ret << dendl;
-    return cephfs_errno_to_ntsatus(ret);;
+    return cephfs_errno_to_ntstatus_map(ret);;
   }
 
   *FreeBytesAvailable   = vfsbuf.f_bsize * vfsbuf.f_bfree;
@@ -874,7 +881,7 @@ NTSTATUS get_volume_serial(PDWORD serial) {
                           fsid_str, sizeof(fsid_str));
   if (ret < 0) {
     dout(2) << "Coudln't retrieve the cluster fsid. Error: " << ret << dendl;
-    return cephfs_errno_to_ntsatus(ret);
+    return cephfs_errno_to_ntstatus_map(ret);
   }
 
   uuid_d fsid;
@@ -931,7 +938,7 @@ int do_map() {
   r = ceph_mount(cmount, g_cfg->root_path.c_str());
   if (r) {
     derr << "ceph_mount failed. Error: " << r << dendl;
-    return cephfs_errno_to_ntsatus(r);
+    return cephfs_errno_to_ntstatus_map(r);
   }
 
   if (g_cfg->win_vol_name.empty()) {
@@ -949,12 +956,23 @@ int do_map() {
     }
   }
 
+  if (g_cfg->max_path_len > 260) {
+    dout(0) << "maximum path length set to " << g_cfg->max_path_len 
+            << ". Some Windows utilities may not be able to handle "
+            << "paths that exceed MAX_PATH (260) characters. "
+            << "CreateDirectoryW, used by Powershell, has also been "
+            << "observed to fail when paths exceed 16384 characters."
+            << dendl;
+  }
+
   atexit(unmount_atexit);
   dout(0) << "Mounted cephfs directory: " << g_cfg->root_path.c_str()
           <<". Mountpoint: " << to_string(g_cfg->mountpoint) << dendl;
 
+  DokanInit();
+
   DWORD status = DokanMain(dokan_options, dokan_operations);
-  switch (status) {
+  switch (static_cast<int>(status)) {
   case DOKAN_SUCCESS:
     dout(2) << "Dokan has returned successfully" << dendl;
     break;
@@ -981,6 +999,8 @@ int do_map() {
     break;
   }
 
+  DokanShutdown();
+
   free(dokan_options);
   free(dokan_operations);
   return 0;
@@ -989,8 +1009,7 @@ int do_map() {
 boost::intrusive_ptr<CephContext> do_global_init(
   int argc, const char **argv, Command cmd)
 {
-  std::vector<const char*> args;
-  argv_to_vec(argc, argv, args);
+  auto args = argv_to_vec(argc, argv);
 
   code_environment_t code_env;
   int flags;
@@ -1032,8 +1051,7 @@ int main(int argc, const char** argv)
   g_cfg = new Config;
 
   Command cmd = Command::None;
-  std::vector<const char*> args;
-  argv_to_vec(argc, argv, args);
+  auto args = argv_to_vec(argc, argv);
   std::ostringstream err_msg;
   int r = parse_args(args, &err_msg, &cmd, g_cfg);
   if (r) {
