@@ -28,6 +28,11 @@
 #undef dout_prefix
 #define dout_prefix *_dout << "librados: "
 
+using std::string;
+using std::map;
+using std::unique_lock;
+using std::vector;
+
 namespace bs = boost::system;
 namespace ca = ceph::async;
 namespace cb = ceph::buffer;
@@ -1230,7 +1235,7 @@ int librados::IoCtxImpl::remove(const object_t& oid)
   ::ObjectOperation op;
   prepare_assert_ops(&op);
   op.remove();
-  return operate(oid, &op, nullptr, librados::OPERATION_FULL_FORCE);
+  return operate(oid, &op, nullptr, CEPH_OSD_FLAG_FULL_FORCE);
 }
 
 int librados::IoCtxImpl::remove(const object_t& oid, int flags)
