@@ -572,9 +572,8 @@ int Mirror<I>::image_disable(I *ictx, bool force) {
     }
   };
 
-  std::unique_lock image_locker{ictx->image_lock};
-  map<librados::snap_t, SnapInfo> snap_info = ictx->snap_info;
-  for (auto &info : snap_info) {
+  std::shared_lock image_locker{ictx->image_lock};
+  for (const auto& info : ictx->snap_info) {
     cls::rbd::ParentImageSpec parent_spec{ictx->md_ctx.get_id(),
                                           ictx->md_ctx.get_namespace(),
                                           ictx->id, info.first};
@@ -1150,7 +1149,7 @@ int Mirror<I>::mode_set(librados::IoCtx& io_ctx,
   }
 
   if (next_mirror_mode == cls::rbd::MIRROR_MODE_POOL) {
-    map<string, string> images;
+    std::map<std::string, std::string> images;
     r = Image<I>::list_images_v2(io_ctx, &images);
     if (r < 0) {
       lderr(cct) << "failed listing images: " << cpp_strerror(r) << dendl;
@@ -1586,7 +1585,7 @@ int Mirror<I>::peer_site_remove(librados::IoCtx& io_ctx,
     return r;
   }
 
-  vector<string> names;
+  std::vector<std::string> names;
   r = Namespace<I>::list(io_ctx, &names);
   if (r < 0) {
     return r;
@@ -1644,7 +1643,7 @@ int Mirror<I>::peer_site_remove(librados::IoCtx& io_ctx,
       {
         std::shared_lock image_locker{img_ctx->image_lock};
         for (auto &it : img_ctx->snap_info) {
-          auto info = boost::get<cls::rbd::MirrorSnapshotNamespace>(
+          auto info = std::get_if<cls::rbd::MirrorSnapshotNamespace>(
             &it.second.snap_namespace);
           if (info && info->mirror_peer_uuids.count(uuid)) {
             snap_ids.push_back(it.first);
@@ -1863,9 +1862,9 @@ int Mirror<I>::image_global_status_list(
   CephContext *cct = reinterpret_cast<CephContext *>(io_ctx.cct());
   int r;
 
-  map<string, string> id_to_name;
+  std::map<std::string, std::string> id_to_name;
   {
-    map<string, string> name_to_id;
+    std::map<std::string, std::string> name_to_id;
     r = Image<I>::list_images_v2(io_ctx, &name_to_id);
     if (r < 0) {
       return r;
@@ -1875,8 +1874,8 @@ int Mirror<I>::image_global_status_list(
     }
   }
 
-  map<std::string, cls::rbd::MirrorImage> images_;
-  map<std::string, cls::rbd::MirrorImageStatus> statuses_;
+  std::map<std::string, cls::rbd::MirrorImage> images_;
+  std::map<std::string, cls::rbd::MirrorImageStatus> statuses_;
 
   r = librbd::cls_client::mirror_image_status_list(&io_ctx, start_id, max,
       					           &images_, &statuses_);
@@ -1945,8 +1944,11 @@ int Mirror<I>::image_status_summary(librados::IoCtx& io_ctx,
                                     MirrorImageStatusStates *states) {
   CephContext *cct = reinterpret_cast<CephContext *>(io_ctx.cct());
 
+  librados::IoCtx default_ns_io_ctx;
+  default_ns_io_ctx.dup(io_ctx);
+  default_ns_io_ctx.set_namespace("");
   std::vector<cls::rbd::MirrorPeer> mirror_peers;
-  int r = cls_client::mirror_peer_list(&io_ctx, &mirror_peers);
+  int r = cls_client::mirror_peer_list(&default_ns_io_ctx, &mirror_peers);
   if (r < 0 && r != -ENOENT) {
     lderr(cct) << "failed to list mirror peers: " << cpp_strerror(r) << dendl;
     return r;
@@ -2003,8 +2005,8 @@ int Mirror<I>::image_info_list(
   entries->clear();
 
   while (entries->size() < max) {
-    map<std::string, cls::rbd::MirrorImage> images;
-    map<std::string, cls::rbd::MirrorImageStatus> statuses;
+    std::map<std::string, cls::rbd::MirrorImage> images;
+    std::map<std::string, cls::rbd::MirrorImageStatus> statuses;
 
     int r = librbd::cls_client::mirror_image_status_list(&io_ctx, last_read,
                                                          max, &images,
