@@ -15,7 +15,7 @@
 #include <ostream>
 
 #include "FSMap.h"
-
+#include "common/debug.h"
 #include "common/StackStringStream.h"
 
 #ifdef WITH_SEASTAR
@@ -26,10 +26,16 @@
 #include "global/global_context.h"
 #include "mon/health_check.h"
 
+#define dout_context g_ceph_context
+#define dout_subsys ceph_subsys_mds
+#undef dout_prefix
+#define dout_prefix *_dout << "FSMap "
+
 using std::list;
 using std::pair;
 using std::ostream;
 using std::string;
+using std::string_view;
 
 using ceph::bufferlist;
 using ceph::Formatter;
@@ -791,7 +797,8 @@ const MDSMap::mds_info_t* FSMap::get_available_standby(const Filesystem& fs) con
       break;
     } else if (info.join_fscid == FS_CLUSTER_ID_NONE) {
       who = &info; /* vanilla standby */
-    } else if (who == nullptr) {
+    } else if (who == nullptr &&
+	       !fs.mds_map.test_flag(CEPH_MDSMAP_REFUSE_STANDBY_FOR_ANOTHER_FS)) {
       who = &info; /* standby for another fs, last resort */
     }
   }
@@ -1089,10 +1096,13 @@ std::vector<mds_gid_t> FSMap::stop(mds_gid_t who)
     if (other_info.rank == info.rank
         && other_info.state == MDSMap::STATE_STANDBY_REPLAY) {
       standbys.push_back(other_gid);
-      erase(other_gid, 0);
     }
   }
 
+  for (const auto &other_gid : standbys) {
+    erase(other_gid, 0);
+  }
+
   fs->mds_map.mds_info.erase(who);
   mds_roles.erase(who);
 
@@ -1209,3 +1219,20 @@ void FSMap::erase_filesystem(fs_cluster_id_t fscid)
     }
   }
 }
+
+void FSMap::swap_fscids(fs_cluster_id_t fscid1, fs_cluster_id_t fscid2)
+{
+  auto fs1 = std::move(filesystems.at(fscid1));
+  filesystems[fscid1] = std::move(filesystems.at(fscid2));
+  filesystems[fscid2] = std::move(fs1);
+
+  auto set_fs1_fscid = [fscid1](auto&& fs) {
+    fs->set_fscid(fscid1);
+  };
+  modify_filesystem(fscid1, std::move(set_fs1_fscid));
+
+  auto set_fs2_fscid = [fscid2](auto&& fs) {
+    fs->set_fscid(fscid2);
+  };
+  modify_filesystem(fscid2, std::move(set_fs2_fscid));
+}
