@@ -56,7 +56,8 @@ enum {
   l_mds_reply_latency,
   l_mds_slow_reply,
   l_mds_forward,
-  l_mds_dir_fetch,
+  l_mds_dir_fetch_complete,
+  l_mds_dir_fetch_keys,
   l_mds_dir_commit,
   l_mds_dir_split,
   l_mds_dir_merge,
@@ -168,7 +169,6 @@ class MDSRank {
 
     MDSRank(
         mds_rank_t whoami_,
-	std::string fs_name_,
         ceph::fair_mutex &mds_lock_,
         LogChannelRef &clog_,
         CommonSafeTimer<ceph::fair_mutex> &timer_,
@@ -182,7 +182,6 @@ class MDSRank {
 	boost::asio::io_context& ioc);
 
     mds_rank_t get_nodeid() const { return whoami; }
-    std::string_view get_fs_name() const { return fs_name; }
     int64_t get_metadata_pool() const
     {
         return metadata_pool;
@@ -191,9 +190,9 @@ class MDSRank {
     mono_time get_starttime() const {
       return starttime;
     }
-    chrono::duration<double> get_uptime() const {
+    std::chrono::duration<double> get_uptime() const {
       mono_time now = mono_clock::now();
-      return chrono::duration<double>(now-starttime);
+      return std::chrono::duration<double>(now-starttime);
     }
 
     bool is_daemon_stopping() const;
@@ -255,6 +254,10 @@ class MDSRank {
       progress_thread.signal();
     }
 
+    uint64_t get_global_id() const {
+      return monc->get_global_id();
+    }
+
     // Daemon lifetime functions: these guys break the abstraction
     // and call up into the parent MDSDaemon instance.  It's kind
     // of unavoidable: if we want any depth into our calls 
@@ -373,7 +376,7 @@ class MDSRank {
 
     void hit_export_target(mds_rank_t rank, double amount=-1.0);
     bool is_export_target(mds_rank_t rank) {
-      const set<mds_rank_t>& map_targets = mdsmap->get_mds_info(get_nodeid()).export_targets;
+      const std::set<mds_rank_t>& map_targets = mdsmap->get_mds_info(get_nodeid()).export_targets;
       return map_targets.count(rank);
     }
 
@@ -382,6 +385,7 @@ class MDSRank {
     int config_client(int64_t session_id, bool remove,
 		      const std::string& option, const std::string& value,
 		      std::ostream& ss);
+    void schedule_inmemory_logger();
 
     double get_inject_journal_corrupt_dentry_first() const {
       return inject_journal_corrupt_dentry_first;
@@ -466,7 +470,6 @@ class MDSRank {
     friend class C_MDS_MonCommand;
 
     const mds_rank_t whoami;
-    std::string fs_name;
 
     ~MDSRank();
 
@@ -492,17 +495,13 @@ class MDSRank {
     void dump_clientreplay_status(Formatter *f) const;
     void command_scrub_start(Formatter *f,
                              std::string_view path, std::string_view tag,
-                             const vector<string>& scrubop_vec, Context *on_finish);
+                             const std::vector<std::string>& scrubop_vec, Context *on_finish);
     void command_tag_path(Formatter *f, std::string_view path,
                           std::string_view tag);
     // scrub control commands
-    void command_scrub_abort(Formatter *f, Context *on_finish);
-    void command_scrub_pause(Formatter *f, Context *on_finish);
     void command_scrub_resume(Formatter *f);
     void command_scrub_status(Formatter *f);
 
-    void command_flush_path(Formatter *f, std::string_view path);
-    void command_flush_journal(Formatter *f);
     void command_get_subtrees(Formatter *f);
     void command_export_dir(Formatter *f,
         std::string_view path, mds_rank_t dest);
@@ -522,7 +521,9 @@ class MDSRank {
         std::ostream &ss);
     void command_openfiles_ls(Formatter *f);
     void command_dump_tree(const cmdmap_t &cmdmap, std::ostream &ss, Formatter *f);
+    int command_lock_path(Formatter *f, const cmdmap_t &cmdmap, std::ostream &ss);
     void command_dump_inode(Formatter *f, const cmdmap_t &cmdmap, std::ostream &ss);
+    void command_dump_dir(Formatter *f, const cmdmap_t &cmdmap, std::ostream &ss);
     void command_cache_drop(uint64_t timeout, Formatter *f, Context *on_finish);
 
     // FIXME the state machine logic should be separable from the dispatch
@@ -574,6 +575,8 @@ class MDSRank {
     // with the provided epoch.
     void apply_blocklist(const std::set<entity_addr_t> &addrs, epoch_t epoch);
 
+    void reset_event_flags();
+
     // Incarnation as seen in MDSMap at the point where a rank is
     // assigned.
     int incarnation = 0;
@@ -589,7 +592,7 @@ class MDSRank {
     MetricsHandler metrics_handler;
     std::unique_ptr<MetricAggregator> metric_aggregator;
 
-    list<cref_t<Message>> waiting_for_nolaggy;
+    std::list<cref_t<Message>> waiting_for_nolaggy;
     MDSContext::que finished_queue;
     // Dispatch, retry, queues
     int dispatch_depth = 0;
@@ -598,7 +601,7 @@ class MDSRank {
     double heartbeat_grace;
     int _heartbeat_reset_grace;
 
-    map<mds_rank_t, version_t> peer_mdsmap_epoch;
+    std::map<mds_rank_t, version_t> peer_mdsmap_epoch;
 
     ceph_tid_t last_tid = 0;    // for mds-initiated requests (e.g. stray rename)
 
@@ -608,9 +611,9 @@ class MDSRank {
     MDSContext::que replay_queue;
     bool replaying_requests_done = false;
 
-    map<mds_rank_t, MDSContext::vec> waiting_for_active_peer;
-    map<mds_rank_t, MDSContext::vec> waiting_for_bootstrapping_peer;
-    map<epoch_t, MDSContext::vec> waiting_for_mdsmap;
+    std::map<mds_rank_t, MDSContext::vec> waiting_for_active_peer;
+    std::map<mds_rank_t, MDSContext::vec> waiting_for_bootstrapping_peer;
+    std::map<epoch_t, MDSContext::vec> waiting_for_mdsmap;
 
     epoch_t osd_epoch_barrier = 0;
 
@@ -620,7 +623,7 @@ class MDSRank {
 
     int mds_slow_req_count = 0;
 
-    map<mds_rank_t,DecayCounter> export_targets; /* targets this MDS is exporting to or wants/tries to */
+    std::map<mds_rank_t,DecayCounter> export_targets; /* targets this MDS is exporting to or wants/tries to */
 
     Messenger *messenger;
     MonClient *monc;
@@ -630,6 +633,7 @@ class MDSRank {
     Context *suicide_hook;
 
     bool standby_replaying = false;  // true if current replay pass is in standby-replay mode
+    uint64_t extraordinary_events_dump_interval = 0;
     double inject_journal_corrupt_dentry_first = 0.0;
 private:
     bool send_status = true;
@@ -641,10 +645,13 @@ private:
     // "task" string that gets displayed in ceph status
     inline static const std::string SCRUB_STATUS_KEY = "scrub status";
 
+    bool client_eviction_dump = false;
+
     void get_task_status(std::map<std::string, std::string> *status);
     void schedule_update_timer_task();
     void send_task_status();
 
+    void inmemory_logger();
     bool is_rank0() const {
       return whoami == (mds_rank_t)0;
     }
@@ -653,9 +660,6 @@ private:
     boost::asio::io_context& ioc;
 };
 
-/* This expects to be given a reference which it is responsible for.
- * The finish function calls functions which
- * will put the Message exactly once.*/
 class C_MDS_RetryMessage : public MDSInternalContext {
 public:
   C_MDS_RetryMessage(MDSRank *mds, const cref_t<Message> &m)
@@ -690,7 +694,6 @@ class MDSRankDispatcher : public MDSRank, public md_config_obs_t
 public:
   MDSRankDispatcher(
       mds_rank_t whoami_,
-      std::string fs_name,
       ceph::fair_mutex &mds_lock_,
       LogChannelRef &clog_,
       CommonSafeTimer<ceph::fair_mutex> &timer_,
