@@ -12,8 +12,37 @@
  * 
  */
 
+
+/*
+
+  -- Storage scheme --
+
+  Pre-quincy:
+
+    - LogSummary contains last N entries for every channel
+    - LogSummary (as "full") written on every commit
+    - LogSummary contains "keys" which LogEntryKey hash_set for the
+      same set of entries (for deduping)
+
+  Quincy+:
+
+    - LogSummary contains, for each channel,
+       - start seq
+       - end seq (last written seq + 1)
+    - LogSummary contains an LRUSet for tracking dups
+    - LogSummary written every N commits
+    - each LogEntry written in a separate key
+       - "%s/%08x" % (channel, seq) -> LogEntry
+    - per-commit record includes channel -> begin (trim bounds)
+    - 'external_log_to' meta records version to which we have logged externally
+
+*/
+
+
+
 #include <boost/algorithm/string/predicate.hpp>
 
+#include <iterator>
 #include <sstream>
 #include <syslog.h>
 
@@ -25,12 +54,14 @@
 #include "messages/MLog.h"
 #include "messages/MLogAck.h"
 #include "common/Graylog.h"
+#include "common/Journald.h"
 #include "common/errno.h"
 #include "common/strtol.h"
 #include "include/ceph_assert.h"
 #include "include/str_list.h"
 #include "include/str_map.h"
 #include "include/compat.h"
+#include "include/utime_fmt.h"
 
 #define dout_subsys ceph_subsys_mon
 
@@ -43,6 +74,7 @@ using std::hex;
 using std::list;
 using std::map;
 using std::make_pair;
+using std::multimap;
 using std::ostream;
 using std::ostringstream;
 using std::pair;
@@ -163,6 +195,34 @@ ceph::logging::Graylog::Ref LogMonitor::log_channel_info::get_graylog(
   return graylogs[channel];
 }
 
+ceph::logging::JournaldClusterLogger &LogMonitor::log_channel_info::get_journald()
+{
+  dout(25) << __func__ << dendl;
+
+  if (!journald) {
+    journald = std::make_unique<ceph::logging::JournaldClusterLogger>();
+  }
+  return *journald;
+}
+
+void LogMonitor::log_channel_info::clear()
+{
+  log_to_syslog.clear();
+  syslog_facility.clear();
+  log_file.clear();
+  expanded_log_file.clear();
+  log_level.clear();
+  log_to_graylog.clear();
+  log_to_graylog_host.clear();
+  log_to_graylog_port.clear();
+  log_to_journald.clear();
+  graylogs.clear();
+  journald.reset();
+}
+
+LogMonitor::log_channel_info::log_channel_info() = default;
+LogMonitor::log_channel_info::~log_channel_info() = default;
+
 
 #undef dout_prefix
 #define dout_prefix _prefix(_dout, mon, get_last_committed())
@@ -198,6 +258,7 @@ void LogMonitor::create_initial()
   e.addrs = mon.messenger->get_myaddrs();
   e.stamp = ceph_clock_now();
   e.prio = CLOG_INFO;
+  e.channel = CLOG_CHANNEL_CLUSTER;
   std::stringstream ss;
   ss << "mkfs " << mon.monmap->get_fsid();
   e.msg = ss.str();
@@ -211,12 +272,13 @@ void LogMonitor::update_from_paxos(bool *need_bootstrap)
   version_t version = get_last_committed();
   dout(10) << __func__ << " version " << version
            << " summary v " << summary.version << dendl;
+
+  log_external_backlog();
+
   if (version == summary.version)
     return;
   ceph_assert(version >= summary.version);
 
-  map<string,bufferlist> channel_blog;
-
   version_t latest_full = get_version_latest_full();
   dout(10) << __func__ << " latest full " << latest_full << dendl;
   if ((latest_full > 0) && (latest_full > summary.version)) {
@@ -237,128 +299,321 @@ void LogMonitor::update_from_paxos(bool *need_bootstrap)
     ceph_assert(bl.length());
 
     auto p = bl.cbegin();
-    __u8 v;
-    decode(v, p);
-    while (!p.end()) {
-      LogEntry le;
-      le.decode(p);
-      dout(7) << "update_from_paxos applying incremental log " << summary.version+1 <<  " " << le << dendl;
-
-      string channel = le.channel;
-      if (channel.empty()) // keep retrocompatibility
-        channel = CLOG_CHANNEL_CLUSTER;
-
-      if (g_conf().get_val<bool>("mon_cluster_log_to_stderr")) {
-	cerr << channel << " " << le << std::endl;
+    __u8 struct_v;
+    decode(struct_v, p);
+    if (struct_v == 1) {
+      // legacy pre-quincy commits
+      while (!p.end()) {
+	LogEntry le;
+	le.decode(p);
+	dout(7) << "update_from_paxos applying incremental log "
+		<< summary.version+1 <<  " " << le << dendl;
+	summary.add_legacy(le);
       }
-
-      if (channels.do_log_to_syslog(channel)) {
-        string level = channels.get_level(channel);
-        string facility = channels.get_facility(channel);
-        if (level.empty() || facility.empty()) {
-          derr << __func__ << " unable to log to syslog -- level or facility"
-               << " not defined (level: " << level << ", facility: "
-               << facility << ")" << dendl;
-          continue;
-        }
-        le.log_to_syslog(channels.get_level(channel),
-                         channels.get_facility(channel));
-      }
-
-      if (channels.do_log_to_graylog(channel)) {
-	ceph::logging::Graylog::Ref graylog = channels.get_graylog(channel);
-	if (graylog) {
-	  graylog->log_log_entry(&le);
-	}
-	dout(7) << "graylog: " << channel << " " << graylog
-		<< " host:" << channels.log_to_graylog_host << dendl;
-      }
-
-      if (g_conf()->mon_cluster_log_to_file) {
-	string log_file = channels.get_log_file(channel);
-	dout(20) << __func__ << " logging for channel '" << channel
-		 << "' to file '" << log_file << "'" << dendl;
-
-	if (!log_file.empty()) {
-	  string log_file_level = channels.get_log_file_level(channel);
-	  if (log_file_level.empty()) {
-	    dout(1) << __func__ << " warning: log file level not defined for"
-		    << " channel '" << channel << "' yet a log file is --"
-		    << " will assume lowest level possible" << dendl;
-	  }
-
-	  int min = string_to_syslog_level(log_file_level);
-	  int l = clog_type_to_syslog_level(le.prio);
-	  if (l <= min) {
-	    stringstream ss;
-	    ss << le << "\n";
-	    // init entry if DNE
-	    bufferlist &blog = channel_blog[channel];
-	    blog.append(ss.str());
-	  }
+    } else {
+      uint32_t num;
+      decode(num, p);
+      while (num--) {
+	LogEntry le;
+	le.decode(p);
+	dout(7) << "update_from_paxos applying incremental log "
+		<< summary.version+1 <<  " " << le << dendl;
+	summary.recent_keys.insert(le.key());
+	summary.channel_info[le.channel].second++;
+	// we may have logged past the (persisted) summary in a prior quorum
+	if (version > external_log_to) {
+	  log_external(le);
 	}
       }
-
-      summary.add(le);
+      map<string,version_t> prune_channels_to;
+      decode(prune_channels_to, p);
+      for (auto& [channel, prune_to] : prune_channels_to) {
+	dout(20) << __func__ << " channel " << channel
+		 << " pruned to " << prune_to << dendl;
+	summary.channel_info[channel].first = prune_to;
+      }
+      // zero out pre-quincy fields (encode_pending needs this to reliably detect
+      // upgrade)
+      summary.tail_by_channel.clear();
+      summary.keys.clear();
     }
 
     summary.version++;
     summary.prune(g_conf()->mon_log_max_summary);
   }
+  dout(10) << " summary.channel_info " << summary.channel_info << dendl;
+  external_log_to = version;
+  mon.store->write_meta("external_log_to", stringify(external_log_to));
 
-  dout(15) << __func__ << " logging for "
-           << channel_blog.size() << " channels" << dendl;
-  for(map<string,bufferlist>::iterator p = channel_blog.begin();
-      p != channel_blog.end(); ++p) {
-    if (!p->second.length()) {
-      dout(15) << __func__ << " channel '" << p->first
-               << "': nothing to log" << dendl;
-      continue;
-    }
+  check_subs();
+}
 
-    dout(15) << __func__ << " channel '" << p->first
-             << "' logging " << p->second.length() << " bytes" << dendl;
-    string log_file = channels.get_log_file(p->first);
+void LogMonitor::log_external(const LogEntry& le)
+{
+  string channel = le.channel;
+  if (channel.empty()) { // keep retrocompatibility
+    channel = CLOG_CHANNEL_CLUSTER;
+  }
 
-    int fd = ::open(log_file.c_str(), O_WRONLY|O_APPEND|O_CREAT|O_CLOEXEC, 0600);
-    if (fd < 0) {
-      int err = -errno;
-      dout(1) << "unable to write to '" << log_file << "' for channel '"
-              << p->first << "': " << cpp_strerror(err) << dendl;
+  string level = channels.get_log_level(channel);
+  if (int log_level = LogEntry::str_to_level(level);log_level > le.prio) {
+    // Do not log LogEntry to any external entity if le.prio is
+    // less than channel log level.
+    return;
+  }
+
+  if (g_conf().get_val<bool>("mon_cluster_log_to_stderr")) {
+    cerr << channel << " " << le << std::endl;
+  }
+
+  if (channels.do_log_to_syslog(channel)) {
+    string facility = channels.get_facility(channel);
+    if (level.empty() || facility.empty()) {
+      derr << __func__ << " unable to log to syslog -- level or facility"
+	   << " not defined (level: " << level << ", facility: "
+	   << facility << ")" << dendl;
     } else {
-      int err = p->second.write_fd(fd);
-      if (err < 0) {
-	dout(1) << "error writing to '" << log_file << "' for channel '"
-                << p->first << ": " << cpp_strerror(err) << dendl;
-      }
-      VOID_TEMP_FAILURE_RETRY(::close(fd));
+      le.log_to_syslog(level, facility);
     }
   }
 
-  check_subs();
+  if (channels.do_log_to_graylog(channel)) {
+    ceph::logging::Graylog::Ref graylog = channels.get_graylog(channel);
+    if (graylog) {
+      graylog->log_log_entry(&le);
+    }
+    dout(7) << "graylog: " << channel << " " << graylog
+	    << " host:" << channels.log_to_graylog_host << dendl;
+  }
+
+  if (channels.do_log_to_journald(channel)) {
+    auto &journald = channels.get_journald();
+    journald.log_log_entry(le);
+    dout(7) << "journald: " << channel << dendl;
+  }
+
+  bool do_stderr = g_conf().get_val<bool>("mon_cluster_log_to_stderr");
+  int fd = -1;
+  if (g_conf()->mon_cluster_log_to_file) {
+    if (this->log_rotated.exchange(false)) {
+      this->log_external_close_fds();
+    }
+
+    auto p = channel_fds.find(channel);
+    if (p == channel_fds.end()) {
+      string log_file = channels.get_log_file(channel);
+      dout(20) << __func__ << " logging for channel '" << channel
+	       << "' to file '" << log_file << "'" << dendl;
+      if (!log_file.empty()) {
+	fd = ::open(log_file.c_str(), O_WRONLY|O_APPEND|O_CREAT|O_CLOEXEC, 0600);
+	if (fd < 0) {
+	  int err = -errno;
+	  dout(1) << "unable to write to '" << log_file << "' for channel '"
+		  << channel << "': " << cpp_strerror(err) << dendl;
+	} else {
+	  channel_fds[channel] = fd;
+	}
+      }
+    } else {
+      fd = p->second;
+    }
+  }
+  if (do_stderr || fd >= 0) {
+    fmt::format_to(std::back_inserter(log_buffer), "{}\n", le);
+
+    if (fd >= 0) {
+      int err = safe_write(fd, log_buffer.data(), log_buffer.size());
+      if (err < 0) {
+	dout(1) << "error writing to '" << channels.get_log_file(channel)
+		<< "' for channel '" << channel
+		<< ": " << cpp_strerror(err) << dendl;
+	::close(fd);
+	channel_fds.erase(channel);
+      }
+    }
+
+    if (do_stderr) {
+      fmt::print(std::cerr, "{} {}", channel, std::string_view(log_buffer.data(), log_buffer.size()));
+    }
+
+    log_buffer.clear();
+  }
+}
+
+void LogMonitor::log_external_close_fds()
+{
+  for (auto& [channel, fd] : channel_fds) {
+    if (fd >= 0) {
+      dout(10) << __func__ << " closing " << channel << " (" << fd << ")" << dendl;
+      ::close(fd);
+    }
+  }
+  channel_fds.clear();
+}
+
+/// catch external logs up to summary.version
+void LogMonitor::log_external_backlog()
+{
+  if (!external_log_to) {
+    std::string cur_str;
+    int r = mon.store->read_meta("external_log_to", &cur_str);
+    if (r == 0) {
+      external_log_to = std::stoull(cur_str);
+      dout(10) << __func__ << " initialized external_log_to = " << external_log_to
+	       << " (recorded log_to position)" << dendl;
+    } else {
+      // pre-quincy, we assumed that anything through summary.version was
+      // logged externally.
+      assert(r == -ENOENT);
+      external_log_to = summary.version;
+      dout(10) << __func__ << " initialized external_log_to = " << external_log_to
+	       << " (summary v " << summary.version << ")" << dendl;
+    }
+  }
+  // we may have logged ahead of summary.version, but never ahead of paxos
+  if (external_log_to > get_last_committed()) {
+    derr << __func__ << " rewinding external_log_to from " << external_log_to
+	 << " -> " << get_last_committed() << " (sync_force? mon rebuild?)" << dendl;
+    external_log_to = get_last_committed();
+  }
+  if (external_log_to >= summary.version) {
+    return;
+  }
+  if (auto first = get_first_committed(); external_log_to < first) {
+    derr << __func__ << " local logs at " << external_log_to
+	 << ", skipping to " << first << dendl;
+    external_log_to = first;
+    // FIXME: write marker in each channel log file?
+  }
+  for (; external_log_to < summary.version; ++external_log_to) {
+    bufferlist bl;
+    int err = get_version(external_log_to+1, bl);
+    ceph_assert(err == 0);
+    ceph_assert(bl.length());
+    auto p = bl.cbegin();
+    __u8 v;
+    decode(v, p);
+    int32_t num = -2;
+    if (v >= 2) {
+      decode(num, p);
+    }
+    while ((num == -2 && !p.end()) || (num >= 0 && num--)) {
+      LogEntry le;
+      le.decode(p);
+      log_external(le);
+    }
+  }
+  mon.store->write_meta("external_log_to", stringify(external_log_to));
 }
 
 void LogMonitor::create_pending()
 {
   pending_log.clear();
-  pending_summary = summary;
+  pending_keys.clear();
   dout(10) << "create_pending v " << (get_last_committed() + 1) << dendl;
 }
 
+void LogMonitor::generate_logentry_key(
+  const std::string& channel,
+  version_t v,
+  std::string *out)
+{
+  out->append(channel);
+  out->append("/");
+  char vs[10];
+  snprintf(vs, sizeof(vs), "%08llx", (unsigned long long)v);
+  out->append(vs);
+}
+
 void LogMonitor::encode_pending(MonitorDBStore::TransactionRef t)
 {
   version_t version = get_last_committed() + 1;
   bufferlist bl;
   dout(10) << __func__ << " v" << version << dendl;
-  __u8 v = 1;
-  encode(v, bl);
-  for (auto p = pending_log.begin(); p != pending_log.end(); ++p)
-    p->second.encode(bl, mon.get_quorum_con_features());
+
+  if (mon.monmap->min_mon_release < ceph_release_t::quincy) {
+    // legacy encoding for pre-quincy quorum
+    __u8 struct_v = 1;
+    encode(struct_v, bl);
+    for (auto& p : pending_log) {
+      p.second.encode(bl, mon.get_quorum_con_features());
+    }
+    put_version(t, version, bl);
+    put_last_committed(t, version);
+    return;
+  }
+  
+  __u8 struct_v = 2;
+  encode(struct_v, bl);
+
+  // first commit after upgrading to quincy?
+  if (!summary.tail_by_channel.empty()) {
+    // include past log entries
+    for (auto& p : summary.tail_by_channel) {
+      for (auto& q : p.second) {
+	pending_log.emplace(make_pair(q.second.stamp, q.second));
+      }
+    }
+  }
+
+  // record new entries
+  auto pending_channel_info = summary.channel_info;
+  uint32_t num = pending_log.size();
+  encode(num, bl);
+  dout(20) << __func__ << " writing " << num << " entries" << dendl;
+  for (auto& p : pending_log) {
+    bufferlist ebl;
+    p.second.encode(ebl, mon.get_quorum_con_features());
+
+    auto& bounds = pending_channel_info[p.second.channel];
+    version_t v = bounds.second++;
+    std::string key;
+    generate_logentry_key(p.second.channel, v, &key);
+    t->put(get_service_name(), key, ebl);
+    
+    bl.claim_append(ebl);
+  }
+
+  // prune log entries?
+  map<string,version_t> prune_channels_to;
+  for (auto& [channel, info] : summary.channel_info) {
+    if (info.second - info.first > g_conf()->mon_log_max) {
+      const version_t from = info.first;
+      const version_t to = info.second - g_conf()->mon_log_max;
+      dout(10) << __func__ << " pruning channel " << channel
+	       << " " << from << " -> " << to << dendl;
+      prune_channels_to[channel] = to;
+      pending_channel_info[channel].first = to;
+      for (version_t v = from; v < to; ++v) {
+	std::string key;
+	generate_logentry_key(channel, v, &key);
+	t->erase(get_service_name(), key);
+      }
+    }
+  }
+  dout(20) << __func__ << " prune_channels_to " << prune_channels_to << dendl;
+  encode(prune_channels_to, bl);
 
   put_version(t, version, bl);
   put_last_committed(t, version);
 }
 
+bool LogMonitor::should_stash_full()
+{
+  if (mon.monmap->min_mon_release < ceph_release_t::quincy) {
+    // commit a LogSummary on every commit
+    return true;
+  }
+
+  // store periodic summary
+  auto period = std::min<uint64_t>(
+    g_conf()->mon_log_full_interval,
+    g_conf()->mon_max_log_epochs
+    );
+  return (get_last_committed() - get_version_latest_full() > period);
+}
+
+
 void LogMonitor::encode_full(MonitorDBStore::TransactionRef t)
 {
   dout(10) << __func__ << " log v " << summary.version << dendl;
@@ -448,7 +703,7 @@ bool LogMonitor::preprocess_log(MonOpRequestRef op)
   for (auto p = m->entries.begin();
        p != m->entries.end();
        ++p) {
-    if (!pending_summary.contains(p->key()))
+    if (!summary.contains(p->key()))
       num_new++;
   }
   if (!num_new) {
@@ -491,13 +746,13 @@ bool LogMonitor::prepare_log(MonOpRequestRef op)
        p != m->entries.end();
        ++p) {
     dout(10) << " logging " << *p << dendl;
-    if (!pending_summary.contains(p->key())) {
-      pending_summary.add(*p);
+    if (!summary.contains(p->key()) &&
+	!pending_keys.count(p->key())) {
+      pending_keys.insert(p->key());
       pending_log.insert(pair<utime_t,LogEntry>(p->stamp, *p));
     }
   }
-  pending_summary.prune(g_conf()->mon_log_max_summary);
-  wait_for_finished_proposal(op, new C_Log(this, op));
+  wait_for_commit(op, new C_Log(this, op));
   return true;
 }
 
@@ -543,8 +798,7 @@ bool LogMonitor::preprocess_command(MonOpRequestRef op)
   string prefix;
   cmd_getval(cmdmap, "prefix", prefix);
 
-  string format;
-  cmd_getval(cmdmap, "format", format, string("plain"));
+  string format = cmd_getval_or<string>(cmdmap, "format", "plain");
   boost::scoped_ptr<Formatter> f(Formatter::create(format));
 
   if (prefix == "log last") {
@@ -578,79 +832,160 @@ bool LogMonitor::preprocess_command(MonOpRequestRef op)
       return entry.prio >= level;
     };
 
-    // Decrement operation that sets to container end when hitting rbegin
     ostringstream ss;
-    if (channel == "*") {
-      list<LogEntry> full_tail;
-      summary.build_ordered_tail(&full_tail);
-      derr << "full " << full_tail << dendl;
-      auto rp = full_tail.rbegin();
-      for (; num > 0 && rp != full_tail.rend(); ++rp) {
-	if (match(*rp)) {
-	  num--;
-	}
-      }
-      if (rp == full_tail.rend()) {
-	--rp;
-      }
-
-      // Decrement a reverse iterator such that going past rbegin()
-      // sets it to rend().  This is for writing a for() loop that
-      // goes up to (and including) rbegin()
-      auto dec = [&rp, &full_tail] () {
-        if (rp == full_tail.rbegin()) {
-          rp = full_tail.rend();
-        } else {
-          --rp;
-        }
-      };
-
-      // Move forward to the end of the container (decrement the reverse
-      // iterator).
-      for (; rp != full_tail.rend(); dec()) {
-	if (!match(*rp)) {
-	  continue;
-	}
-	if (f) {
-	  f->dump_object("entry", *rp);
-	} else {
-	  ss << *rp << "\n";
-	}
-      }
-    } else {
-      auto p = summary.tail_by_channel.find(channel);
-      if (p != summary.tail_by_channel.end()) {
-	auto rp = p->second.rbegin();
-	for (; num > 0 && rp != p->second.rend(); ++rp) {
-	  if (match(rp->second)) {
+    if (!summary.tail_by_channel.empty()) {
+      // pre-quincy compat
+      // Decrement operation that sets to container end when hitting rbegin
+      if (channel == "*") {
+	list<LogEntry> full_tail;
+	summary.build_ordered_tail_legacy(&full_tail);
+	auto rp = full_tail.rbegin();
+	for (; num > 0 && rp != full_tail.rend(); ++rp) {
+	  if (match(*rp)) {
 	    num--;
 	  }
 	}
-	if (rp == p->second.rend()) {
+	if (rp == full_tail.rend()) {
 	  --rp;
 	}
 
-        // Decrement a reverse iterator such that going past rbegin()
-        // sets it to rend().  This is for writing a for() loop that
-        // goes up to (and including) rbegin()
-        auto dec = [&rp, &p] () {
-          if (rp == p->second.rbegin()) {
-            rp = p->second.rend();
-          } else {
-            --rp;
-          }
-        };
+	// Decrement a reverse iterator such that going past rbegin()
+	// sets it to rend().  This is for writing a for() loop that
+	// goes up to (and including) rbegin()
+	auto dec = [&rp, &full_tail] () {
+		     if (rp == full_tail.rbegin()) {
+		       rp = full_tail.rend();
+		     } else {
+		       --rp;
+		     }
+		   };
 
-        // Move forward to the end of the container (decrement the reverse
-        // iterator).
-	for (; rp != p->second.rend(); dec()) {
-	  if (!match(rp->second)) {
+	// Move forward to the end of the container (decrement the reverse
+	// iterator).
+	for (; rp != full_tail.rend(); dec()) {
+	  if (!match(*rp)) {
 	    continue;
 	  }
 	  if (f) {
-	    f->dump_object("entry", rp->second);
+	    f->dump_object("entry", *rp);
 	  } else {
-	    ss << rp->second << "\n";
+	    ss << *rp << "\n";
+	  }
+	}
+      } else {
+	auto p = summary.tail_by_channel.find(channel);
+	if (p != summary.tail_by_channel.end()) {
+	  auto rp = p->second.rbegin();
+	  for (; num > 0 && rp != p->second.rend(); ++rp) {
+	    if (match(rp->second)) {
+	      num--;
+	    }
+	  }
+	  if (rp == p->second.rend()) {
+	    --rp;
+	  }
+
+	  // Decrement a reverse iterator such that going past rbegin()
+	  // sets it to rend().  This is for writing a for() loop that
+	  // goes up to (and including) rbegin()
+	  auto dec = [&rp, &p] () {
+		       if (rp == p->second.rbegin()) {
+			 rp = p->second.rend();
+		       } else {
+			 --rp;
+		       }
+		     };
+	  
+	  // Move forward to the end of the container (decrement the reverse
+	  // iterator).
+	  for (; rp != p->second.rend(); dec()) {
+	    if (!match(rp->second)) {
+	      continue;
+	    }
+	    if (f) {
+	      f->dump_object("entry", rp->second);
+	    } else {
+	      ss << rp->second << "\n";
+	    }
+	  }
+	}
+      }
+    } else {
+      // quincy+
+      if (channel == "*") {
+	// tail all channels; we need to mix by timestamp
+	multimap<utime_t,LogEntry> entries;  // merge+sort all channels by timestamp
+	for (auto& p : summary.channel_info) {
+	  version_t from = p.second.first;
+	  version_t to = p.second.second;
+	  version_t start;
+	  if (to > (version_t)num) {
+	    start = std::max(to - num, from);
+	  } else {
+	    start = from;
+	  }
+	  dout(10) << __func__ << " channel " << p.first
+		   << " from " << from << " to " << to << dendl;
+	  for (version_t v = start; v < to; ++v) {
+	    bufferlist ebl;
+	    string key;
+	    generate_logentry_key(p.first, v, &key);
+	    int r = mon.store->get(get_service_name(), key, ebl);
+	    if (r < 0) {
+	      derr << __func__ << " missing key " << key << dendl;
+	      continue;
+	    }
+	    LogEntry le;
+	    auto p = ebl.cbegin();
+	    decode(le, p);
+	    entries.insert(make_pair(le.stamp, le));
+	  }
+	}
+	while ((int)entries.size() > num) {
+	  entries.erase(entries.begin());
+	}
+	for (auto& p : entries) {
+	  if (!match(p.second)) {
+	    continue;
+	  }
+	  if (f) {
+	    f->dump_object("entry", p.second);
+	  } else {
+	    ss << p.second << "\n";
+	  }
+	}
+      } else {
+	// tail one channel
+	auto p = summary.channel_info.find(channel);
+	if (p != summary.channel_info.end()) {
+	  version_t from = p->second.first;
+	  version_t to = p->second.second;
+	  version_t start;
+	  if (to > (version_t)num) {
+	    start = std::max(to - num, from);
+	  } else {
+	    start = from;
+	  }
+	  dout(10) << __func__ << " from " << from << " to " << to << dendl;
+	  for (version_t v = start; v < to; ++v) {
+	    bufferlist ebl;
+	    string key;
+	    generate_logentry_key(channel, v, &key);
+	    int r = mon.store->get(get_service_name(), key, ebl);
+	    if (r < 0) {
+	      derr << __func__ << " missing key " << key << dendl;
+	      continue;
+	    }
+	    LogEntry le;
+	    auto p = ebl.cbegin();
+	    decode(le, p);
+	    if (match(le)) {
+	      if (f) {
+	        f->dump_object("entry", le);
+	      } else {
+	        ss << le << "\n";
+	      }
+	    }
 	  }
 	}
       }
@@ -700,7 +1035,6 @@ bool LogMonitor::prepare_command(MonOpRequestRef op)
 
   if (prefix == "log") {
     vector<string> logtext;
-    string level_str;
     cmd_getval(cmdmap, "logtext", logtext);
     LogEntry le;
     le.rank = m->get_orig_source();
@@ -708,14 +1042,13 @@ bool LogMonitor::prepare_command(MonOpRequestRef op)
     le.name = session->entity_name;
     le.stamp = m->get_recv_stamp();
     le.seq = 0;
-    cmd_getval(cmdmap, "level", level_str, string("info"));
+    string level_str = cmd_getval_or<string>(cmdmap, "level", "info");
     le.prio = LogEntry::str_to_level(level_str);
     le.channel = CLOG_CHANNEL_DEFAULT;
     le.msg = str_join(logtext, " ");
-    pending_summary.add(le);
-    pending_summary.prune(g_conf()->mon_log_max_summary);
+    pending_keys.insert(le.key());
     pending_log.insert(pair<utime_t,LogEntry>(le.stamp, le));
-    wait_for_finished_proposal(op, new Monitor::C_Command(
+    wait_for_commit(op, new Monitor::C_Command(
           mon, op, 0, string(), get_last_committed() + 1));
     return true;
   }
@@ -725,6 +1058,11 @@ bool LogMonitor::prepare_command(MonOpRequestRef op)
   return false;
 }
 
+void LogMonitor::dump_info(Formatter *f)
+{
+  f->dump_unsigned("logm_first_committed", get_first_committed());
+  f->dump_unsigned("logm_last_committed", get_last_committed());
+}
 
 int LogMonitor::sub_name_to_id(const string& n)
 {
@@ -823,17 +1161,20 @@ void LogMonitor::_create_sub_incremental(MLog *mlog, int level, version_t sv)
     ceph_assert(bl.length());
     auto p = bl.cbegin();
     __u8 v;
-    decode(v,p);
-    while (!p.end()) {
+    decode(v, p);
+    int32_t num = -2;
+    if (v >= 2) {
+      decode(num, p);
+      dout(20) << __func__ << " sv " << sv << " has " << num << " entries" << dendl;
+    }
+    while ((num == -2 && !p.end()) || (num >= 0 && num--)) {
       LogEntry le;
       le.decode(p);
-
       if (le.prio < level) {
 	dout(20) << __func__ << " requested " << level 
-		 << " entry " << le.prio << dendl;
+		 << ", skipping " << le << dendl;
 	continue;
       }
-
       mlog->entries.push_back(le);
     }
     mlog->version = sv++;
@@ -858,16 +1199,6 @@ void LogMonitor::update_log_channels()
     return;
   }
 
-  r = get_conf_str_map_helper(
-    g_conf().get_val<string>("mon_cluster_log_to_syslog_level"),
-    oss, &channels.syslog_level,
-    CLOG_CONFIG_DEFAULT_KEY);
-  if (r < 0) {
-    derr << __func__ << " error parsing 'mon_cluster_log_to_syslog_level'"
-         << dendl;
-    return;
-  }
-
   r = get_conf_str_map_helper(
     g_conf().get_val<string>("mon_cluster_log_to_syslog_facility"),
     oss, &channels.syslog_facility,
@@ -888,11 +1219,11 @@ void LogMonitor::update_log_channels()
   }
 
   r = get_conf_str_map_helper(
-    g_conf().get_val<string>("mon_cluster_log_file_level"), oss,
-    &channels.log_file_level,
+    g_conf().get_val<string>("mon_cluster_log_level"), oss,
+    &channels.log_level,
     CLOG_CONFIG_DEFAULT_KEY);
   if (r < 0) {
-    derr << __func__ << " error parsing 'mon_cluster_log_file_level'"
+    derr << __func__ << " error parsing 'mon_cluster_log_level'"
          << dendl;
     return;
   }
@@ -927,7 +1258,18 @@ void LogMonitor::update_log_channels()
     return;
   }
 
+  r = get_conf_str_map_helper(
+    g_conf().get_val<string>("mon_cluster_log_to_journald"), oss,
+    &channels.log_to_journald,
+    CLOG_CONFIG_DEFAULT_KEY);
+  if (r < 0) {
+    derr << __func__ << " error parsing 'mon_cluster_log_to_journald'"
+         << dendl;
+    return;
+  }
+
   channels.expand_channel_meta();
+  log_external_close_fds();
 }
 
 
@@ -935,13 +1277,14 @@ void LogMonitor::handle_conf_change(const ConfigProxy& conf,
                                     const std::set<std::string> &changed)
 {
   if (changed.count("mon_cluster_log_to_syslog") ||
-      changed.count("mon_cluster_log_to_syslog_level") ||
       changed.count("mon_cluster_log_to_syslog_facility") ||
       changed.count("mon_cluster_log_file") ||
-      changed.count("mon_cluster_log_file_level") ||
+      changed.count("mon_cluster_log_level") ||
       changed.count("mon_cluster_log_to_graylog") ||
       changed.count("mon_cluster_log_to_graylog_host") ||
-      changed.count("mon_cluster_log_to_graylog_port")) {
+      changed.count("mon_cluster_log_to_graylog_port") ||
+      changed.count("mon_cluster_log_to_journald") ||
+      changed.count("mon_cluster_log_to_file")) {
     update_log_channels();
   }
 }
