@@ -487,6 +487,8 @@ namespace ceph {
       constexpr mon_feature_t FEATURE_PACIFIC(    (1ULL << 6));
       // elector pinging and CONNECTIVITY mode:
       constexpr mon_feature_t FEATURE_PINGING(    (1ULL << 7));
+      constexpr mon_feature_t FEATURE_QUINCY(    (1ULL << 8));
+      constexpr mon_feature_t FEATURE_REEF(    (1ULL << 9));
 
       constexpr mon_feature_t FEATURE_RESERVED(   (1ULL << 63));
       constexpr mon_feature_t FEATURE_NONE(       (0ULL));
@@ -506,6 +508,8 @@ namespace ceph {
 	  FEATURE_OCTOPUS |
 	  FEATURE_PACIFIC |
 	  FEATURE_PINGING |
+	  FEATURE_QUINCY |
+	  FEATURE_REEF |
 	  FEATURE_NONE
 	  );
       }
@@ -529,6 +533,8 @@ namespace ceph {
 	  FEATURE_OCTOPUS |
 	  FEATURE_PACIFIC |
 	  FEATURE_PINGING |
+	  FEATURE_QUINCY |
+	  FEATURE_REEF |
 	  FEATURE_NONE
 	  );
       }
@@ -547,6 +553,12 @@ namespace ceph {
 
 static inline ceph_release_t infer_ceph_release_from_mon_features(mon_feature_t f)
 {
+  if (f.contains_all(ceph::features::mon::FEATURE_REEF)) {
+    return ceph_release_t::reef;
+  }
+  if (f.contains_all(ceph::features::mon::FEATURE_QUINCY)) {
+    return ceph_release_t::quincy;
+  }
   if (f.contains_all(ceph::features::mon::FEATURE_PACIFIC)) {
     return ceph_release_t::pacific;
   }
@@ -587,6 +599,10 @@ static inline const char *ceph::features::mon::get_feature_name(uint64_t b) {
     return "octopus";
   } else if (f == FEATURE_PACIFIC) {
     return "pacific";
+  } else if (f == FEATURE_QUINCY) {
+    return "quincy";
+  } else if (f == FEATURE_REEF) {
+    return "reef";
   } else if (f == FEATURE_RESERVED) {
     return "reserved";
   }
@@ -611,6 +627,10 @@ inline mon_feature_t ceph::features::mon::get_feature_by_name(const std::string
     return FEATURE_OCTOPUS;
   } else if (n == "pacific") {
     return FEATURE_PACIFIC;
+  } else if (n == "quincy") {
+    return FEATURE_QUINCY;
+  } else if (n == "reef") {
+    return FEATURE_REEF;
   } else if (n == "reserved") {
     return FEATURE_RESERVED;
   }
