@@ -1,5 +1,5 @@
 import { Component, OnInit, Type, ViewChild } from '@angular/core';
-import { FormControl, Validators } from '@angular/forms';
+import { UntypedFormControl, Validators } from '@angular/forms';
 import { ActivatedRoute, Router } from '@angular/router';
 
 import { NgbNav, NgbTooltip } from '@ng-bootstrap/ng-bootstrap';
@@ -11,7 +11,7 @@ import { CrushRuleService } from '~/app/shared/api/crush-rule.service';
 import { ErasureCodeProfileService } from '~/app/shared/api/erasure-code-profile.service';
 import { PoolService } from '~/app/shared/api/pool.service';
 import { CrushNodeSelectionClass } from '~/app/shared/classes/crush.node.selection.class';
-import { CriticalConfirmationModalComponent } from '~/app/shared/components/critical-confirmation-modal/critical-confirmation-modal.component';
+import { DeleteConfirmationModalComponent } from '~/app/shared/components/delete-confirmation-modal/delete-confirmation-modal.component';
 import { SelectOption } from '~/app/shared/components/select/select-option.model';
 import { ActionLabelsI18n, URLVerbs } from '~/app/shared/constants/app.constants';
 import { Icons } from '~/app/shared/enum/icons.enum';
@@ -82,6 +82,7 @@ export class PoolFormComponent extends CdForm implements OnInit {
   pgAutoscaleModes: string[];
   crushUsage: string[] = undefined; // Will only be set if a rule is used by some pool
   ecpUsage: string[] = undefined; // Will only be set if a rule is used by some pool
+  crushRuleMaxSize = 10;
 
   private modalSubscription: Subscription;
 
@@ -119,22 +120,22 @@ export class PoolFormComponent extends CdForm implements OnInit {
 
   private createForm() {
     const compressionForm = new CdFormGroup({
-      mode: new FormControl('none'),
-      algorithm: new FormControl(''),
-      minBlobSize: new FormControl('', {
+      mode: new UntypedFormControl('none'),
+      algorithm: new UntypedFormControl(''),
+      minBlobSize: new UntypedFormControl('', {
         updateOn: 'blur'
       }),
-      maxBlobSize: new FormControl('', {
+      maxBlobSize: new UntypedFormControl('', {
         updateOn: 'blur'
       }),
-      ratio: new FormControl('', {
+      ratio: new UntypedFormControl('', {
         updateOn: 'blur'
       })
     });
 
     this.form = new CdFormGroup(
       {
-        name: new FormControl('', {
+        name: new UntypedFormControl('', {
           validators: [
             Validators.pattern(/^[.A-Za-z0-9_/-]+$/),
             Validators.required,
@@ -148,14 +149,14 @@ export class PoolFormComponent extends CdForm implements OnInit {
             })
           ]
         }),
-        poolType: new FormControl('', {
+        poolType: new UntypedFormControl('', {
           validators: [Validators.required]
         }),
-        crushRule: new FormControl(null, {
+        crushRule: new UntypedFormControl(null, {
           validators: [
             CdValidators.custom(
               'tooFewOsds',
-              (rule: any) => this.info && rule && this.info.osd_count < rule.min_size
+              (rule: any) => this.info && rule && this.info.osd_count < 1
             ),
             CdValidators.custom(
               'required',
@@ -164,18 +165,18 @@ export class PoolFormComponent extends CdForm implements OnInit {
             )
           ]
         }),
-        size: new FormControl('', {
+        size: new UntypedFormControl('', {
           updateOn: 'blur'
         }),
-        erasureProfile: new FormControl(null),
-        pgNum: new FormControl('', {
+        erasureProfile: new UntypedFormControl(null),
+        pgNum: new UntypedFormControl('', {
           validators: [Validators.required]
         }),
-        pgAutoscaleMode: new FormControl(null),
-        ecOverwrites: new FormControl(false),
+        pgAutoscaleMode: new UntypedFormControl(null),
+        ecOverwrites: new UntypedFormControl(false),
         compression: compressionForm,
-        max_bytes: new FormControl(''),
-        max_objects: new FormControl(0)
+        max_bytes: new UntypedFormControl(''),
+        max_objects: new UntypedFormControl(0)
       },
       [CdValidators.custom('form', (): null => null)]
     );
@@ -420,10 +421,6 @@ export class PoolFormComponent extends CdForm implements OnInit {
     if (!this.info || this.info.osd_count < 1) {
       return 0;
     }
-    const rule = this.form.getValue('crushRule');
-    if (rule) {
-      return rule.min_size;
-    }
     return 1;
   }
 
@@ -467,8 +464,8 @@ export class PoolFormComponent extends CdForm implements OnInit {
       rule.steps[0].item_name
     );
     const currentDomain = domains[rule.steps[1].type];
-    const usable = currentDomain ? currentDomain.length : rule.max_size;
-    rule.usable_size = Math.min(usable, rule.max_size);
+    const usable = currentDomain ? currentDomain.length : this.crushRuleMaxSize;
+    rule.usable_size = Math.min(usable, this.crushRuleMaxSize);
   }
 
   private replicatedPgCalc(pgs: number): number {
@@ -677,7 +674,7 @@ export class PoolFormComponent extends CdForm implements OnInit {
       return;
     }
     const name = value[nameAttribute];
-    this.modalService.show(CriticalConfirmationModalComponent, {
+    this.modalService.show(DeleteConfirmationModalComponent, {
       itemDescription,
       itemNames: [name],
       submitActionObservable: () => {
