@@ -6,7 +6,7 @@ LGPL2.1.  See file COPYING.
 import cephfs
 import rados
 from contextlib import contextmanager
-from mgr_util import CephfsClient, open_filesystem, CephfsConnectionException
+from mgr_util import CephfsClient, open_filesystem
 from collections import OrderedDict
 from datetime import datetime, timezone
 import logging
@@ -15,9 +15,8 @@ from typing import cast, Any, Callable, Dict, Iterator, List, Set, Optional, \
     Tuple, TypeVar, Union, Type
 from types import TracebackType
 import sqlite3
-from .schedule import Schedule, parse_retention
+from .schedule import Schedule
 import traceback
-import errno
 
 
 SNAP_SCHEDULE_NAMESPACE = 'cephfs-snap-schedule'
@@ -37,8 +36,12 @@ SNAPSHOT_PREFIX = 'scheduled'
 log = logging.getLogger(__name__)
 
 
+CephfsClientT = TypeVar('CephfsClientT', bound=CephfsClient)
+
+
 @contextmanager
-def open_ioctx(self, pool):
+def open_ioctx(self: CephfsClientT,
+               pool: Union[int, str]) -> Iterator[rados.Ioctx]:
     try:
         if type(pool) is int:
             with self.mgr.rados.open_ioctx2(pool) as ioctx:
@@ -53,14 +56,18 @@ def open_ioctx(self, pool):
         raise
 
 
-def updates_schedule_db(func):
-    def f(self, fs, schedule_or_path, *args):
-        func(self, fs, schedule_or_path, *args)
+FuncT = TypeVar('FuncT', bound=Callable[..., None])
+
+
+def updates_schedule_db(func: FuncT) -> FuncT:
+    def f(self: 'SnapSchedClient', fs: str, schedule_or_path: str, *args: Any) -> None:
+        ret = func(self, fs, schedule_or_path, *args)
         path = schedule_or_path
         if isinstance(schedule_or_path, Schedule):
             path = schedule_or_path.path
         self.refresh_snap_timers(fs, path)
-    return f
+        return ret
+    return cast(FuncT, f)
 
 
 def get_prune_set(candidates: Set[Tuple[cephfs.DirEntry, datetime]],
@@ -72,11 +79,11 @@ def get_prune_set(candidates: Set[Tuple[cephfs.DirEntry, datetime]],
         # NOTE: prune set has tz suffix stripped out.
         ("n", SNAPSHOT_TS_FORMAT),
         # TODO remove M for release
-        ("M", '%Y-%m-%d-%H_%M'),
+        ("m", '%Y-%m-%d-%H_%M'),
         ("h", '%Y-%m-%d-%H'),
         ("d", '%Y-%m-%d'),
         ("w", '%G-%V'),
-        ("m", '%Y-%m'),
+        ("M", '%Y-%m'),
         ("y", '%Y'),
     ])
     keep = []
@@ -96,7 +103,8 @@ def get_prune_set(candidates: Set[Tuple[cephfs.DirEntry, datetime]],
             if snap_ts != last:
                 last = snap_ts
                 if snap not in keep:
-                    log.debug(f'keeping {snap[0].d_name} due to {period_count}{period}')
+                    log.debug((f'keeping {snap[0].d_name} due to '
+                               f'{period_count}{period}'))
                     keep.append(snap)
                     kept_for_this_period += 1
                     if kept_for_this_period == period_count:
@@ -145,7 +153,7 @@ class DBConnectionManager():
 
 class SnapSchedClient(CephfsClient):
 
-    def __init__(self, mgr):
+    def __init__(self, mgr: Any) -> None:
         super(SnapSchedClient, self).__init__(mgr)
         # Each db connection is now guarded by a Lock; this is required to
         # avoid concurrent DB transactions when more than one paths in a
@@ -165,56 +173,90 @@ class SnapSchedClient(CephfsClient):
                     self.refresh_snap_timers(fs_name, sched.path, db)
 
     @property
-    def allow_minute_snaps(self):
+    def allow_minute_snaps(self) -> None:
         return self.mgr.get_module_option('allow_m_granularity')
 
     @property
     def dump_on_update(self) -> None:
         return self.mgr.get_module_option('dump_on_update')
 
-    def get_schedule_db(self, fs):
+    def _create_snap_schedule_kv_db(self, db: sqlite3.Connection) -> None:
+        SQL = """
+        CREATE TABLE IF NOT EXISTS SnapScheduleModuleKV (
+          key TEXT PRIMARY KEY,
+          value NOT NULL
+        ) WITHOUT ROWID;
+        INSERT OR IGNORE INTO SnapScheduleModuleKV (key, value) VALUES ('__snap_schedule_db_version', 1);
+        """
+        db.executescript(SQL)
+
+    def _get_snap_schedule_db_version(self, db: sqlite3.Connection) -> int:
+        SQL = """
+        SELECT value
+        FROM SnapScheduleModuleKV
+        WHERE key = '__snap_schedule_db_version';
+        """
+        cur = db.execute(SQL)
+        row = cur.fetchone()
+        assert row is not None
+        return int(row[0])
+
+    # add all upgrades here
+    def _upgrade_snap_schedule_db_schema(self, db: sqlite3.Connection) -> None:
+        # add a column to hold the subvolume group name
+        if self._get_snap_schedule_db_version(db) < 2:
+            SQL = """
+            ALTER TABLE schedules
+            ADD COLUMN group_name TEXT;
+            """
+            db.executescript(SQL)
+
+            # bump up the snap-schedule db version to 2
+            SQL = "UPDATE OR ROLLBACK SnapScheduleModuleKV SET value = ? WHERE key = '__snap_schedule_db_version';"
+            db.execute(SQL, (2,))
+
+    def get_schedule_db(self, fs: str) -> DBConnectionManager:
         dbinfo = None
         self.conn_lock.acquire()
         if fs not in self.sqlite_connections:
-            db = sqlite3.connect(':memory:', check_same_thread=False)
-            with db:
-                db.row_factory = sqlite3.Row
-                db.execute("PRAGMA FOREIGN_KEYS = 1")
-                pool = self.get_metadata_pool(fs)
-                with open_ioctx(self, pool) as ioctx:
-                    try:
-                        size, _mtime = ioctx.stat(SNAP_DB_OBJECT_NAME)
-                        ddl = ioctx.read(SNAP_DB_OBJECT_NAME,
-                                         size).decode('utf-8')
-                        db.executescript(ddl)
-                    except rados.ObjectNotFound:
-                        log.debug(f'No schedule DB found in {fs}, creating one.')
-                        db.executescript(Schedule.CREATE_TABLES)
+            poolid = self.get_metadata_pool(fs)
+            assert poolid, f'fs "{fs}" not found'
+            uri = f"file:///*{poolid}:/{SNAP_DB_OBJECT_NAME}.db?vfs=ceph"
+            log.debug(f"using uri {uri}")
+            db = sqlite3.connect(uri, check_same_thread=False, uri=True)
+            db.execute('PRAGMA FOREIGN_KEYS = 1')
+            db.execute('PRAGMA JOURNAL_MODE = PERSIST')
+            db.execute('PRAGMA PAGE_SIZE = 65536')
+            db.execute('PRAGMA CACHE_SIZE = 256')
+            db.execute('PRAGMA TEMP_STORE = memory')
+            db.row_factory = sqlite3.Row
+            # check for legacy dump store
+            pool_param = cast(Union[int, str], poolid)
+            with open_ioctx(self, pool_param) as ioctx:
+                try:
+                    size, _mtime = ioctx.stat(SNAP_DB_OBJECT_NAME)
+                    dump = ioctx.read(SNAP_DB_OBJECT_NAME, size).decode('utf-8')
+                    db.executescript(dump)
+                    ioctx.remove_object(SNAP_DB_OBJECT_NAME)
+                except rados.ObjectNotFound:
+                    log.debug(f'No legacy schedule DB found in {fs}')
+            db.executescript(Schedule.CREATE_TABLES)
+            self._create_snap_schedule_kv_db(db)
+            self._upgrade_snap_schedule_db_schema(db)
             self.sqlite_connections[fs] = DBInfo(fs, db)
         dbinfo = self.sqlite_connections[fs]
         self.conn_lock.release()
         return DBConnectionManager(dbinfo)
 
-    def store_schedule_db(self, fs, db):
-        # only store db is it exists, otherwise nothing to do
-        metadata_pool = self.get_metadata_pool(fs)
-        if not metadata_pool:
-            raise CephfsConnectionException(
-                -errno.ENOENT, "Filesystem {} does not exist".format(fs))
-        db_content = []
-        for row in db.iterdump():
-            db_content.append(row)
-        with open_ioctx(self, metadata_pool) as ioctx:
-            ioctx.write_full(SNAP_DB_OBJECT_NAME,
-                             '\n'.join(db_content).encode('utf-8'))
-
-    def _is_allowed_repeat(self, exec_row, path):
-        if Schedule.parse_schedule(exec_row['schedule'])[1] == 'M':
+    def _is_allowed_repeat(self, exec_row: Dict[str, str], path: str) -> bool:
+        if Schedule.parse_schedule(exec_row['schedule'])[1] == 'm':
             if self.allow_minute_snaps:
-                log.debug(f'Minute repeats allowed, scheduling snapshot on path {path}')
+                log.debug(('Minute repeats allowed, '
+                           f'scheduling snapshot on path {path}'))
                 return True
             else:
-                log.info(f'Minute repeats disabled, skipping snapshot on path {path}')
+                log.info(('Minute repeats disabled, '
+                          f'skipping snapshot on path {path}'))
                 return False
         else:
             return True
@@ -260,16 +302,21 @@ class SnapSchedClient(CephfsClient):
         except Exception:
             self._log_exception('refresh_snap_timers')
 
-    def _log_exception(self, fct):
+    def _log_exception(self, fct: str) -> None:
         log.error(f'{fct} raised an exception:')
         log.error(traceback.format_exc())
 
-    def create_scheduled_snapshot(self, fs_name, path, retention, start, repeat):
+    def create_scheduled_snapshot(self,
+                                  fs_name: str,
+                                  path: str,
+                                  retention: str,
+                                  start: str,
+                                  repeat: str) -> None:
         log.debug(f'Scheduled snapshot of {path} triggered')
         set_schedule_to_inactive = False
-        with self.get_schedule_db(fs_name) as conn_mgr:
-            db = conn_mgr.dbinfo.db
-            try:
+        try:
+            with self.get_schedule_db(fs_name) as conn_mgr:
+                db = conn_mgr.dbinfo.db
                 try:
                     sched = Schedule.get_db_schedules(path,
                                                       db,
@@ -301,19 +348,20 @@ class SnapSchedClient(CephfsClient):
                 finally:
                     if set_schedule_to_inactive:
                         sched.set_inactive(db)
-            finally:
+        finally:
+            with self.get_schedule_db(fs_name) as conn_mgr:
+                db = conn_mgr.dbinfo.db
                 self.refresh_snap_timers(fs_name, path, db)
-                self.prune_snapshots(sched, db)
-                self.store_schedule_db(fs_name, db)
+            self.prune_snapshots(sched)
 
-    def prune_snapshots(self, sched, db):
+    def prune_snapshots(self, sched: Schedule) -> None:
         try:
             log.debug('Pruning snapshots')
             ret = sched.retention
             path = sched.path
             prune_candidates = set()
             time = datetime.now(timezone.utc)
-            mds_max_snaps_per_dir = self.mgr.get_ceph_option('mds_max_snaps_per_dir')
+            mds_max_snaps_per_dir = self.mgr.get_foreign_ceph_option('mds', 'mds_max_snaps_per_dir')
             with open_filesystem(self, sched.fs) as fs_handle:
                 snap_dir = self.mgr.rados.conf_get('client_snapdir')
                 with fs_handle.opendir(f'{path}/{snap_dir}') as d_handle:
@@ -335,7 +383,9 @@ class SnapSchedClient(CephfsClient):
                     log.debug(f'rmdir on {dirname}')
                     fs_handle.rmdir(f'{path}/{snap_dir}/{dirname}')
                 if to_prune:
-                    sched.update_pruned(time, db, len(to_prune))
+                    with self.get_schedule_db(sched.fs) as conn_mgr:
+                        db = conn_mgr.dbinfo.db
+                        sched.update_pruned(time, db, len(to_prune))
         except Exception:
             self._log_exception('prune_snapshots')
 
@@ -354,17 +404,21 @@ class SnapSchedClient(CephfsClient):
 
     @updates_schedule_db
     # TODO improve interface
-    def store_snap_schedule(self, fs, path_, args):
+    def store_snap_schedule(self,
+                            fs: str, path_: str,
+                            args: Tuple[str, str, str, str,
+                                        Optional[str], Optional[str],
+                                        Optional[str]]) -> None:
         sched = Schedule(*args)
         log.debug(f'repeat is {sched.repeat}')
-        if sched.parse_schedule(sched.schedule)[1] == 'M' and not self.allow_minute_snaps:
+        if sched.parse_schedule(sched.schedule)[1] == 'm' and not self.allow_minute_snaps:
             log.error('not allowed')
-            raise ValueError('no minute snaps allowed')
+            raise ValueError('invalid schedule specified - multiplier should '
+                             'be one of h,d,w,M,y')
         log.debug(f'attempting to add schedule {sched}')
         with self.get_schedule_db(fs) as conn_mgr:
             db = conn_mgr.dbinfo.db
             sched.store_schedule(db)
-            self.store_schedule_db(sched.fs, db)
 
     @updates_schedule_db
     def rm_snap_schedule(self,
@@ -374,35 +428,32 @@ class SnapSchedClient(CephfsClient):
         with self.get_schedule_db(fs) as conn_mgr:
             db = conn_mgr.dbinfo.db
             Schedule.rm_schedule(db, path, schedule, start)
-            self.store_schedule_db(fs, db)
 
     @updates_schedule_db
     def add_retention_spec(self,
-                           fs,
-                           path,
-                           retention_spec_or_period,
-                           retention_count):
+                           fs: str,
+                           path: str,
+                           retention_spec_or_period: str,
+                           retention_count: Optional[str]) -> None:
         retention_spec = retention_spec_or_period
         if retention_count:
             retention_spec = retention_count + retention_spec
         with self.get_schedule_db(fs) as conn_mgr:
             db = conn_mgr.dbinfo.db
             Schedule.add_retention(db, path, retention_spec)
-            self.store_schedule_db(fs, db)
 
     @updates_schedule_db
     def rm_retention_spec(self,
-                          fs,
-                          path,
-                          retention_spec_or_period,
-                          retention_count):
+                          fs: str,
+                          path: str,
+                          retention_spec_or_period: str,
+                          retention_count: Optional[str]) -> None:
         retention_spec = retention_spec_or_period
         if retention_count:
             retention_spec = retention_count + retention_spec
         with self.get_schedule_db(fs) as conn_mgr:
             db = conn_mgr.dbinfo.db
             Schedule.rm_retention(db, path, retention_spec)
-            self.store_schedule_db(fs, db)
 
     @updates_schedule_db
     def activate_snap_schedule(self,
@@ -417,7 +468,6 @@ class SnapSchedClient(CephfsClient):
                                                   start=start)
             for s in schedules:
                 s.set_active(db)
-            self.store_schedule_db(fs, db)
 
     @updates_schedule_db
     def deactivate_snap_schedule(self,
@@ -431,4 +481,3 @@ class SnapSchedClient(CephfsClient):
                                                   start=start)
             for s in schedules:
                 s.set_inactive(db)
-            self.store_schedule_db(fs, db)
