@@ -10,6 +10,7 @@
 
 #include "rgw_acl_s3.h"
 #include "rgw_user.h"
+#include "rgw_sal.h"
 
 #define dout_subsys ceph_subsys_rgw
 
@@ -18,6 +19,8 @@
 #define RGW_URI_ALL_USERS	"http://acs.amazonaws.com/groups/global/AllUsers"
 #define RGW_URI_AUTH_USERS	"http://acs.amazonaws.com/groups/global/AuthenticatedUsers"
 
+using namespace std;
+
 static string rgw_uri_all_users = RGW_URI_ALL_USERS;
 static string rgw_uri_auth_users = RGW_URI_AUTH_USERS;
 
@@ -290,15 +293,13 @@ static const char *get_acl_header(const RGWEnv *env,
   return env->get(header, NULL);
 }
 
-static int parse_grantee_str(const DoutPrefixProvider *dpp, RGWUserCtl *user_ctl, string& grantee_str,
+static int parse_grantee_str(const DoutPrefixProvider *dpp, rgw::sal::Driver* driver, string& grantee_str,
         const struct s3_acl_header *perm, ACLGrant& grant)
 {
   string id_type, id_val_quoted;
   int rgw_perm = perm->rgw_perm;
   int ret;
 
-  RGWUserInfo info;
-
   ret = parse_key_value(grantee_str, id_type, id_val_quoted);
   if (ret < 0)
     return ret;
@@ -306,18 +307,19 @@ static int parse_grantee_str(const DoutPrefixProvider *dpp, RGWUserCtl *user_ctl
   string id_val = rgw_trim_quotes(id_val_quoted);
 
   if (strcasecmp(id_type.c_str(), "emailAddress") == 0) {
-    ret = user_ctl->get_info_by_email(dpp, id_val, &info, null_yield);
+    std::unique_ptr<rgw::sal::User> user;
+    ret = driver->get_user_by_email(dpp, id_val, null_yield, &user);
     if (ret < 0)
       return ret;
 
-    grant.set_canon(info.user_id, info.display_name, rgw_perm);
+    grant.set_canon(user->get_id(), user->get_display_name(), rgw_perm);
   } else if (strcasecmp(id_type.c_str(), "id") == 0) {
-    rgw_user user(id_val);
-    ret = user_ctl->get_info_by_uid(dpp, user, &info, null_yield);
+    std::unique_ptr<rgw::sal::User> user = driver->get_user(rgw_user(id_val));
+    ret = user->load_user(dpp, null_yield);
     if (ret < 0)
       return ret;
 
-    grant.set_canon(info.user_id, info.display_name, rgw_perm);
+    grant.set_canon(user->get_id(), user->get_display_name(), rgw_perm);
   } else if (strcasecmp(id_type.c_str(), "uri") == 0) {
     ACLGroupTypeEnum gid = grant.uri_to_group(id_val);
     if (gid == ACL_GROUP_NONE)
@@ -331,8 +333,9 @@ static int parse_grantee_str(const DoutPrefixProvider *dpp, RGWUserCtl *user_ctl
   return 0;
 }
 
-static int parse_acl_header(const DoutPrefixProvider *dpp, RGWUserCtl *user_ctl, const RGWEnv *env,
-         const struct s3_acl_header *perm, std::list<ACLGrant>& _grants)
+static int parse_acl_header(const DoutPrefixProvider *dpp, rgw::sal::Driver* driver,
+			    const RGWEnv *env, const struct s3_acl_header *perm,
+			    std::list<ACLGrant>& _grants)
 {
   std::list<string> grantees;
   std::string hacl_str;
@@ -346,7 +349,7 @@ static int parse_acl_header(const DoutPrefixProvider *dpp, RGWUserCtl *user_ctl,
 
   for (list<string>::iterator it = grantees.begin(); it != grantees.end(); ++it) {
     ACLGrant grant;
-    int ret = parse_grantee_str(dpp, user_ctl, *it, perm, grant);
+    int ret = parse_grantee_str(dpp, driver, *it, perm, grant);
     if (ret < 0)
       return ret;
 
@@ -451,13 +454,15 @@ static const s3_acl_header acl_header_perms[] = {
   {0, NULL}
 };
 
-int RGWAccessControlPolicy_S3::create_from_headers(const DoutPrefixProvider *dpp, RGWUserCtl *user_ctl, const RGWEnv *env, ACLOwner& _owner)
+int RGWAccessControlPolicy_S3::create_from_headers(const DoutPrefixProvider *dpp,
+						   rgw::sal::Driver* driver,
+						   const RGWEnv *env, ACLOwner& _owner)
 {
   std::list<ACLGrant> grants;
   int r = 0;
 
   for (const struct s3_acl_header *p = acl_header_perms; p->rgw_perm; p++) {
-    r = parse_acl_header(dpp, user_ctl, env, p, grants);
+    r = parse_acl_header(dpp, driver, env, p, grants);
     if (r < 0) {
       return r;
     }
@@ -474,8 +479,9 @@ int RGWAccessControlPolicy_S3::create_from_headers(const DoutPrefixProvider *dpp
 /*
   can only be called on object that was parsed
  */
-int RGWAccessControlPolicy_S3::rebuild(const DoutPrefixProvider *dpp, RGWUserCtl *user_ctl, ACLOwner *owner, RGWAccessControlPolicy& dest,
-                                       std::string &err_msg)
+int RGWAccessControlPolicy_S3::rebuild(const DoutPrefixProvider *dpp,
+				       rgw::sal::Driver* driver, ACLOwner *owner,
+				       RGWAccessControlPolicy& dest, std::string &err_msg)
 {
   if (!owner)
     return -EINVAL;
@@ -487,15 +493,15 @@ int RGWAccessControlPolicy_S3::rebuild(const DoutPrefixProvider *dpp, RGWUserCtl
       return -EPERM;
   }
 
-  RGWUserInfo owner_info;
-  if (user_ctl->get_info_by_uid(dpp, owner->get_id(), &owner_info, null_yield) < 0) {
+  std::unique_ptr<rgw::sal::User> user = driver->get_user(owner->get_id());
+  if (user->load_user(dpp, null_yield) < 0) {
     ldpp_dout(dpp, 10) << "owner info does not exist" << dendl;
     err_msg = "Invalid id";
     return -EINVAL;
   }
   ACLOwner& dest_owner = dest.get_owner();
   dest_owner.set_id(owner->get_id());
-  dest_owner.set_name(owner_info.display_name);
+  dest_owner.set_name(user->get_display_name());
 
   ldpp_dout(dpp, 20) << "owner id=" << owner->get_id() << dendl;
   ldpp_dout(dpp, 20) << "dest owner id=" << dest.get_owner().get_id() << dendl;
@@ -522,11 +528,12 @@ int RGWAccessControlPolicy_S3::rebuild(const DoutPrefixProvider *dpp, RGWUserCtl
         }
         email = u.id;
         ldpp_dout(dpp, 10) << "grant user email=" << email << dendl;
-        if (user_ctl->get_info_by_email(dpp, email, &grant_user, null_yield) < 0) {
+	if (driver->get_user_by_email(dpp, email, null_yield, &user) < 0) {
           ldpp_dout(dpp, 10) << "grant user email not found or other error" << dendl;
           err_msg = "The e-mail address you provided does not match any account on record.";
           return -ERR_UNRESOLVABLE_EMAIL;
         }
+	grant_user = user->get_info();
         uid = grant_user.user_id;
       }
     case ACL_TYPE_CANON_USER:
@@ -539,18 +546,22 @@ int RGWAccessControlPolicy_S3::rebuild(const DoutPrefixProvider *dpp, RGWUserCtl
           }
         }
     
-        if (grant_user.user_id.empty() && user_ctl->get_info_by_uid(dpp, uid, &grant_user, null_yield) < 0) {
-          ldpp_dout(dpp, 10) << "grant user does not exist:" << uid << dendl;
-          err_msg = "Invalid id";
-          return -EINVAL;
-        } else {
-          ACLPermission& perm = src_grant.get_permission();
-          new_grant.set_canon(uid, grant_user.display_name, perm.get_permissions());
-          grant_ok = true;
-          rgw_user new_id;
-          new_grant.get_id(new_id);
-          ldpp_dout(dpp, 10) << "new grant: " << new_id << ":" << grant_user.display_name << dendl;
+        if (grant_user.user_id.empty()) {
+	  user = driver->get_user(uid);
+	  if (user->load_user(dpp, null_yield) < 0) {
+	    ldpp_dout(dpp, 10) << "grant user does not exist:" << uid << dendl;
+	    err_msg = "Invalid id";
+	    return -EINVAL;
+	  } else {
+	    grant_user = user->get_info();
+	  }
         }
+	ACLPermission& perm = src_grant.get_permission();
+	new_grant.set_canon(uid, grant_user.display_name, perm.get_permissions());
+	grant_ok = true;
+	rgw_user new_id;
+	new_grant.get_id(new_id);
+	ldpp_dout(dpp, 10) << "new grant: " << new_id << ":" << grant_user.display_name << dendl;
       }
       break;
     case ACL_TYPE_GROUP:
@@ -620,3 +631,13 @@ XMLObj *RGWACLXMLParser_S3::alloc_obj(const char *el)
   return obj;
 }
 
+ACLGroupTypeEnum ACLGrant_S3::uri_to_group(string& uri)
+{
+  if (uri.compare(rgw_uri_all_users) == 0)
+    return ACL_GROUP_ALL_USERS;
+  else if (uri.compare(rgw_uri_auth_users) == 0)
+    return ACL_GROUP_AUTHENTICATED_USERS;
+
+  return ACL_GROUP_NONE;
+}
+
