@@ -1,8 +1,7 @@
 // -*- mode:C++; tab-width:8; c-basic-offset:2; indent-tabs-mode:t -*-
 // vim: ts=8 sw=2 smarttab ft=cpp
 
-#ifndef RGW_FILE_H
-#define RGW_FILE_H
+#pragma once
 
 #include "include/rados/rgw_file.h"
 
@@ -259,7 +258,7 @@ namespace rgw {
     static constexpr uint32_t FLAG_ROOT =    0x0002;
     static constexpr uint32_t FLAG_CREATE =  0x0004;
     static constexpr uint32_t FLAG_CREATING =  0x0008;
-    static constexpr uint32_t FLAG_SYMBOLIC_LINK = 0x0009;
+    static constexpr uint32_t FLAG_SYMBOLIC_LINK = 0x0009; // XXXX bug?
     static constexpr uint32_t FLAG_DIRECTORY = 0x0010;
     static constexpr uint32_t FLAG_BUCKET = 0x0020;
     static constexpr uint32_t FLAG_LOCK =   0x0040;
@@ -432,7 +431,7 @@ namespace rgw {
     }
 
     directory* get_directory() {
-      return get<directory>(&variant_type);
+      return boost::get<directory>(&variant_type);
     }
 
     size_t get_size() const { return state.size; }
@@ -869,10 +868,10 @@ namespace rgw {
 
     RGWFileHandle::FHCache fh_cache;
     RGWFileHandle::FhLRU fh_lru;
-    
+
     std::string uid; // should match user.user_id, iiuc
 
-    RGWUserInfo user;
+    std::unique_ptr<rgw::sal::User> user;
     RGWAccessKey key; // XXXX acc_key
 
     static std::atomic<uint32_t> fs_inst_counter;
@@ -1000,17 +999,17 @@ namespace rgw {
       (void) fh_lru.unref(fh, cohort::lru::FLAG_NONE);
     }
 
-    int authorize(const DoutPrefixProvider *dpp, rgw::sal::RGWRadosStore* store) {
-      int ret = store->ctl()->user->get_info_by_access_key(dpp, key.id, &user, null_yield);
+    int authorize(const DoutPrefixProvider *dpp, rgw::sal::Driver* driver) {
+      int ret = driver->get_user_by_access_key(dpp, key.id, null_yield, &user);
       if (ret == 0) {
-	RGWAccessKey* k = user.get_key(key.id);
+	RGWAccessKey* k = user->get_info().get_key(key.id);
 	if (!k || (k->key != key.key))
 	  return -EINVAL;
-	if (user.suspended)
+	if (user->get_info().suspended)
 	  return -ERR_USER_SUSPENDED;
       } else {
 	/* try external authenticators (ldap for now) */
-	rgw::LDAPHelper* ldh = rgwlib.get_ldh(); /* !nullptr */
+	rgw::LDAPHelper* ldh = g_rgwlib->get_ldh(); /* !nullptr */
 	RGWToken token;
 	/* boost filters and/or string_ref may throw on invalid input */
 	try {
@@ -1019,14 +1018,12 @@ namespace rgw {
 	  token = std::string("");
 	}
 	if (token.valid() && (ldh->auth(token.id, token.key) == 0)) {
-	  /* try to store user if it doesn't already exist */
-	  if (store->ctl()->user->get_info_by_uid(dpp, rgw_user(token.id), &user, null_yield) < 0) {
-	    int ret = store->ctl()->user->store_info(dpp, user, null_yield,
-                                                  RGWUserCtl::PutParams()
-                                                  .set_exclusive(true));
+	  /* try to driver user if it doesn't already exist */
+	  if (user->load_user(dpp, null_yield) < 0) {
+	    int ret = user->store_user(dpp, null_yield, true);
 	    if (ret < 0) {
 	      lsubdout(get_context(), rgw, 10)
-		<< "NOTICE: failed to store new user's info: ret=" << ret
+		<< "NOTICE: failed to driver new user's info: ret=" << ret
 		<< dendl;
 	    }
 	  }
@@ -1046,7 +1043,7 @@ namespace rgw {
 			     const uint32_t flags = RGWFileHandle::FLAG_NONE) {
       using std::get;
 
-      // cast int32_t(RGWFileHandle::FLAG_NONE) due to strictness of Clang 
+      // cast int32_t(RGWFileHandle::FLAG_NONE) due to strictness of Clang
       // the cast transfers a lvalue into a rvalue  in the ctor
       // check the commit message for the full details
       LookupFHResult fhr { nullptr, uint32_t(RGWFileHandle::FLAG_NONE) };
@@ -1089,7 +1086,7 @@ namespace rgw {
 			     const uint32_t flags = RGWFileHandle::FLAG_NONE) {
       using std::get;
 
-      // cast int32_t(RGWFileHandle::FLAG_NONE) due to strictness of Clang 
+      // cast int32_t(RGWFileHandle::FLAG_NONE) due to strictness of Clang
       // the cast transfers a lvalue into a rvalue  in the ctor
       // check the commit message for the full details
       LookupFHResult fhr { nullptr, uint32_t(RGWFileHandle::FLAG_NONE) };
@@ -1239,8 +1236,8 @@ namespace rgw {
 
     MkObjResult create(RGWFileHandle* parent, const char *name, struct stat *st,
 		      uint32_t mask, uint32_t flags);
-    
-    MkObjResult symlink(RGWFileHandle* parent, const char *name, 
+
+    MkObjResult symlink(RGWFileHandle* parent, const char *name,
                const char *link_path, struct stat *st, uint32_t mask, uint32_t flags);
 
     MkObjResult mkdir(RGWFileHandle* parent, const char *name, struct stat *st,
@@ -1313,14 +1310,10 @@ namespace rgw {
 
     uint64_t get_fsid() { return root_fh.state.dev; }
 
-    RGWUserInfo* get_user() { return &user; }
+    RGWUserInfo* get_user() { return &user->get_info(); }
 
     void update_user(const DoutPrefixProvider *dpp) {
-      RGWUserInfo _user = user;
-      auto user_ctl = rgwlib.get_store()->ctl()->user;
-      int ret = user_ctl->get_info_by_access_key(dpp, key.id, &user, null_yield);
-      if (ret != 0)
-        user = _user;
+      (void) g_rgwlib->get_driver()->get_user_by_access_key(dpp, key.id, null_yield, &user);
     }
 
     void close();
@@ -1354,7 +1347,7 @@ public:
   uint32_t d_count;
   bool rcb_eof; // caller forced early stop in readdir cycle
 
-  RGWListBucketsRequest(CephContext* _cct, std::unique_ptr<rgw::sal::RGWUser> _user,
+  RGWListBucketsRequest(CephContext* _cct, std::unique_ptr<rgw::sal::User> _user,
 			RGWFileHandle* _rgw_fh, rgw_readdir_cb _rcb,
 			void* _cb_arg, RGWFileHandle::readdir_offset& _offset)
     : RGWLibRequest(_cct, std::move(_user)), rgw_fh(_rgw_fh), offset(_offset),
@@ -1381,18 +1374,15 @@ public:
   bool only_bucket() override { return false; }
 
   int op_init() override {
-    // assign store, s, and dialect_handler
-    RGWObjectCtx* rados_ctx
-      = static_cast<RGWObjectCtx*>(get_state()->obj_ctx);
+    // assign driver, s, and dialect_handler
     // framework promises to call op_init after parent init
-    ceph_assert(rados_ctx);
-    RGWOp::init(rados_ctx->get_store(), get_state(), this);
+    RGWOp::init(RGWHandler::driver, get_state(), this);
     op = this; // assign self as op: REQUIRED
     return 0;
   }
 
   int header_init() override {
-    struct req_state* state = get_state();
+    req_state* state = get_state();
     state->info.method = "GET";
     state->op = OP_GET;
 
@@ -1415,7 +1405,7 @@ public:
     sent_data = true;
   }
 
-  void send_response_data(rgw::sal::RGWBucketList& buckets) override {
+  void send_response_data(rgw::sal::BucketList& buckets) override {
     if (!sent_data)
       return;
     auto& m = buckets.get_buckets();
@@ -1453,6 +1443,8 @@ public:
   }
 
   bool eof() {
+    using boost::get;
+
     if (unlikely(cct->_conf->subsys.should_gather(ceph_subsys_rgw, 15))) {
       bool is_offset =
 	unlikely(! get<const char*>(&offset)) ||
@@ -1484,7 +1476,7 @@ public:
   uint32_t d_count;
   bool rcb_eof; // caller forced early stop in readdir cycle
 
-  RGWReaddirRequest(CephContext* _cct, std::unique_ptr<rgw::sal::RGWUser> _user,
+  RGWReaddirRequest(CephContext* _cct, std::unique_ptr<rgw::sal::User> _user,
 		    RGWFileHandle* _rgw_fh, rgw_readdir_cb _rcb,
 		    void* _cb_arg, RGWFileHandle::readdir_offset& _offset)
     : RGWLibRequest(_cct, std::move(_user)), rgw_fh(_rgw_fh), offset(_offset),
@@ -1517,18 +1509,15 @@ public:
   bool only_bucket() override { return true; }
 
   int op_init() override {
-    // assign store, s, and dialect_handler
-    RGWObjectCtx* rados_ctx
-      = static_cast<RGWObjectCtx*>(get_state()->obj_ctx);
+    // assign driver, s, and dialect_handler
     // framework promises to call op_init after parent init
-    ceph_assert(rados_ctx);
-    RGWOp::init(rados_ctx->get_store(), get_state(), this);
+    RGWOp::init(RGWHandler::driver, get_state(), this);
     op = this; // assign self as op: REQUIRED
     return 0;
   }
 
   int header_init() override {
-    struct req_state* state = get_state();
+    req_state* state = get_state();
     state->info.method = "GET";
     state->op = OP_GET;
 
@@ -1588,7 +1577,7 @@ public:
   }
 
   void send_response() override {
-    struct req_state* state = get_state();
+    req_state* state = get_state();
     auto cnow = real_clock::now();
 
     /* enumerate objs and common_prefixes in parallel,
@@ -1597,11 +1586,11 @@ public:
 
     class DirIterator
     {
-      vector<rgw_bucket_dir_entry>& objs;
-      vector<rgw_bucket_dir_entry>::iterator obj_iter;
+      std::vector<rgw_bucket_dir_entry>& objs;
+      std::vector<rgw_bucket_dir_entry>::iterator obj_iter;
 
-      map<string, bool>& common_prefixes;
-      map<string, bool>::iterator cp_iter;
+      std::map<std::string, bool>& common_prefixes;
+      std::map<string, bool>::iterator cp_iter;
 
       boost::optional<std::string_view> obj_sref;
       boost::optional<std::string_view> cp_sref;
@@ -1609,8 +1598,8 @@ public:
 
     public:
 
-      DirIterator(vector<rgw_bucket_dir_entry>& objs,
-		  map<string, bool>& common_prefixes)
+      DirIterator(std::vector<rgw_bucket_dir_entry>& objs,
+		  std::map<string, bool>& common_prefixes)
 	: objs(objs), common_prefixes(common_prefixes), _skip_cp(false)
 	{
 	  obj_iter = objs.begin();
@@ -1692,11 +1681,11 @@ public:
 	return cp_sref.get();
       }
 
-      vector<rgw_bucket_dir_entry>::iterator& get_obj_iter() {
+      std::vector<rgw_bucket_dir_entry>::iterator& get_obj_iter() {
 	return obj_iter;
       }
 
-      map<string, bool>::iterator& get_cp_iter() {
+      std::map<string, bool>::iterator& get_cp_iter() {
 	return cp_iter;
       }
 
@@ -1781,6 +1770,8 @@ public:
   }
 
   bool eof() {
+    using boost::get;
+
     if (unlikely(cct->_conf->subsys.should_gather(ceph_subsys_rgw, 15))) {
       bool is_offset =
 	unlikely(! get<const char*>(&offset)) ||
@@ -1808,7 +1799,7 @@ public:
   bool valid;
   bool has_children;
 
-  RGWRMdirCheck (CephContext* _cct, std::unique_ptr<rgw::sal::RGWUser> _user,
+  RGWRMdirCheck (CephContext* _cct, std::unique_ptr<rgw::sal::User> _user,
 		 const RGWFileHandle* _rgw_fh)
     : RGWLibRequest(_cct, std::move(_user)), rgw_fh(_rgw_fh), valid(false),
       has_children(false) {
@@ -1819,18 +1810,15 @@ public:
   bool only_bucket() override { return true; }
 
   int op_init() override {
-    // assign store, s, and dialect_handler
-    RGWObjectCtx* rados_ctx
-      = static_cast<RGWObjectCtx*>(get_state()->obj_ctx);
+    // assign driver, s, and dialect_handler
     // framework promises to call op_init after parent init
-    ceph_assert(rados_ctx);
-    RGWOp::init(rados_ctx->get_store(), get_state(), this);
+    RGWOp::init(RGWHandler::driver, get_state(), this);
     op = this; // assign self as op: REQUIRED
     return 0;
   }
 
   int header_init() override {
-    struct req_state* state = get_state();
+    req_state* state = get_state();
     state->info.method = "GET";
     state->op = OP_GET;
 
@@ -1887,7 +1875,7 @@ class RGWCreateBucketRequest : public RGWLibRequest,
 public:
   const std::string& bucket_name;
 
-  RGWCreateBucketRequest(CephContext* _cct, std::unique_ptr<rgw::sal::RGWUser> _user,
+  RGWCreateBucketRequest(CephContext* _cct, std::unique_ptr<rgw::sal::User> _user,
 			std::string& _bname)
     : RGWLibRequest(_cct, std::move(_user)), bucket_name(_bname) {
     op = this;
@@ -1901,19 +1889,16 @@ public:
   }
 
   int op_init() override {
-    // assign store, s, and dialect_handler
-    RGWObjectCtx* rados_ctx
-      = static_cast<RGWObjectCtx*>(get_state()->obj_ctx);
+    // assign driver, s, and dialect_handler
     // framework promises to call op_init after parent init
-    ceph_assert(rados_ctx);
-    RGWOp::init(rados_ctx->get_store(), get_state(), this);
+    RGWOp::init(RGWHandler::driver, get_state(), this);
     op = this; // assign self as op: REQUIRED
     return 0;
   }
 
   int header_init() override {
 
-    struct req_state* state = get_state();
+    req_state* state = get_state();
     state->info.method = "PUT";
     state->op = OP_PUT;
 
@@ -1929,7 +1914,7 @@ public:
   }
 
   int get_params(optional_yield) override {
-    struct req_state* state = get_state();
+    req_state* state = get_state();
     RGWAccessControlPolicy_S3 s3policy(state->cct);
     /* we don't have (any) headers, so just create canned ACLs */
     int ret = s3policy.create_canned(state->owner, state->bucket_owner, state->canned_acl);
@@ -1952,7 +1937,7 @@ class RGWDeleteBucketRequest : public RGWLibRequest,
 public:
   const std::string& bucket_name;
 
-  RGWDeleteBucketRequest(CephContext* _cct, std::unique_ptr<rgw::sal::RGWUser> _user,
+  RGWDeleteBucketRequest(CephContext* _cct, std::unique_ptr<rgw::sal::User> _user,
 			std::string& _bname)
     : RGWLibRequest(_cct, std::move(_user)), bucket_name(_bname) {
     op = this;
@@ -1961,19 +1946,16 @@ public:
   bool only_bucket() override { return true; }
 
   int op_init() override {
-    // assign store, s, and dialect_handler
-    RGWObjectCtx* rados_ctx
-      = static_cast<RGWObjectCtx*>(get_state()->obj_ctx);
+    // assign driver, s, and dialect_handler
     // framework promises to call op_init after parent init
-    ceph_assert(rados_ctx);
-    RGWOp::init(rados_ctx->get_store(), get_state(), this);
+    RGWOp::init(RGWHandler::driver, get_state(), this);
     op = this; // assign self as op: REQUIRED
     return 0;
   }
 
   int header_init() override {
 
-    struct req_state* state = get_state();
+    req_state* state = get_state();
     state->info.method = "DELETE";
     state->op = OP_DELETE;
 
@@ -2004,7 +1986,7 @@ public:
   buffer::list& bl; /* XXX */
   size_t bytes_written;
 
-  RGWPutObjRequest(CephContext* _cct, std::unique_ptr<rgw::sal::RGWUser> _user,
+  RGWPutObjRequest(CephContext* _cct, std::unique_ptr<rgw::sal::User> _user,
 		  const std::string& _bname, const std::string& _oname,
 		  buffer::list& _bl)
     : RGWLibRequest(_cct, std::move(_user)), bucket_name(_bname), obj_name(_oname),
@@ -2015,12 +1997,9 @@ public:
   bool only_bucket() override { return true; }
 
   int op_init() override {
-    // assign store, s, and dialect_handler
-    RGWObjectCtx* rados_ctx
-      = static_cast<RGWObjectCtx*>(get_state()->obj_ctx);
+    // assign driver, s, and dialect_handler
     // framework promises to call op_init after parent init
-    ceph_assert(rados_ctx);
-    RGWOp::init(rados_ctx->get_store(), get_state(), this);
+    RGWOp::init(RGWHandler::driver, get_state(), this);
     op = this; // assign self as op: REQUIRED
 
     int rc = valid_s3_object_name(obj_name);
@@ -2032,7 +2011,7 @@ public:
 
   int header_init() override {
 
-    struct req_state* state = get_state();
+    req_state* state = get_state();
     state->info.method = "PUT";
     state->op = OP_PUT;
 
@@ -2051,7 +2030,7 @@ public:
   }
 
   int get_params(optional_yield) override {
-    struct req_state* state = get_state();
+    req_state* state = get_state();
     RGWAccessControlPolicy_S3 s3policy(state->cct);
     /* we don't have (any) headers, so just create canned ACLs */
     int ret = s3policy.create_canned(state->owner, state->bucket_owner, state->canned_acl);
@@ -2096,7 +2075,7 @@ public:
   size_t read_resid; /* initialize to len, <= sizeof(ulp_buffer) */
   bool do_hexdump = false;
 
-  RGWReadRequest(CephContext* _cct, std::unique_ptr<rgw::sal::RGWUser> _user,
+  RGWReadRequest(CephContext* _cct, std::unique_ptr<rgw::sal::User> _user,
 		 RGWFileHandle* _rgw_fh, uint64_t off, uint64_t len,
 		 void *_ulp_buffer)
     : RGWLibRequest(_cct, std::move(_user)), rgw_fh(_rgw_fh), ulp_buffer(_ulp_buffer),
@@ -2114,19 +2093,16 @@ public:
   bool only_bucket() override { return false; }
 
   int op_init() override {
-    // assign store, s, and dialect_handler
-    RGWObjectCtx* rados_ctx
-      = static_cast<RGWObjectCtx*>(get_state()->obj_ctx);
+    // assign driver, s, and dialect_handler
     // framework promises to call op_init after parent init
-    ceph_assert(rados_ctx);
-    RGWOp::init(rados_ctx->get_store(), get_state(), this);
+    RGWOp::init(RGWHandler::driver, get_state(), this);
     op = this; // assign self as op: REQUIRED
     return 0;
   }
 
   int header_init() override {
 
-    struct req_state* state = get_state();
+    req_state* state = get_state();
     state->info.method = "GET";
     state->op = OP_GET;
 
@@ -2189,7 +2165,7 @@ public:
   const std::string& bucket_name;
   const std::string& obj_name;
 
-  RGWDeleteObjRequest(CephContext* _cct, std::unique_ptr<rgw::sal::RGWUser> _user,
+  RGWDeleteObjRequest(CephContext* _cct, std::unique_ptr<rgw::sal::User> _user,
 		      const std::string& _bname, const std::string& _oname)
     : RGWLibRequest(_cct, std::move(_user)), bucket_name(_bname), obj_name(_oname) {
     op = this;
@@ -2198,19 +2174,16 @@ public:
   bool only_bucket() override { return true; }
 
   int op_init() override {
-    // assign store, s, and dialect_handler
-    RGWObjectCtx* rados_ctx
-      = static_cast<RGWObjectCtx*>(get_state()->obj_ctx);
+    // assign driver, s, and dialect_handler
     // framework promises to call op_init after parent init
-    ceph_assert(rados_ctx);
-    RGWOp::init(rados_ctx->get_store(), get_state(), this);
+    RGWOp::init(RGWHandler::driver, get_state(), this);
     op = this; // assign self as op: REQUIRED
     return 0;
   }
 
   int header_init() override {
 
-    struct req_state* state = get_state();
+    req_state* state = get_state();
     state->info.method = "DELETE";
     state->op = OP_DELETE;
 
@@ -2240,7 +2213,7 @@ public:
 
   static constexpr uint32_t FLAG_NONE = 0x000;
 
-  RGWStatObjRequest(CephContext* _cct, std::unique_ptr<rgw::sal::RGWUser> _user,
+  RGWStatObjRequest(CephContext* _cct, std::unique_ptr<rgw::sal::User> _user,
 		    const std::string& _bname, const std::string& _oname,
 		    uint32_t _flags)
     : RGWLibRequest(_cct, std::move(_user)), bucket_name(_bname), obj_name(_oname),
@@ -2276,19 +2249,16 @@ public:
   bool only_bucket() override { return false; }
 
   int op_init() override {
-    // assign store, s, and dialect_handler
-    RGWObjectCtx* rados_ctx
-      = static_cast<RGWObjectCtx*>(get_state()->obj_ctx);
+    // assign driver, s, and dialect_handler
     // framework promises to call op_init after parent init
-    ceph_assert(rados_ctx);
-    RGWOp::init(rados_ctx->get_store(), get_state(), this);
+    RGWOp::init(RGWHandler::driver, get_state(), this);
     op = this; // assign self as op: REQUIRED
     return 0;
   }
 
   int header_init() override {
 
-    struct req_state* state = get_state();
+    req_state* state = get_state();
     state->info.method = "GET";
     state->op = OP_GET;
 
@@ -2333,7 +2303,7 @@ public:
   std::map<std::string, buffer::list> attrs;
   RGWLibFS::BucketStats& bs;
 
-  RGWStatBucketRequest(CephContext* _cct, std::unique_ptr<rgw::sal::RGWUser> _user,
+  RGWStatBucketRequest(CephContext* _cct, std::unique_ptr<rgw::sal::User> _user,
 		       const std::string& _path,
 		       RGWLibFS::BucketStats& _stats)
     : RGWLibRequest(_cct, std::move(_user)), bs(_stats) {
@@ -2353,19 +2323,16 @@ public:
   bool only_bucket() override { return false; }
 
   int op_init() override {
-    // assign store, s, and dialect_handler
-    RGWObjectCtx* rados_ctx
-      = static_cast<RGWObjectCtx*>(get_state()->obj_ctx);
+    // assign driver, s, and dialect_handler
     // framework promises to call op_init after parent init
-    ceph_assert(rados_ctx);
-    RGWOp::init(rados_ctx->get_store(), get_state(), this);
+    RGWOp::init(RGWHandler::driver, get_state(), this);
     op = this; // assign self as op: REQUIRED
     return 0;
   }
 
   int header_init() override {
 
-    struct req_state* state = get_state();
+    req_state* state = get_state();
     state->info.method = "GET";
     state->op = OP_GET;
 
@@ -2408,7 +2375,7 @@ public:
   bool is_dir;
   bool exact_matched;
 
-  RGWStatLeafRequest(CephContext* _cct, std::unique_ptr<rgw::sal::RGWUser> _user,
+  RGWStatLeafRequest(CephContext* _cct, std::unique_ptr<rgw::sal::User> _user,
 		     RGWFileHandle* _rgw_fh, const std::string& _path)
     : RGWLibRequest(_cct, std::move(_user)), rgw_fh(_rgw_fh), path(_path),
       matched(false), is_dir(false), exact_matched(false) {
@@ -2419,19 +2386,16 @@ public:
   bool only_bucket() override { return true; }
 
   int op_init() override {
-    // assign store, s, and dialect_handler
-    RGWObjectCtx* rados_ctx
-      = static_cast<RGWObjectCtx*>(get_state()->obj_ctx);
+    // assign driver, s, and dialect_handler
     // framework promises to call op_init after parent init
-    ceph_assert(rados_ctx);
-    RGWOp::init(rados_ctx->get_store(), get_state(), this);
+    RGWOp::init(RGWHandler::driver, get_state(), this);
     op = this; // assign self as op: REQUIRED
     return 0;
   }
 
   int header_init() override {
 
-    struct req_state* state = get_state();
+    req_state* state = get_state();
     state->info.method = "GET";
     state->op = OP_GET;
 
@@ -2458,7 +2422,7 @@ public:
   }
 
   void send_response() override {
-    struct req_state* state = get_state();
+    req_state* state = get_state();
     // try objects
     for (const auto& iter : objs) {
       auto& name = iter.key.name;
@@ -2513,8 +2477,8 @@ public:
   const std::string& obj_name;
   RGWFileHandle* rgw_fh;
   std::optional<rgw::BlockingAioThrottle> aio;
-  std::optional<rgw::putobj::AtomicObjectProcessor> processor;
-  rgw::putobj::DataProcessor* filter;
+  std::unique_ptr<rgw::sal::Writer> processor;
+  rgw::sal::DataProcessor* filter;
   boost::optional<RGWPutObj_Compress> compressor;
   CompressorRef plugin;
   buffer::list data;
@@ -2524,11 +2488,11 @@ public:
   size_t bytes_written;
   bool eio;
 
-  RGWWriteRequest(rgw::sal::RGWRadosStore* store,
-		  std::unique_ptr<rgw::sal::RGWUser> _user,
+  RGWWriteRequest(rgw::sal::Driver* driver, const RGWProcessEnv& penv,
+		  std::unique_ptr<rgw::sal::User> _user,
 		  RGWFileHandle* _fh, const std::string& _bname,
 		  const std::string& _oname)
-    : RGWLibContinuedReq(store->ctx(), std::move(_user)),
+    : RGWLibContinuedReq(driver->ctx(), penv, std::move(_user)),
       bucket_name(_bname), obj_name(_oname),
       rgw_fh(_fh), filter(nullptr), timer_id(0), real_ofs(0),
       bytes_written(0), eio(false) {
@@ -2544,19 +2508,16 @@ public:
   bool only_bucket() override { return true; }
 
   int op_init() override {
-    // assign store, s, and dialect_handler
-    RGWObjectCtx* rados_ctx
-      = static_cast<RGWObjectCtx*>(get_state()->obj_ctx);
+    // assign driver, s, and dialect_handler
     // framework promises to call op_init after parent init
-    ceph_assert(rados_ctx);
-    RGWOp::init(rados_ctx->get_store(), get_state(), this);
+    RGWOp::init(RGWHandler::driver, get_state(), this);
     op = this; // assign self as op: REQUIRED
     return 0;
   }
 
   int header_init() override {
 
-    struct req_state* state = get_state();
+    req_state* state = get_state();
     state->info.method = "PUT";
     state->op = OP_PUT;
 
@@ -2572,7 +2533,7 @@ public:
   }
 
   int get_params(optional_yield) override {
-    struct req_state* state = get_state();
+    req_state* state = get_state();
     RGWAccessControlPolicy_S3 s3policy(state->cct);
     /* we don't have (any) headers, so just create canned ACLs */
     int ret = s3policy.create_canned(state->owner, state->bucket_owner, state->canned_acl);
@@ -2620,7 +2581,7 @@ public:
   const std::string& src_name;
   const std::string& dst_name;
 
-  RGWCopyObjRequest(CephContext* _cct, std::unique_ptr<rgw::sal::RGWUser> _user,
+  RGWCopyObjRequest(CephContext* _cct, std::unique_ptr<rgw::sal::User> _user,
 		    RGWFileHandle* _src_parent, RGWFileHandle* _dst_parent,
 		    const std::string& _src_name, const std::string& _dst_name)
     : RGWLibRequest(_cct, std::move(_user)), src_parent(_src_parent),
@@ -2635,12 +2596,9 @@ public:
   bool only_bucket() override { return true; }
 
   int op_init() override {
-    // assign store, s, and dialect_handler
-    RGWObjectCtx* rados_ctx
-      = static_cast<RGWObjectCtx*>(get_state()->obj_ctx);
+    // assign driver, s, and dialect_handler
     // framework promises to call op_init after parent init
-    ceph_assert(rados_ctx);
-    RGWOp::init(rados_ctx->get_store(), get_state(), this);
+    RGWOp::init(RGWHandler::driver, get_state(), this);
     op = this; // assign self as op: REQUIRED
 
     return 0;
@@ -2648,21 +2606,21 @@ public:
 
   int header_init() override {
 
-    struct req_state* state = get_state();
+    req_state* state = get_state();
     state->info.method = "PUT"; // XXX check
     state->op = OP_PUT;
 
-    src_bucket_name = src_parent->bucket_name();
-    // need s->src_bucket_name?
-    dest_bucket_name = dst_parent->bucket_name();
-    // need s->bucket.name?
-    dest_obj_name = dst_parent->format_child_name(dst_name, false);
-    // need s->object_name?
+    state->src_bucket_name = src_parent->bucket_name();
+    state->bucket_name = dst_parent->bucket_name();
+
+    std::string dest_obj_name = dst_parent->format_child_name(dst_name, false);
 
     int rc = valid_s3_object_name(dest_obj_name);
     if (rc != 0)
       return rc;
 
+    state->object = RGWHandler::driver->get_object(rgw_obj_key(dest_obj_name));
+
     /* XXX and fixup key attr (could optimize w/string ref and
      * dest_obj_name) */
     buffer::list ux_key;
@@ -2682,15 +2640,15 @@ public:
   }
 
   int get_params(optional_yield) override {
-    struct req_state* s = get_state();
+    req_state* s = get_state();
     RGWAccessControlPolicy_S3 s3policy(s->cct);
     /* we don't have (any) headers, so just create canned ACLs */
     int ret = s3policy.create_canned(s->owner, s->bucket_owner, s->canned_acl);
     dest_policy = s3policy;
     /* src_object required before RGWCopyObj::verify_permissions() */
     rgw_obj_key k = rgw_obj_key(src_name);
-    src_object = rgwlib.get_store()->get_object(k);
-    s->object = src_object->clone(); // needed to avoid trap at rgw_op.cc:5150
+    s->src_object = s->bucket->get_object(k);
+    s->object = s->src_object->clone(); // needed to avoid trap at rgw_op.cc:5150
     return ret;
   }
 
@@ -2707,7 +2665,7 @@ public:
   const std::string& obj_name;
 
   RGWGetAttrsRequest(CephContext* _cct,
-		     std::unique_ptr<rgw::sal::RGWUser> _user,
+		     std::unique_ptr<rgw::sal::User> _user,
 		     const std::string& _bname, const std::string& _oname)
     : RGWLibRequest(_cct, std::move(_user)), RGWGetAttrs(),
       bucket_name(_bname), obj_name(_oname) {
@@ -2721,19 +2679,16 @@ public:
   virtual bool only_bucket() { return false; }
 
   virtual int op_init() {
-    // assign store, s, and dialect_handler
-    RGWObjectCtx* rados_ctx
-      = static_cast<RGWObjectCtx*>(get_state()->obj_ctx);
+    // assign driver, s, and dialect_handler
     // framework promises to call op_init after parent init
-    assert(rados_ctx);
-    RGWOp::init(rados_ctx->get_store(), get_state(), this);
+    RGWOp::init(RGWHandler::driver, get_state(), this);
     op = this; // assign self as op: REQUIRED
     return 0;
   }
 
   virtual int header_init() {
 
-    struct req_state* s = get_state();
+    req_state* s = get_state();
     s->info.method = "GET";
     s->op = OP_GET;
 
@@ -2762,7 +2717,7 @@ public:
   const std::string& bucket_name;
   const std::string& obj_name;
 
-  RGWSetAttrsRequest(CephContext* _cct, std::unique_ptr<rgw::sal::RGWUser> _user,
+  RGWSetAttrsRequest(CephContext* _cct, std::unique_ptr<rgw::sal::User> _user,
 		     const std::string& _bname, const std::string& _oname)
     : RGWLibRequest(_cct, std::move(_user)), bucket_name(_bname), obj_name(_oname) {
     op = this;
@@ -2775,19 +2730,16 @@ public:
   bool only_bucket() override { return false; }
 
   int op_init() override {
-    // assign store, s, and dialect_handler
-    RGWObjectCtx* rados_ctx
-      = static_cast<RGWObjectCtx*>(get_state()->obj_ctx);
+    // assign driver, s, and dialect_handler
     // framework promises to call op_init after parent init
-    ceph_assert(rados_ctx);
-    RGWOp::init(rados_ctx->get_store(), get_state(), this);
+    RGWOp::init(RGWHandler::driver, get_state(), this);
     op = this; // assign self as op: REQUIRED
     return 0;
   }
 
   int header_init() override {
 
-    struct req_state* state = get_state();
+    req_state* state = get_state();
     state->info.method = "PUT";
     state->op = OP_PUT;
 
@@ -2818,33 +2770,30 @@ public:
   const std::string& obj_name;
 
   RGWRMAttrsRequest(CephContext* _cct,
-		     std::unique_ptr<rgw::sal::RGWUser> _user,
+		     std::unique_ptr<rgw::sal::User> _user,
 		     const std::string& _bname, const std::string& _oname)
     : RGWLibRequest(_cct, std::move(_user)), RGWRMAttrs(),
       bucket_name(_bname), obj_name(_oname) {
     op = this;
   }
 
-  const rgw::sal::RGWAttrs& get_attrs() {
+  const rgw::sal::Attrs& get_attrs() {
     return attrs;
   }
 
   virtual bool only_bucket() { return false; }
 
   virtual int op_init() {
-    // assign store, s, and dialect_handler
-    RGWObjectCtx* rados_ctx
-      = static_cast<RGWObjectCtx*>(get_state()->obj_ctx);
+    // assign driver, s, and dialect_handler
     // framework promises to call op_init after parent init
-    assert(rados_ctx);
-    RGWOp::init(rados_ctx->get_store(), get_state(), this);
+    RGWOp::init(RGWHandler::driver, get_state(), this);
     op = this; // assign self as op: REQUIRED
     return 0;
   }
 
   virtual int header_init() {
 
-    struct req_state* s = get_state();
+    req_state* s = get_state();
     s->info.method = "DELETE";
     s->op = OP_PUT;
 
@@ -2873,25 +2822,22 @@ class RGWGetClusterStatReq : public RGWLibRequest,
         public RGWGetClusterStat {
 public:
   struct rados_cluster_stat_t& stats_req;
-  RGWGetClusterStatReq(CephContext* _cct, std::unique_ptr<rgw::sal::RGWUser> _user,
+  RGWGetClusterStatReq(CephContext* _cct, std::unique_ptr<rgw::sal::User> _user,
                        rados_cluster_stat_t& _stats):
   RGWLibRequest(_cct, std::move(_user)), stats_req(_stats){
     op = this;
   }
 
   int op_init() override {
-    // assign store, s, and dialect_handler
-    RGWObjectCtx* rados_ctx
-      = static_cast<RGWObjectCtx*>(get_state()->obj_ctx);
+    // assign driver, s, and dialect_handler
     // framework promises to call op_init after parent init
-    ceph_assert(rados_ctx);
-    RGWOp::init(rados_ctx->get_store(), get_state(), this);
+    RGWOp::init(RGWHandler::driver, get_state(), this);
     op = this; // assign self as op: REQUIRED
     return 0;
   }
 
   int header_init() override {
-    struct req_state* state = get_state();
+    req_state* state = get_state();
     state->info.method = "GET";
     state->op = OP_GET;
     return 0;
@@ -2909,5 +2855,3 @@ public:
 
 
 } /* namespace rgw */
-
-#endif /* RGW_FILE_H */
