@@ -25,8 +25,8 @@ public:
   RGWRESTMgr_Usage() = default;
   ~RGWRESTMgr_Usage() override = default;
 
-  RGWHandler_REST* get_handler(rgw::sal::RGWRadosStore* store,
-			       struct req_state*,
+  RGWHandler_REST* get_handler(rgw::sal::Driver* driver,
+			       req_state*,
                                const rgw::auth::StrategyRegistry& auth_registry,
                                const std::string&) override {
     return new RGWHandler_Usage(auth_registry);
