@@ -1,8 +1,7 @@
 // -*- mode:C++; tab-width:8; c-basic-offset:2; indent-tabs-mode:t -*-
 // vim: ts=8 sw=2 smarttab ft=cpp
 
-#ifndef RGW_TAG_S3_H
-#define RGW_TAG_S3_H
+#pragma once
 
 #include <map>
 #include <string>
@@ -48,6 +47,3 @@ public:
     return tagset.rebuild(dest);
   }
 };
-
-
-#endif /* RGW_TAG_S3_H */
