@@ -5,10 +5,12 @@
 #include "svc_rados.h"
 #include "svc_zone.h"
 
-#include "rgw/rgw_tools.h"
+#include "rgw_tools.h"
 
 #define dout_subsys ceph_subsys_rgw
 
+using namespace std;
+
 int RGWSI_SysObj_Core_GetObjState::get_rados_obj(const DoutPrefixProvider *dpp,
                                                  RGWSI_RADOS *rados_svc,
                                                  RGWSI_Zone *zone_svc,
@@ -51,68 +53,9 @@ int RGWSI_SysObj_Core::get_rados_obj(const DoutPrefixProvider *dpp,
   return 0;
 }
 
-int RGWSI_SysObj_Core::get_system_obj_state_impl(RGWSysObjectCtxBase *rctx,
-                                                 const rgw_raw_obj& obj,
-                                                 RGWSysObjState **state,
-                                                 RGWObjVersionTracker *objv_tracker,
-                                                 optional_yield y,
-                                                 const DoutPrefixProvider *dpp)
-{
-  if (obj.empty()) {
-    return -EINVAL;
-  }
-
-  RGWSysObjState *s = rctx->get_state(obj);
-  ldpp_dout(dpp, 20) << "get_system_obj_state: rctx=" << (void *)rctx << " obj=" << obj << " state=" << (void *)s << " s->prefetch_data=" << s->prefetch_data << dendl;
-  *state = s;
-  if (s->has_attrs) {
-    return 0;
-  }
-
-  s->obj = obj;
-
-  int r = raw_stat(dpp, obj, &s->size, &s->mtime, &s->epoch, &s->attrset,
-                   (s->prefetch_data ? &s->data : nullptr), objv_tracker, y);
-  if (r == -ENOENT) {
-    s->exists = false;
-    s->has_attrs = true;
-    s->mtime = real_time();
-    return 0;
-  }
-  if (r < 0)
-    return r;
-
-  s->exists = true;
-  s->has_attrs = true;
-  s->obj_tag = s->attrset[RGW_ATTR_ID_TAG];
-
-  if (s->obj_tag.length()) {
-    ldpp_dout(dpp, 20) << "get_system_obj_state: setting s->obj_tag to " << s->obj_tag.c_str() << dendl;
-  } else {
-    ldpp_dout(dpp, 20) << "get_system_obj_state: s->obj_tag was set empty" << dendl;
-  }
-
-  return 0;
-}
-
-int RGWSI_SysObj_Core::get_system_obj_state(RGWSysObjectCtxBase *rctx,
-                                            const rgw_raw_obj& obj,
-                                            RGWSysObjState **state,
-                                            RGWObjVersionTracker *objv_tracker,
-                                            optional_yield y,
-                                            const DoutPrefixProvider *dpp)
-{
-  int ret;
-
-  do {
-    ret = get_system_obj_state_impl(rctx, obj, state, objv_tracker, y, dpp);
-  } while (ret == -EAGAIN);
-
-  return ret;
-}
-
-int RGWSI_SysObj_Core::raw_stat(const DoutPrefixProvider *dpp, const rgw_raw_obj& obj, uint64_t *psize, real_time *pmtime, uint64_t *epoch,
-                                map<string, bufferlist> *attrs, bufferlist *first_chunk,
+int RGWSI_SysObj_Core::raw_stat(const DoutPrefixProvider *dpp, const rgw_raw_obj& obj,
+                                uint64_t *psize, real_time *pmtime,
+                                map<string, bufferlist> *attrs,
                                 RGWObjVersionTracker *objv_tracker,
                                 optional_yield y)
 {
@@ -133,16 +76,8 @@ int RGWSI_SysObj_Core::raw_stat(const DoutPrefixProvider *dpp, const rgw_raw_obj
   if (psize || pmtime) {
     op.stat2(&size, &mtime_ts, nullptr);
   }
-  if (first_chunk) {
-    op.read(0, cct->_conf->rgw_max_chunk_size, first_chunk, nullptr);
-  }
   bufferlist outbl;
   r = rados_obj.operate(dpp, &op, &outbl, y);
-
-  if (epoch) {
-    *epoch = rados_obj.get_last_version();
-  }
-
   if (r < 0)
     return r;
 
@@ -154,8 +89,7 @@ int RGWSI_SysObj_Core::raw_stat(const DoutPrefixProvider *dpp, const rgw_raw_obj
   return 0;
 }
 
-int RGWSI_SysObj_Core::stat(RGWSysObjectCtxBase& obj_ctx,
-                            RGWSI_SysObj_Obj_GetObjState& _state,
+int RGWSI_SysObj_Core::stat(RGWSI_SysObj_Obj_GetObjState& _state,
                             const rgw_raw_obj& obj,
                             map<string, bufferlist> *attrs,
 			    bool raw_attrs,
@@ -165,21 +99,19 @@ int RGWSI_SysObj_Core::stat(RGWSysObjectCtxBase& obj_ctx,
                             optional_yield y,
                             const DoutPrefixProvider *dpp)
 {
-  RGWSysObjState *astate = nullptr;
+  uint64_t size = 0;
+  ceph::real_time mtime;
+  std::map<std::string, bufferlist> attrset;
 
-  int r = get_system_obj_state(&obj_ctx, obj, &astate, objv_tracker, y, dpp);
+  int r = raw_stat(dpp, obj, &size, &mtime, &attrset, objv_tracker, y);
   if (r < 0)
     return r;
 
-  if (!astate->exists) {
-    return -ENOENT;
-  }
-
   if (attrs) {
     if (raw_attrs) {
-      *attrs = astate->attrset;
+      *attrs = std::move(attrset);
     } else {
-      rgw_filter_attrset(astate->attrset, RGW_ATTR_PREFIX, attrs);
+      rgw_filter_attrset(attrset, RGW_ATTR_PREFIX, attrs);
     }
     if (cct->_conf->subsys.should_gather<ceph_subsys_rgw, 20>()) {
       map<string, bufferlist>::iterator iter;
@@ -190,19 +122,19 @@ int RGWSI_SysObj_Core::stat(RGWSysObjectCtxBase& obj_ctx,
   }
 
   if (obj_size)
-    *obj_size = astate->size;
+    *obj_size = size;
   if (lastmod)
-    *lastmod = astate->mtime;
+    *lastmod = mtime;
 
   return 0;
 }
 
 int RGWSI_SysObj_Core::read(const DoutPrefixProvider *dpp,
-                            RGWSysObjectCtxBase& obj_ctx,
                             RGWSI_SysObj_Obj_GetObjState& _read_state,
                             RGWObjVersionTracker *objv_tracker,
                             const rgw_raw_obj& obj,
                             bufferlist *bl, off_t ofs, off_t end,
+                            ceph::real_time* pmtime, uint64_t* psize,
                             map<string, bufferlist> *attrs,
 			    bool raw_attrs,
                             rgw_cache_entry_info *cache_info,
@@ -212,6 +144,7 @@ int RGWSI_SysObj_Core::read(const DoutPrefixProvider *dpp,
   auto& read_state = static_cast<GetObjState&>(_read_state);
 
   uint64_t len;
+  struct timespec mtime_ts;
   librados::ObjectReadOperation op;
 
   if (end < 0)
@@ -222,6 +155,9 @@ int RGWSI_SysObj_Core::read(const DoutPrefixProvider *dpp,
   if (objv_tracker) {
     objv_tracker->prepare_op_for_read(&op);
   }
+  if (psize || pmtime) {
+    op.stat2(psize, &mtime_ts, nullptr);
+  }
 
   ldpp_dout(dpp, 20) << "rados->read ofs=" << ofs << " len=" << len << dendl;
   op.read(ofs, len, bl, nullptr);
@@ -242,21 +178,25 @@ int RGWSI_SysObj_Core::read(const DoutPrefixProvider *dpp,
     ldpp_dout(dpp, 20) << "get_rados_obj() on obj=" << obj << " returned " << r << dendl;
     return r;
   }
-  r = rados_obj.operate(dpp, &op, nullptr, y);
+
+  auto& ref = rados_obj.get_ref();
+  version_t op_ver = 0;
+  r = rgw_rados_operate(dpp, ref.pool.ioctx(), obj.oid, &op, nullptr, y, 0, &op_ver);
   if (r < 0) {
     ldpp_dout(dpp, 20) << "rados_obj.operate() r=" << r << " bl.length=" << bl->length() << dendl;
     return r;
   }
   ldpp_dout(dpp, 20) << "rados_obj.operate() r=" << r << " bl.length=" << bl->length() << dendl;
 
-  uint64_t op_ver = rados_obj.get_last_version();
-
   if (read_state.last_ver > 0 &&
       read_state.last_ver != op_ver) {
     ldpp_dout(dpp, 5) << "raced with an object write, abort" << dendl;
     return -ECANCELED;
   }
 
+  if (pmtime) {
+    *pmtime = ceph::real_clock::from_timespec(mtime_ts);
+  }
   if (attrs && !raw_attrs) {
     rgw_filter_attrset(unfiltered_attrset, RGW_ATTR_PREFIX, attrs);
   }
@@ -303,7 +243,7 @@ int RGWSI_SysObj_Core::set_attrs(const DoutPrefixProvider *dpp,
                                  map<string, bufferlist>& attrs,
                                  map<string, bufferlist> *rmattrs,
                                  RGWObjVersionTracker *objv_tracker,
-                                 optional_yield y)
+                                 bool exclusive, optional_yield y)
 {
   RGWSI_RADOS::Obj rados_obj;
   int r = get_rados_obj(dpp, zone_svc, obj, &rados_obj);
@@ -314,6 +254,9 @@ int RGWSI_SysObj_Core::set_attrs(const DoutPrefixProvider *dpp,
 
   librados::ObjectWriteOperation op;
 
+  if (exclusive) {
+    op.create(true); // exclusive create
+  }
   if (objv_tracker) {
     objv_tracker->prepare_op_for_write(&op);
   }
@@ -510,7 +453,6 @@ int RGWSI_SysObj_Core::notify(const DoutPrefixProvider *dpp, const rgw_raw_obj&
 }
 
 int RGWSI_SysObj_Core::remove(const DoutPrefixProvider *dpp, 
-                              RGWSysObjectCtxBase& obj_ctx,
                               RGWObjVersionTracker *objv_tracker,
                               const rgw_raw_obj& obj,
                               optional_yield y)
@@ -658,7 +600,7 @@ int RGWSI_SysObj_Core::pool_list_prefixed_objs(const DoutPrefixProvider *dpp,
   do {
     vector<string> oids;
 #define MAX_OBJS_DEFAULT 1000
-    int r = op.get_next(MAX_OBJS_DEFAULT, &oids, &is_truncated);
+    int r = op.get_next(dpp, MAX_OBJS_DEFAULT, &oids, &is_truncated);
     if (r < 0) {
       return r;
     }
@@ -693,7 +635,8 @@ int RGWSI_SysObj_Core::pool_list_objects_init(const DoutPrefixProvider *dpp,
   return 0;
 }
 
-int RGWSI_SysObj_Core::pool_list_objects_next(RGWSI_SysObj::Pool::ListCtx& _ctx,
+int RGWSI_SysObj_Core::pool_list_objects_next(const DoutPrefixProvider *dpp,
+                                              RGWSI_SysObj::Pool::ListCtx& _ctx,
                                               int max,
                                               vector<string> *oids,
                                               bool *is_truncated)
@@ -702,10 +645,10 @@ int RGWSI_SysObj_Core::pool_list_objects_next(RGWSI_SysObj::Pool::ListCtx& _ctx,
     return -EINVAL;
   }
   auto& ctx = static_cast<PoolListImplInfo&>(*_ctx.impl);
-  int r = ctx.op.get_next(max, oids, is_truncated);
+  int r = ctx.op.get_next(dpp, max, oids, is_truncated);
   if (r < 0) {
     if(r != -ENOENT)
-      ldout(cct, 10) << "failed to list objects pool_iterate returned r=" << r << dendl;
+      ldpp_dout(dpp, 10) << "failed to list objects pool_iterate returned r=" << r << dendl;
     return r;
   }
 
