@@ -89,6 +89,9 @@ public:
     virtual int get_sockopt(file_desc& _fd, int level, int optname, void* data, size_t len) const {
         return _fd.getsockopt(level, optname, reinterpret_cast<char*>(data), socklen_t(len));
     }
+    virtual socket_address local_address(file_desc& _fd) const {
+        return _fd.get_address();
+    }
 };
 
 thread_local posix_ap_server_socket_impl::sockets_map_t posix_ap_server_socket_impl::sockets{};
@@ -197,6 +200,18 @@ get_posix_connected_socket_ops(sa_family_t family, int protocol) {
     }
 }
 
+static void shutdown_socket_fd(pollable_fd& fd, int how) noexcept {
+    try {
+        // file_desc::shutdown ignores ENOTCONN. Other reasons for exception
+        // EINVAL (wrong "how") -- impossible
+        // ENOTSOCK (not a socket) -- incredible
+        // EBADF (invalid file descriptor) -- irretrievable
+        fd.shutdown(how);
+    } catch (...) {
+        on_internal_error(seastar_logger, format("socket shutdown({}, {}) failed: {}", fd.get_file_desc().fdinfo(), how, std::current_exception()));
+    }
+}
+
 class posix_connected_socket_impl final : public connected_socket_impl {
     pollable_fd _fd;
     const posix_connected_socket_operations* _ops;
@@ -219,10 +234,10 @@ public:
         return data_sink(std::make_unique< posix_data_sink_impl>(_fd));
     }
     virtual void shutdown_input() override {
-        _fd.shutdown(SHUT_RD);
+        shutdown_socket_fd(_fd, SHUT_RD);
     }
     virtual void shutdown_output() override {
-        _fd.shutdown(SHUT_WR);
+        shutdown_socket_fd(_fd, SHUT_WR);
     }
     virtual void set_nodelay(bool nodelay) override {
         return _ops->set_nodelay(_fd.get_file_desc(), nodelay);
@@ -248,6 +263,13 @@ public:
     int get_sockopt(int level, int optname, void* data, size_t len) const override {
         return _ops->get_sockopt(_fd.get_file_desc(), level, optname, data, len);
     }
+    socket_address local_address() const noexcept override {
+        return _ops->local_address(_fd.get_file_desc());
+    }
+    future<> wait_input_shutdown() override {
+        return _fd.poll_rdhup();
+    }
+
     friend class posix_server_socket_impl;
     friend class posix_ap_server_socket_impl;
     friend class posix_reuseport_server_socket_impl;
@@ -490,7 +512,7 @@ posix_server_socket_impl::accept() {
 
 void
 posix_server_socket_impl::abort_accept() {
-    _lfd.abort_reader();
+    _lfd.shutdown(SHUT_RD, pollable_fd::shutdown_kernel_only::no);
 }
 
 socket_address posix_server_socket_impl::local_address() const {
@@ -546,7 +568,7 @@ posix_reuseport_server_socket_impl::accept() {
 
 void
 posix_reuseport_server_socket_impl::abort_accept() {
-    _lfd.abort_reader();
+    _lfd.shutdown(SHUT_RD, pollable_fd::shutdown_kernel_only::no);
 }
 
 socket_address posix_reuseport_server_socket_impl::local_address() const {
@@ -572,7 +594,7 @@ posix_ap_server_socket_impl::move_connected_socket(int protocol, socket_address
 
 future<temporary_buffer<char>>
 posix_data_source_impl::get() {
-    return _fd.read_some(static_cast<internal::buffer_allocator*>(this)).then([this] (temporary_buffer<char> b) {
+    return _fd.recv_some(static_cast<internal::buffer_allocator*>(this)).then([this] (temporary_buffer<char> b) {
         if (b.size() >= _config.buffer_size) {
             _config.buffer_size *= 2;
             _config.buffer_size = std::min(_config.buffer_size, _config.max_buffer_size);
@@ -629,7 +651,7 @@ posix_data_sink_impl::close() {
     return make_ready_future<>();
 }
 
-posix_network_stack::posix_network_stack(boost::program_options::variables_map opts, std::pmr::polymorphic_allocator<char>* allocator)
+posix_network_stack::posix_network_stack(const program_options::option_group& opts, std::pmr::polymorphic_allocator<char>* allocator)
         : _reuseport(engine().posix_reuseport_available()), _allocator(allocator) {
 }
 
@@ -654,8 +676,8 @@ posix_network_stack::listen(socket_address sa, listen_options opt) {
     return ::seastar::socket(std::make_unique<posix_socket_impl>(_allocator));
 }
 
-posix_ap_network_stack::posix_ap_network_stack(boost::program_options::variables_map opts, std::pmr::polymorphic_allocator<char>* allocator)
-        : posix_network_stack(std::move(opts), allocator), _reuseport(engine().posix_reuseport_available()) {
+posix_ap_network_stack::posix_ap_network_stack(const program_options::option_group& opts, std::pmr::polymorphic_allocator<char>* allocator)
+        : posix_network_stack(opts, allocator), _reuseport(engine().posix_reuseport_available()) {
 }
 
 server_socket
@@ -755,10 +777,10 @@ public:
     virtual future<> send(const socket_address& dst, const char *msg) override;
     virtual future<> send(const socket_address& dst, packet p) override;
     virtual void shutdown_input() override {
-        _fd.abort_reader();
+        _fd.shutdown(SHUT_RD, pollable_fd::shutdown_kernel_only::no);
     }
     virtual void shutdown_output() override {
-        _fd.abort_writer();
+        _fd.shutdown(SHUT_WR, pollable_fd::shutdown_kernel_only::no);
     }
     virtual void close() override {
         _closed = true;
@@ -840,13 +862,14 @@ posix_udp_channel::receive() {
     });
 }
 
-void register_posix_stack() {
-    register_network_stack("posix", boost::program_options::options_description(),
-        [](boost::program_options::variables_map ops) {
+network_stack_entry register_posix_stack() {
+    return network_stack_entry{
+        "posix", std::make_unique<program_options::option_group>(nullptr, "Posix"),
+        [](const program_options::option_group& ops) {
             return smp::main_thread() ? posix_network_stack::create(ops)
                                       : posix_ap_network_stack::create(ops);
         },
-        true);
+        true};
 }
 
 // nw interface stuff
