@@ -19,9 +19,11 @@
  * Copyright (C) 2015 Cloudius Systems, Ltd.
  */
 
+#include <unistd.h>
 #include <fmt/core.h>
 #if FMT_VERSION >= 60000
 #include <fmt/chrono.h>
+#include <fmt/color.h>
 #elif FMT_VERSION >= 50000
 #include <fmt/time.h>
 #endif
@@ -45,11 +47,59 @@
 #include <map>
 #include <regex>
 #include <string>
+#include <string_view>
 #include <system_error>
 #include <chrono>
+#include <algorithm>
+
+#include "core/program_options.hh"
 
 using namespace std::chrono_literals;
 
+struct wrapped_log_level {
+    seastar::log_level level;
+};
+
+namespace fmt {
+template <> struct formatter<wrapped_log_level> {
+    using log_level = seastar::log_level;
+    static constexpr size_t nr_levels = static_cast<size_t>(log_level::trace) + 1;
+    static bool colored;
+
+    // format specifier not supported
+    template <typename ParseContext>
+    constexpr auto parse(ParseContext &ctx) { return ctx.begin(); }
+
+    template <typename FormatContext>
+    auto format(wrapped_log_level wll, FormatContext& ctx) const {
+        static seastar::array_map<seastar::sstring, nr_levels> text = {
+            { int(log_level::debug), "DEBUG" },
+            { int(log_level::info),  "INFO " },
+            { int(log_level::trace), "TRACE" },
+            { int(log_level::warn),  "WARN " },
+            { int(log_level::error), "ERROR" },
+        };
+        int index = static_cast<int>(wll.level);
+        std::string_view name = text[index];
+#if FMT_VERSION >= 60000
+        static seastar::array_map<text_style, nr_levels> style = {
+            { int(log_level::debug), fg(terminal_color::green)  },
+            { int(log_level::info),  fg(terminal_color::white)  },
+            { int(log_level::trace), fg(terminal_color::blue)   },
+            { int(log_level::warn),  fg(terminal_color::yellow) },
+            { int(log_level::error), fg(terminal_color::red)    },
+        };
+        if (colored) {
+            return fmt::format_to(ctx.out(), "{}",
+                fmt::format(style[index], "{}", name));
+        }
+#endif
+        return fmt::format_to(ctx.out(), "{}", name);
+    }
+};
+bool formatter<wrapped_log_level>::colored = true;
+}
+
 namespace seastar {
 
 namespace internal {
@@ -60,7 +110,13 @@ void log_buf::free_buffer() noexcept {
     }
 }
 
-void log_buf::realloc_buffer() {
+void log_buf::realloc_buffer_and_append(char c) noexcept {
+  if (_alloc_failure) {
+    // Already failed to reallocate once, don't try again
+    return;
+  }
+
+  try {
     const auto old_size = size();
     const auto new_size = old_size * 2;
 
@@ -72,6 +128,13 @@ void log_buf::realloc_buffer() {
     _current = _begin + old_size;
     _end = _begin + new_size;
     _own_buf = true;
+    *_current++ = c;
+  } catch (...) {
+    _alloc_failure = true;
+    std::string_view msg = "(log buffer allocation failure)";
+    auto can_copy = std::min(msg.size(), size_t(_current - _begin));
+    std::memcpy(_current - can_copy, msg.begin(), can_copy);
+  }
 }
 
 log_buf::log_buf()
@@ -168,18 +231,20 @@ static internal::log_buf::inserter_iterator print_boot_timestamp(internal::log_b
 static internal::log_buf::inserter_iterator print_real_timestamp(internal::log_buf::inserter_iterator it) {
     struct a_second {
         time_t t;
-        std::string s;
+        std::array<char, 32> static_buf; // big enough to hold '2023-01-14 15:06:33'
+        internal::log_buf buf{static_buf.data(), static_buf.size()};
     };
     static thread_local a_second this_second;
     using clock = std::chrono::system_clock;
     auto n = clock::now();
     auto t = clock::to_time_t(n);
     if (this_second.t != t) {
-        this_second.s = fmt::format("{:%Y-%m-%d %T}", fmt::localtime(t));
         this_second.t = t;
+        this_second.buf.clear();
+        fmt::format_to(this_second.buf.back_insert_begin(), "{:%Y-%m-%d %T}", fmt::localtime(t));
     }
     auto ms = (n - clock::from_time_t(t)) / 1ms;
-    return fmt::format_to(it, "{},{:03d}", this_second.s, ms);
+    return fmt::format_to(it, "{},{:03d}", this_second.buf.view(), ms);
 }
 
 static internal::log_buf::inserter_iterator (*print_timestamp)(internal::log_buf::inserter_iterator) = print_no_timestamp;
@@ -215,6 +280,7 @@ std::istream& operator>>(std::istream& in, log_level& level) {
 std::ostream* logger::_out = &std::cerr;
 std::atomic<bool> logger::_ostream = { true };
 std::atomic<bool> logger::_syslog = { false };
+unsigned logger::_shard_field_width = 1;
 
 logger::logger(sstring name) : _name(std::move(name)) {
     global_logger_registry().register_logger(this);
@@ -251,28 +317,27 @@ logger::do_log(log_level level, log_writer& writer) {
     if(!is_ostream_enabled && !is_syslog_enabled) {
       return;
     }
-    static array_map<sstring, 20> level_map = {
-            { int(log_level::debug), "DEBUG" },
-            { int(log_level::info),  "INFO "  },
-            { int(log_level::trace), "TRACE" },
-            { int(log_level::warn),  "WARN "  },
-            { int(log_level::error), "ERROR" },
-    };
     auto print_once = [&] (internal::log_buf::inserter_iterator it) {
       if (local_engine) {
-          it = fmt::format_to(it, " [shard {}]", this_shard_id());
+          it = fmt::format_to(it, " [shard {:{}}]", this_shard_id(), _shard_field_width);
       }
       it = fmt::format_to(it, " {} - ", _name);
       return writer(it);
     };
 
+    // Mainly this protects us from re-entrance via malloc()'s
+    // oversized allocation warnings and failed allocation errors
+    silencer be_silent;
+
     if (is_ostream_enabled) {
         internal::log_buf buf(static_log_buf.data(), static_log_buf.size());
         auto it = buf.back_insert_begin();
-        it = fmt::format_to(it, "{} ", level_map[int(level)]);
+        it = fmt::format_to(it, "{} ", wrapped_log_level{level});
         it = print_timestamp(it);
         it = print_once(it);
-        *_out << buf.view() << std::endl;
+        *it++ = '\n';
+        *_out << buf.view();
+        _out->flush();
     }
     if (is_syslog_enabled) {
         internal::log_buf buf(static_log_buf.data(), static_log_buf.size());
@@ -296,11 +361,15 @@ logger::do_log(log_level level, log_writer& writer) {
     }
 }
 
-void logger::failed_to_log(std::exception_ptr ex) noexcept
+void logger::failed_to_log(std::exception_ptr ex, format_info fmt) noexcept
 {
     try {
-        lambda_log_writer writer([ex = std::move(ex)] (internal::log_buf::inserter_iterator it) {
-            return fmt::format_to(it, "failed to log message: {}", ex);
+        lambda_log_writer writer([ex = std::move(ex), fmt = std::move(fmt)] (internal::log_buf::inserter_iterator it) {
+            it = fmt::format_to(it, "{}:{} @{}: failed to log message", fmt.loc.file_name(), fmt.loc.line(), fmt.loc.function_name());
+            if (!fmt.format.empty()) {
+                it = fmt::format_to(it, ": fmt='{}'", fmt.format);
+            }
+            return fmt::format_to(it, ": {}", ex);
         });
         do_log(log_level::error, writer);
     } catch (...) {
@@ -328,6 +397,16 @@ logger::set_syslog_enabled(bool enabled) noexcept {
     _syslog.store(enabled, std::memory_order_relaxed);
 }
 
+void
+logger::set_shard_field_width(unsigned width) noexcept {
+    _shard_field_width = width;
+}
+
+void
+logger::set_with_color(bool enabled) noexcept {
+    fmt::formatter<wrapped_log_level>::colored = enabled;
+}
+
 bool logger::is_shard_zero() noexcept {
     return this_shard_id() == 0;
 }
@@ -408,6 +487,7 @@ void apply_logging_settings(const logging_settings& s) {
         break;
     }
     logger::set_syslog_enabled(s.syslog_enabled);
+    logger::set_with_color(s.with_color);
 
     switch (s.stdout_timestamp_style) {
     case logger_timestamp_style::none:
@@ -452,29 +532,52 @@ log_level parse_log_level(const sstring& s) {
     }
 }
 
-bpo::options_description get_options_description() {
-    bpo::options_description opts("Logging options");
+void parse_map_associations(const std::string& v, std::function<void(std::string, std::string)> consume_key_value) {
+    static const std::regex colon(":");
 
-    opts.add_options()
-            ("default-log-level",
-             bpo::value<sstring>()->default_value("info"),
+    std::sregex_token_iterator s(v.begin(), v.end(), colon, -1);
+    const std::sregex_token_iterator e;
+    while (s != e) {
+        const sstring p = std::string(*s++);
+
+        const auto i = p.find('=');
+        if (i == sstring::npos) {
+            throw bpo::invalid_option_value(p);
+        }
+
+        auto k = p.substr(0, i);
+        auto v = p.substr(i + 1, p.size());
+        consume_key_value(std::move(k), std::move(v));
+    };
+}
+
+bpo::options_description get_options_description() {
+    program_options::options_description_building_visitor descriptor;
+    options(nullptr).describe(descriptor);
+    return std::move(descriptor).get_options_description();
+}
+
+options::options(program_options::option_group* parent_group)
+    : program_options::option_group(parent_group, "Logging options")
+    , default_log_level(*this, "default-log-level",
+             log_level::info,
              "Default log level for log messages. Valid values are trace, debug, info, warn, error."
-            )
-            ("logger-log-level",
-             bpo::value<program_options::string_map>()->default_value({}),
+             )
+    , logger_log_level(*this, "logger-log-level",
+             {{}},
              "Map of logger name to log level. The format is \"NAME0=LEVEL0[:NAME1=LEVEL1:...]\". "
              "Valid logger names can be queried with --help-loggers. "
              "Valid values for levels are trace, debug, info, warn, error. "
              "This option can be specified multiple times."
             )
-            ("logger-stdout-timestamps", bpo::value<logger_timestamp_style>()->default_value(logger_timestamp_style::real),
+    , logger_stdout_timestamps(*this, "logger-stdout-timestamps", logger_timestamp_style::real,
                     "Select timestamp style for stdout logs: none|boot|real")
-            ("log-to-stdout", bpo::value<bool>()->default_value(true), "Send log output to output stream, as selected by --logger-ostream-type")
-            ("logger-ostream-type", bpo::value<logger_ostream_type>()->default_value(logger_ostream_type::stderr), "Send log output to: none|stdout|stderr")
-            ("log-to-syslog", bpo::value<bool>()->default_value(false), "Send log output to syslog.")
-            ("help-loggers", bpo::bool_switch(), "Print a list of logger names and exit.");
-
-    return opts;
+    , log_to_stdout(*this, "log-to-stdout", true, "Send log output to output stream, as selected by --logger-ostream-type")
+    , logger_ostream_type(*this, "logger-ostream-type", logger_ostream_type::stderr,
+            "Send log output to: none|stdout|stderr")
+    , log_to_syslog(*this, "log-to-syslog", false, "Send log output to syslog.")
+    , log_with_color(*this, "log-with-color", isatty(STDOUT_FILENO), "Print colored tag prefix in log message written to ostream")
+{
 }
 
 void print_available_loggers(std::ostream& os) {
@@ -490,18 +593,21 @@ void print_available_loggers(std::ostream& os) {
 }
 
 logging_settings extract_settings(const boost::program_options::variables_map& vars) {
-    const auto& raw_levels = vars["logger-log-level"].as<program_options::string_map>();
-
-    std::unordered_map<sstring, log_level> levels;
-    parse_logger_levels(raw_levels, std::inserter(levels, levels.begin()));
+    options opts(nullptr);
+    program_options::variables_map_extracting_visitor visitor(vars);
+    opts.mutate(visitor);
+    return extract_settings(opts);
+}
 
+logging_settings extract_settings(const options& opts) {
     return logging_settings{
-        std::move(levels),
-        parse_log_level(vars["default-log-level"].as<sstring>()),
-        vars["log-to-stdout"].as<bool>(),
-        vars["log-to-syslog"].as<bool>(),
-        vars["logger-stdout-timestamps"].as<logger_timestamp_style>(),
-        vars["logger-ostream-type"].as<logger_ostream_type>(),
+        opts.logger_log_level.get_value(),
+        opts.default_log_level.get_value(),
+        opts.log_to_stdout.get_value(),
+        opts.log_to_syslog.get_value(),
+        opts.log_with_color.get_value(),
+        opts.logger_stdout_timestamps.get_value(),
+        opts.logger_ostream_type.get_value(),
     };
 }
 
@@ -542,18 +648,21 @@ std::ostream& operator<<(std::ostream& out, const std::exception_ptr& eptr) {
             throw;
         } catch (const seastar::nested_exception& ne) {
             out << fmt::format(": {} (while cleaning up after {})", ne.inner, ne.outer);
-        } catch(const std::system_error &e) {
+        } catch (const std::system_error& e) {
             out << " (error " << e.code() << ", " << e.what() << ")";
-        } catch(const std::exception& e) {
+        } catch (const std::exception& e) {
             out << " (" << e.what() << ")";
-            try {
-                std::rethrow_if_nested(e);
-            } catch (...) {
-                out << ": " << std::current_exception();
-            }
-        } catch(...) {
+        } catch (...) {
             // no extra info
         }
+
+        try {
+            throw;
+        } catch (const std::nested_exception& ne) {
+            out << ": " << ne.nested_ptr();
+        } catch (...) {
+            // do nothing
+        }
     }
     return out;
 }
