@@ -32,11 +32,11 @@ extern "C"{
 #include "common/ceph_argparse.h"
 #include "common/Finisher.h"
 #include "global/global_init.h"
-#include "rgw/rgw_common.h"
-#include "rgw/rgw_datalog.h"
-#include "rgw/rgw_mdlog.h"
-#include "rgw/rgw_bucket.h"
-#include "rgw/rgw_rados.h"
+#include "rgw_common.h"
+#include "rgw_datalog.h"
+#include "rgw_mdlog.h"
+#include "rgw_bucket.h"
+#include "rgw_rados.h"
 #include "include/utime.h"
 #include "include/object.h"
 #include <gtest/gtest.h>
@@ -1560,8 +1560,7 @@ TEST(TestRGWAdmin, bilog_trim) {
 }
 
 int main(int argc, char *argv[]){
-  vector<const char*> args;
-  argv_to_vec(argc, (const char **)argv, args);
+  auto args = argv_to_vec(argc, argv);
 
   auto cct = global_init(NULL, args, CEPH_ENTITY_TYPE_CLIENT,
 			 CODE_ENVIRONMENT_UTILITY,
